define(["properties/App.Settings", "properties/App.Properties", "system/power/PowerInterface", "system/SystemInterface", "system/io/FileEx", "system/io/Directory", "models/RebootReasons"], function(Settings, Properties, PowerInterface, SystemInterface, File, Directory, RebootReasons) {
    function ApplicationUpdateController() {
        var self = this;

        /**
         * Starts the app update process
         * @return None
         */
        self.start = function() {
            logger.info('App Update process started.');
            isUpgradeToIPKRequired().done(function(upgradeRequired) {
                if (upgradeRequired) {
                    upgradeToIPK();
                } else {
                    setTimeout(function() {
                        checkForUpdate();
                        setInterval(checkForUpdate, Settings.application.remoteApplicationUpdates.frequency);
                    }, 30000);
                }
            });
        };

        /**
         * This method checks for an app update and will call updateApplication()
         * to perform the update if required.
         * The checking process uses AJAX for clients that do not support SCAP 1.2+ but uses
         * copyFile + readFile for those that do. This fixes an issue with the inode count increasing
         * on the panel which can ultimately render the filesystem useless.
         * @return {[type]} [description]
         */
        function checkForUpdate() {
            logger.info('Checking if application update is required...');
            if (Settings.application.sdkVersion < 1.2) {
                logger.warning("WARNING: Client does not support SCAP 1.2. Using AJAX to request XML file. This may result in inode count increase. Please update the firmware to a version that supports SCAP 1.2.");
                checkForUpdatePreSCAP12();
            } else {
                logger.info("SCAP 1.2 Supported. Downloading XML config and checking for update...");
                checkForUpdateSCAP12Plus();
            }
        }

        function isUpgradeToIPKRequired() {
            var dfd = $.Deferred();
            logger.info('Checking if IPK upgrade is required');
            if (Settings.application.webOSVersion <= 2) {
                logger.warning("IPK Upgade not supported for supported for model " + Settings.application.model + " with webOS version " + Settings.application.webOSVersion + ". Application will continue to run as a zip package. Some advanced application feature such as support for synchronisation will be unavailable.");
                setTimeout(function() {
                    dfd.resolve(false);
                }, 100);
            } else {
                SystemInterface.getServerProperties().done(function(properties) {
                    if (properties.appType && (properties.appType == "zip" || properties.appType == "ZIP")) {
                        logger.info('IPK Upgrade required');
                        dfd.resolve(true);
                    } else {
                        logger.info('IPK Upgrade not required.');
                        try {
                            logger.info('Server settings appType: ' + properties.appType);
                        } catch (e) {

                        }
                        dfd.resolve(false);
                    }
                }).fail(function() {
                    logger.error('Unable to determine if IPK upgrade is required - failed to load server properties');
                    dfd.resolve(false);
                });
            }
            return dfd;
        }

        function upgradeToIPK() {

            logger.info('Updated server settings to IPK format.');
            // Copying data files from local storage to normal file storage (otherwise app will effectively reset)
            logger.info('Copying data files from local storage to file system...');
            File.read(Settings.application.fileSystem.systemFiles.settingsFilePath).done(function(settingsFileData) {
                File.read(Settings.application.fileSystem.systemFiles.clientIdFilePath).done(function(clientIdFileData) {
                    File.read(Settings.application.fileSystem.systemFiles.clientKeyFilePath).done(function(clientKeyFileData) {
                        File.read(Settings.application.fileSystem.systemFiles.lastConfigFilePath).done(function(lastConfigFileData) {
                            File.read(Settings.application.fileSystem.systemFiles.fileListFilePath).done(function(fileListFileData) {
                                // update file list data
                                if (fileListFileData) {
                                    var files = JSON.parse(localStorage[Settings.application.fileSystem.systemFiles.fileListFilePath]);
                                    files.forEach(function(file) {
                                        if (file.path.lastIndexOf("data/", 0) === 0) {
                                            file.path = file.path.replace("data/", "file://internal/data/");
                                        }
                                    });
                                    fileListFileData = JSON.stringify(files);
                                }
                                // Now Update IPK Setting so that files are written to the file system
                                Settings.application.packageType = 'IPK';

                                Settings.application.fileSystem.dataDirectory = "data";
                                logger.info('Creating data directory');
                                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.dataDirectory).always(function() {
                                    // Update file paths
                                    Settings.application.fileSystem.systemFiles.settingsFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.settingsFilePath;
                                    Settings.application.fileSystem.systemFiles.propertiesFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.propertiesFilePath;
                                    Settings.application.fileSystem.systemFiles.lastConfigFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.lastConfigFilePath;
                                    Settings.application.fileSystem.systemFiles.clientIdFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.clientIdFilePath;
                                    Settings.application.fileSystem.systemFiles.clientKeyFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.clientKeyFilePath;
                                    Settings.application.fileSystem.systemFiles.appVersionFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.appVersionFilePath;
                                    Settings.application.fileSystem.systemFiles.fileListFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.fileListFilePath;
                                    Settings.application.fileSystem.systemFiles.fileListUpgradeFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.fileListUpgradeFilePath;
                                    Settings.application.fileSystem.systemFiles.proofOfPlayFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.proofOfPlayFilePath;
                                    Settings.application.fileSystem.systemFiles.proofOfPlayUploadFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.proofOfPlayUploadFilePath;
                                    // Write files
                                    writeDataFileToFileSystem(Settings.application.fileSystem.systemFiles.settingsFilePath, settingsFileData).done(function() {
                                        writeDataFileToFileSystem(Settings.application.fileSystem.systemFiles.clientIdFilePath, clientIdFileData).done(function() {
                                            writeDataFileToFileSystem(Settings.application.fileSystem.systemFiles.clientKeyFilePath, clientKeyFileData).done(function() {
                                                writeDataFileToFileSystem(Settings.application.fileSystem.systemFiles.lastConfigFilePath, lastConfigFileData).done(function() {
                                                    writeDataFileToFileSystem(Settings.application.fileSystem.systemFiles.fileListFilePath, fileListFileData).done(function() {
                                                        // Update server settings
                                                        logger.info('Updating server settings to switch to IPK format....');
                                                        var options = {};
                                                        options.serverIp = "192.168.0.2";
                                                        options.serverPort = 80;
                                                        options.secureConnection = false;
                                                        options.appLaunchMode = Configuration.AppMode.LOCAL;
                                                        options.appType = "ipk";
                                                        options.fqdnMode = true;
                                                        options.fqdnAddr = Settings.application.remoteApplicationUpdates.location + "com.lg.app.signage.ipk";

                                                        var configuration = new Configuration();
                                                        configuration.setServerProperty(
                                                            function() {
                                                                // Download IPK and then reboot
                                                                logger.info('Copying data files from local storage to file system...');
                                                                updateApplication().fail(function() {
                                                                    // If the update fails then return settings back to what they were
                                                                    Settings.application.packageType = 'ZIP';
                                                                    options.fqdnAddr = Settings.application.remoteApplicationUpdates.location + "lgapp.zip";
                                                                    options.appType = "zip";
                                                                    configuration.setServerProperty(
                                                                        function() {
                                                                            // Download IPK and then reboot
                                                                            logger.info('Reverted server settings');
                                                                        },
                                                                        function() {
                                                                            // body
                                                                            var errorCode = cbObject.errorCode;
                                                                            var errorText = cbObject.errorText;
                                                                            logger.error("Error setting server settings. Error Code [" + errorCode + "]: " + errorText);
                                                                        },
                                                                        options
                                                                    );
                                                                });
                                                            },
                                                            function() {
                                                                // body
                                                                var errorCode = cbObject.errorCode;
                                                                var errorText = cbObject.errorText;
                                                                logger.error("Error setting server settings. Error Code [" + errorCode + "]: " + errorText);
                                                            },
                                                            options
                                                        );
                                                    });
                                                });
                                            });
                                        });
                                    });
                                })
                            });
                        });
                    });
                });
            });
        }

        function writeDataFileToFileSystem(path, data) {
            var dfd = $.Deferred();
            if (data) {
                return File.write(path, data);
            } else {
                setTimeout(function() {
                    dfd.resolve();
                }, 50);
            }
            return dfd;
        }

        /**
         * This method calls out to our servers using AJAX and gets an XML file that has details about the current app version.
         * Due to a bug with the OS this causes the inode count to increase each time the AJAX request is made which will ultimately
         * cause the panel to become unable to persist new data.
         * All panels should be running SCAP 1.2+ but if they are not then the app will fall back to this method.
         * @return None
         */
        function checkForUpdatePreSCAP12() {
            $.ajax({
                url: Settings.application.remoteApplicationUpdates.location + Settings.application.remoteApplicationUpdates.configFileName,
                contentType: 'application/xml',
                timeout: 10000,
                cache: false
            }).done(function(data, textStatus, jqXHR) {
                // File Received - get version information
                var updatedVersion = $(data).find('application version').first().text();
                var updatedBuildNumber = $(data).find('application buildnumber').first().text();
                var updatedAppName = $(data).find('application name').first().text();

                // Check the new build against the old build to determine if an update is required
                if (updatedBuildNumber != Properties.appBuildNumber) {
                    logger.info('Application update: REQUIRED. Current version: ' + Properties.appBuildNumber + '. New version: ' + updatedBuildNumber);
                    // Do Update
                    updateApplication(updatedBuildNumber);
                } else {
                    logger.info('Application update: NOT REQUIRED. Current version: ' + Properties.appBuildNumber + '. New version: ' + updatedBuildNumber);
                }
            }).fail(function(jqXHR, textStatus, errorThrown) {
                logger.error('Error retrieving update file ' + errorThrown);
            });
        }

        /**
         * This method checks for an update on panels that support SCAP 1.2+.
         * This allows us to not use an AJAX request which causes the inode count to increase.
         * Instead this uses the copyFile() and readFile() method from the SCAP API. This means
         * that the same file is always overwritten and read back from local so that inode count for this process
         * is only ever 1.
         * @return None
         */
        function checkForUpdateSCAP12Plus() {
            var remoteUrl = Settings.application.remoteApplicationUpdates.location + Settings.application.remoteApplicationUpdates.configFileName;
            var localUrl = Settings.application.fileSystem.root + "/" + Settings.application.remoteApplicationUpdates.configFileName;
            logger.info('App Updater: Downloading XML Config File from: ' + remoteUrl + ' to ' + localUrl);
            new Storage().copyFile(
                // Download Succeeded
                function(result) {
                    logger.info('App Updater: Downloaded XML Config File: ' + localUrl);
                    // Read the file - Note this is a small file so no need to worry about buffer limitations
                    new Storage().readFile(
                        // Success Callback
                        function(result) {
                            // Get the data
                            var data = result.data;
                            logger.info('App Updater: Successfully read XML Data from local: ' + data);
                            // parse as XML
                            var xmlDoc = $.parseXML(data);
                            // Get the version and build number
                            var updatedVersion = $(xmlDoc).find('application version').first().text();
                            var updatedBuildNumber = $(xmlDoc).find('application buildnumber').first().text();
                            var updatedAppName = $(xmlDoc).find('application name').first().text();
                            // Check the new build against the old build to determine if an update is required
                            if (updatedBuildNumber != Properties.appBuildNumber) {
                                logger.info('Application update: REQUIRED. Current version: ' + Properties.appBuildNumber + '. New version: ' + updatedBuildNumber);
                                // Do Update
                                updateApplication(updatedBuildNumber);
                            } else {
                                logger.info('Application update: NOT REQUIRED. Current version: ' + Properties.appBuildNumber + '. New version: ' + updatedBuildNumber);
                            }
                        },
                        // Failed Callback
                        function(error) {
                            logger.error('App Updated: Error whilst reading XML Config File from: ' + localUrl + '. Error: [' + result.errorCode + '] ' + result.errorText);
                        },
                        // Options
                        {
                            path: localUrl,
                            encoding: 'utf8',
                            position: 0

                        }
                    );
                },
                // Download Failed
                function(result) {
                    logger.error('App Updated: Error whilst downloading XML Config File from: ' + remoteUrl + '. Error: [' + result.errorCode + '] ' + result.errorText);
                },
                // Copy FilesOptions
                {
                    source: remoteUrl,
                    destination: localUrl
                }
            );
        }

        /**
         * This method performs the application update.
         * The panel will download the new zip archive, install it and restart itself.
         * @return None
         */
        function updateApplication(appBuildNumber) {
            var dfd = $.Deferred();
            // Update Options         
            var options = {
                to: Storage.AppMode.LOCAL,
                recovery: true
            };
            // Do Update
            new Storage().upgradeApplication(
                function(upgradeSuccessResult) {
                    // Updated Succeeded so restart the panel
                    logger.info('Application updated successfully. Attempting restart...');
                    PowerInterface.reboot(RebootReasons.APPLICATION_UPDATE, appBuildNumber);
                },
                function(upgradeFailedResult) {
                    logger.error("Failed to update application: [" + upgradeFailedResult.errorCode + "] " + upgradeFailedResult.errorText);
                    dfd.reject();
                },
                options
            );
            return dfd;
        }
    }

    return ApplicationUpdateController;
});