define(["properties/App.Settings", "models/FirmwareUpdateStatusConstants", "system/SystemInterface", "system/storage/StorageInterface", "core/SLAPIMethodsInterface", "models/API"], function(Settings, FirmwareUpdateStatusConstants, SystemInterface, StorageInterface, APIMethodsInterface, API) {
	function FirmwareUpdateController() {
		var self = this;

		var _model = null;
		var _currentFirmwareVersion = null;
		var _started = false;
		var _updateStatusCheckIntervalId = null;
		var _performingUpdate = false;
		var _updateCheckIntervalId = null;

		var _APIMethodsInterface = new APIMethodsInterface();

		self.start = function() {
			var dfd = $.Deferred();
			// If the controller is already running then just return the resolved promise
			if (_started === true) {
				logger.warning('Attempted to start FW Update controller but process is already running');
				return dfd.resolve();
			}
			// Check remote firmware updates are supported and enabled
			if (Settings.application.sdkVersion < 1.2) {
				logger.warning('Attempted to start FW Update controller but firmware updates are not supported or not enabled.');
				return dfd.resolve();
			}
			// Start updates
			logger.info('Starting firmware update process...');
			// Get the model
			SystemInterface.getModel()
				.done(function(model){
					_model = model;
					logger.info("Successfully loaded model data: " + _model);
					// Get the firmware version
					SystemInterface.getFirmwareVersion()
					.done(function(fwVersion) {
						_currentFirmwareVersion = fwVersion;
						logger.info("Successfully loaded firmware version data: " + _currentFirmwareVersion);
						// Resolve the firmware upgrade controller has started
						_started = true;
						dfd.resolve();

					})
					.fail(function() {
						logger.error("Error starting firmware update process. Unable to determine current firmware version that is installed.");
						dfd.reject();
					});
				})
				.fail(function() {
					logger.error("Error starting firmware update process. Unable to determine model of panel.");
					dfd.reject();
				});
			return dfd;
		};

		self.stop = function() {
			logger.info("Stopping firmware update controller...");
			if (!_performingUpdate) {
				clearInterval(_updateStatusCheckIntervalId);
				clearTimeout(_updateCheckIntervalId);
				_started = false;
				logger.info("Firmware Controller stopped.");
			}
			else {
				logger.error("Cannot stop firmware update controller because an update is currently in progress.");
			}
		};

		self.isFirmwareUpdateRequired = function(firmwareUpdate) {
			// Check Remote FW Updates are enabled	
			if (Settings.application.sdkVersion < 1.2 || Settings.application.remoteFirmwareUpdates.enabled == false) {
				logger.info('Will not check firmware update. Remote firmware updates are disabled');
				return false;
			}
			logger.info('Checking if firmware update is required...');
			// Check we are able to determine player fw family and version
			if (!_model) {
				logger.warning('Unable to determine player model. Unable to continue to check firmware update.');
				return false;
			}			
			if (!_currentFirmwareVersion) {
				logger.warning('Unable to determine player firmware version. Unable to continue to check firmware update.');
				return false;
			}
			// Compare player firmware family and version with firmwareUpdate family and version
			if (convertFirmwareVersionFromStringToNumber(firmwareUpdate.version) > convertFirmwareVersionFromStringToNumber(_currentFirmwareVersion)) {
				logger.info('Firmware update is required from ' + _currentFirmwareVersion + ' to ' + firmwareUpdate.version);
				return true;
			}
			else {
				logger.warning('Firmware update is not required or invalid from current installed version ' + _currentFirmwareVersion + ' to ' + firmwareUpdate.version);
				return false;
			}
		};

		self.updateFirmware = function(firmwareUpdate) {
			logger.info('Starting firmware update...');
			var dfd = $.Deferred();
			notifyFirmwareDownloadRequested(firmwareUpdate.id)
			downloadFirmware(firmwareUpdate)
				.done(function() {
					logger.info('Firmware downloaded');
					dfd.resolve();
				})
				.fail(function() {
					dfd.reject();
				});
			return dfd;
		};

		self.cancelUpdate = function() {
			logger.info("Cancelling requested firmware update as it is not required.");
			notifyFirmwareUpdateCompleted();
		};

		function downloadFirmware(firmwareUpdate) {
			var dfd = $.Deferred();
			// Check there is enough space (allow 800mb - most firmwares are < 500mb)
			logger.debug("Checking available disk space...");
			StorageInterface.getTotalStorageAvailable()
				.done(function(freeSpace){
					logger.info("Free space: " + freeSpace + "MB");
					logger.info("Required space: " + Settings.application.remoteFirmwareUpdates.requiredDiskSpace + "MB");
					// If not enough space then cancel update
					if (freeSpace < Settings.application.remoteFirmwareUpdates.requiredDiskSpace) {
						logger.error("Error updating firmware. Insufficient available space to download firmware file.");
						clearInterval(_updateStatusCheckIntervalId);
						_performingUpdate = false;
						return dfd.reject();
					}
					// Download update
					logger.info("Finished checking available disk space. Starting download process.");
					// Set url
					var options = {
						uri : firmwareUpdate.url
					};
					// Start downloading the firmware
					logger.info("Firmware download started. Polling Update status every 15 seconds");
					// Poll for firmware upgrade status every 30 seconds
					_updateStatusCheckIntervalId = setInterval(function() {
						getFirmwareUpdateStatus();
					}, 15000);
					// Start firmware download
					SystemInterface.downloadFirmware(options)
						.done(function() {
							logger.info('Firmware downloaded successfully.');
							dfd.resolve();
						})
						.fail(function(){
							logger.error("Error attempting to start to download firmware file.");
							clearInterval(_updateStatusCheckIntervalId);
							_performingUpdate = false;
							dfd.reject();
						});
				})
				.fail(function(){
					logger.error("Error updating firmware. Unable to determine available disk space.");
					clearInterval(_updateStatusCheckIntervalId);
					_performingUpdate = false;
					dfd.reject();
				});
			return dfd;
		};

		function performFirmwareUpdate() {
			var dfd = $.Deferred();
			logger.info("Starting firmware installation...");		
			SystemInterface.upgradeFirmware()
				.done(function(){
					logger.info("Successfully started firmware installation");
					notifyFirmwareUpdateCompleted();
					dfd.resolve();
				})
				.fail(function(){
					logger.error("Failed to start firmware upgrade");
					notifyFirmwareUpdateCompleted();
					dfd.reject();
				});
			return dfd;
		}

		function getFirmwareUpdateStatus() {
			SystemInterface.getFirmwareUpgradeStatus()
				.done(function(result) {
					switch(result.status) {
						case FirmwareUpdateStatusConstants.READY:
							logger.info("Firmware downloaded and ready to be installed. Starting firmware installation..");
							performFirmwareUpdate()
								.alyways(function() {
									// Don't need to do anything here
								});
							break;
						case FirmwareUpdateStatusConstants.DOWNLOADING:
							logger.info("Firmware download progress: " + result.downloadProgress +"%");
							break;
						case FirmwareUpdateStatusConstants.IN_PROGRESS:
							logger.info("Firmware download progress: " + result.upgradeStatus + "%");
							break;
						case FirmwareUpdateStatusConstants.COMPLETED:
						case FirmwareUpdateStatusConstants.FAILED:
						case FirmwareUpdateStatusConstants.IDLE:
							// Update process has completed or failed
							// stop checking status
							clearInterval(_updateStatusCheckIntervalId);
							_performingUpdate = false;
							notifyFirmwareUpdateCompleted();
							break;
					}
				});
		}

		function convertFirmwareVersionFromStringToNumber(fwVersionString) {
			try {
				// String is in the format 03.00.90
				// This always works because LG never break this format
				var fwNumber = parseInt(fwVersionString.split('.').join(""));
				return fwNumber;
			}
			catch (e) {
				logger.error("Error converting version string to number: " + e.message);
				return null;
			}
		}

		function notifyFirmwareDownloadRequested(objectId) {
            notify({
                type: "DownloadStart",
                object_id: objectId
            });
		}	

		function notifyFirmwareUpdateCompleted() {
            notify({
                type: "FirmwareUpdateComplete"
            });
		}
		
		function notify(data) {
			logger.info('Notify ' + data.type + ', Object Id: ' + data.object_id);
            var jqXHR = executeAPICall('notifications', {
                type: 'POST',
                data: data
            });
            jqXHR.done(function(data, textStatus, jqXHR) {
                logger.info('Notify Response Code ' + jqXHR.status);
            }).fail(function(jqXHR, textStatus, errorThrown) {
                logger.error('Error sending notification');
            });
		}
		
		function executeAPICall(operationName, settings) {
			return _APIMethodsInterface.executeAPICall(API.PLAYER_API, operationName, settings);
        }
	}

	return FirmwareUpdateController;
});
