define(["controllers/InterruptController", "properties/App.Settings", "system/sensors/GPIONotificationInterface"], function(InterruptController, Settings, GPIONotificationInterface) {
	function GPIOBasedInterruptsController(){

		var self = this;

		var _started = false;
		var _initialized = false;
		var _options = null;

		/**
		 * initializes the module
		 * @param  {any initialization options for this module}
		 * @return {Jquery Promise}
		 */
		self.initialize = function(options){
			var dfd = $.Deferred();

			logger.debug("Initializing GPIO Interrupt Controller");

			_options = options;
				
			GPIONotificationInterface.onStateChange.addListener(GPIOStateChangeHandler)
			.done(function() {
				logger.debug("GPIO Interrupt Controller Initialized");
				_initialized = true;
				dfd.resolve();
			})
			.fail(function()
			{
				logger.error("Error initializing GPIO Interrupt Controller");
				dfd.reject();
			});

			return dfd;
		};

		/**
		 * Handles GPIO state changes
		 * @param {Array of GPIOPins}
		 */
		function GPIOStateChangeHandler(pins)
		{
			logger.debug("GPIOBasedInterruptsController: GPIO Pin states changed.");

			if (_started)
			{
				// Iterate over all pins and attempt to fire an interrupt
				// for each pin
				$.each(pins, function(index, pin) {
					// Work out the trigger key based on pin details
					var triggerKey = "GPI" + pin.number + pin.state;

					// Attempt to trigger an interrupt (if one exists) for each pin
					// For the time being play to set duration
					InterruptController.triggerInterrupt(triggerKey, _options.trigger.type, false, false)
					.done(function() {
						logger.debug("Playing Interrupt for GPIO Key: " + triggerKey);
					})
					.fail(function()
					{
						logger.debug("No interrupt found for GPIO Key: " + triggerKey);
					});
				});
			}
		}

		/**
		 * Starts the module
		 * @return {Jquery Promise}
		 */
		self.start = function()
		{
			var dfd = $.Deferred();

			setTimeout(function()
			{
				if (self.isInitialized())
				{
					_started = true;
					logger.debug("GPIO Interrupt Controller Started");
					dfd.resolve();
				}
				else
				{
					logger.error("Failed to start GPIO Interrupt Controller. Can not start a module that is not already initialized");
					dfd.resolve();
				}
				
			}, 50);

			return dfd;
		};
		
		/**
		 * Stops the module
		 * @return {Jquery Promise}
		 */
		self.stop = function()
		{
			var dfd = $.Deferred();

			// Only stop the ocntroller is it is already started
			if (self.isRunning())
			{
				GPIONotificationInterface.onStateChange.removeListener()
				.done(function() {
					_started = false;
					logger.debug("GPIO Interrupts Controller Stopped.");
					dfd.resolve();
				})
				.fail(function()
				{
					logger.error("Error stopping GPIO Interrupt Controller");
					dfd.reject();
				});
			}
			else
			{
				logger.warning("Stop called on GPIO Interrupts Controlle but controller not started.");
				dfd.reject();
			}

			return dfd;
		};

		/**
		 * Returns true if module is initizialized
		 * @return {Boolean}
		 */
		self.isInitialized = function()
		{
			return _initialized;
		};

		/**
		 * Returns true if sensor is available. In this case
		 * if initialization was successful we know the sensor was available.
		 * No other way to test this.
		 * @return {Boolean}
		 */
		self.isAvailable = function()
		{
			if (self.isInitialized())
				return true;
			else
				return false;
		};

		/**
		 * Returns true if the module has been started, false otherwise
		 * @return {Boolean}
		 */
		self.isRunning = function()
		{
			return _started;
		};

		/**
		 * Returns true if supported in this client, false otherwise
		 * @return {Boolean}
		 */
		self.isSupported = function()
		{
			return Settings.application.modules.GPIOBasedInterrupts.supported;
		};

		/**
		 * Returns true if the module is enabled, false otherwise
		 * @return {Boolean}
		 */
		self.isEnabled = function()
		{
			return Settings.application.modules.GPIOBasedInterrupts.supported;
		};
	}
	
	return new GPIOBasedInterruptsController();
}); 