define(["controllers/InterruptController", "system/sensors/IdleNotificationInterface", "models/IdleState", "properties/App.Settings"], function(InterruptController, IdleNotificationInterface, IdleState, Settings) {
	function InteractiveKioskController(){

		var self = this;

		var _started = false;
		var _initialized = false;
		var _options = null;

		// We should look to add options to SL specifically so that content can be published
		// and defined to be triggered by idle/active status. Until then the trigger mechanism
		// is set in the application settings
		var _triggerKey = null;
		var _triggerType = null;

		/**
		 * Initialize Interactive Kiosk Mode
		 * @param  {[type]} options [description]
		 * @return {[type]}         [description]
		 */
		self.initialize = function(options){
			var dfd = $.Deferred();

			logger.debug("Configuring Interactive Kiosk Mode");

			// Configure module options
			_options = options;
			_triggerKey = _options.trigger.key;
			_triggerType = _options.trigger.type;

			// Set Idle Interval (note this method never returns a failure)
			IdleNotificationInterface.setIdleInterval(Settings.application.modules.interactiveKiosk.options.idleIntervalInSeconds)
			.always(function(){

				// Register Callback function on idle state change
				IdleNotificationInterface.onStateChange.addListener(idleStateChangeHandler)
				.done(function()
				{
					logger.debug("Idle State Change Handler Method registered.");

					// Check the current state
					IdleNotificationInterface.queryState(Settings.application.modules.interactiveKiosk.options.idleIntervalInSeconds)
					.done(function(currentState)
					{
						// If currents state is active then play the interactive content
						if (currentState === "active")
						{
							playInterrupt(_triggerKey, _triggerType, true);
						}

						// Mark as initialized
						_initialized = true;

						// Resolve initialization
						dfd.resolve();
					});
				})
				.fail(function(){
					logger.error("Failed to register idle state change handler for Interactive kiosk mode");
					dfd.reject();
				});
			})
			.fail(function(){
				logger.error("Failed to set Idle Interval for Interactive kiosk mode");
				dfd.reject();
			});

			return dfd;
		};

		/**
		 * Handle Idle State Changes for Interactive Kiosk Mode
		 * @param  {[type]} newState [description]
		 * @return {[type]}          [description]
		 */
		function idleStateChangeHandler(newState)
		{
			logger.debug("idleStateChangeHandler called");
			if (_started && newState != null)
			{
				switch (newState)
				{
					case "active":
						playInterrupt(_triggerKey, _triggerType, true);
						break;
					case "idle":
						playScheduledContent();
						break;
					case "locked":
						// DO Nothing
						break;
				}
			}
		}

		/**
		 * Attempts to trigger an interupt using the interrupt controller
		 * @param  {[type]} triggerKey       [description]
		 * @param  {[type]} triggerType      [description]
		 * @param  {[type]} playIndefinitely [description]
		 * @return {[type]}                  [description]
		 */
		function playInterrupt(triggerKey, triggerType, playIndefinitely)
		{
			InterruptController.triggerInterrupt(triggerKey, triggerType, playIndefinitely, false)
			.done(function() {
				logger.info("Playing Interactive Kiosk Active Content");
			})
			.fail(function() {
				logger.error("Failed to load Interactive Kiosk Active Content");
			});
		}

		/**
		 * Returns to standard scheduled content
		 * @return {[type]} [description]
		 */
		function playScheduledContent()
		{
			if(!InterruptController.isPlayingInterrupt() || (InterruptController.isPlayingInterrupt() && !InterruptController.currentInterruptIsHighPriority()))
			{
				InterruptController.cancelInterrupt()
				.done(function() {
					logger.info("Interrupt cancelled. Playing Scheduled Content");
				})
				.fail(function() {
					logger.error("Failed to cancel interrupt content");
				});
			}
			else
			{
				logger.warning("Attempted to cancel current interrupt but either no interrupt is playing or the current interrupt can not be cancelled.");
			}

		}

		/**
		 * Starts the module
		 * @return {Jquery Promise}
		 */
		self.start = function()
		{
			var dfd = $.Deferred();

			setTimeout(function()
			{
				if (self.isInitialized())
				{
					_started = true;
					logger.debug("Interactive Kiosk Controller Started");
					dfd.resolve();
				}
				else
				{
					logger.error("Failed to start Interactive Kiosk Controller. Can not start a module that is not already initialized");
					dfd.resolve();
				}
				
			}, 50);

			return dfd;
		};
		
		/**
		 * Stops the module
		 * @return {Jquery Promise}
		 */
		self.stop = function()
		{
			var dfd = $.Deferred();

			// Only stop the controller is it is already started
			setTimeout(function(){

				if (self.isRunning())
				{
					_started = false;
					logger.debug("Interactive Kiosk Stopped.");
					dfd.resolve();
					
				}
				else
				{
					logger.warning("Stop called on Interactive Kiosk Controller but controller not started.");
					dfd.reject();
				}
			});

			return dfd;
		};

		/**
		 * Returns true if module is initizialized
		 * @return {Boolean}
		 */
		self.isInitialized = function()
		{
			return _initialized;
		};

		/**
		 * Returns true if sensor is available. In this case
		 * if initialization was successful we know the sensor was available.
		 * No other way to test this.
		 * @return {Boolean}
		 */
		self.isAvailable = function()
		{
			if (self.isInitialized())
				return true;
			else
				return false;
		};

		/**
		 * Returns true if the module has been started, false otherwise
		 * @return {Boolean}
		 */
		self.isRunning = function()
		{
			return _started;
		};

		/**
		 * Returns true if supported in this client, false otherwise
		 * @return {Boolean}
		 */
		self.isSupported = function()
		{
			return Settings.application.modules.GPIOBasedInterrupts.supported;
		};

		/**
		 * Returns true if the module is enabled, false otherwise
		 * @return {Boolean}
		 */
		self.isEnabled = function()
		{
			return Settings.application.modules.GPIOBasedInterrupts.supported;
		};
	}
	
	return new InteractiveKioskController();
}); 