define(["models/ContentModel", "models/InterruptModel", "models/ActionModel", "models/TriggerModel", "properties/App.Settings", "system/power/PowerInterface"], function(ContentModel, InterruptModel, ActionModel, TriggerModel, Settings, PowerInterface) {
	function InterruptController(){
		var self = this;

		var _mediaPlayer = null;
		var _interrupts = null;
		var _scheduleTimerId = null;
		var _isPlayingInterrupt = false;
		var _highPriorityInterruptPlaying = false;
		var _screenControl = null;
		var _indefiniteInterruptPlaying = false;
		var _currentlyPlayingInterrupt = null;

		/**
		 * Initialize the interrupt Controller
		 * @param  {Reference to media player}
		 * @param  {Interrupt Model}
		 * @return {JQuery Promise}
		 */
		self.initialize = function(mediaPlayer, interrupts, screenControl){
			var dfd = $.Deferred();

			setTimeout(function()
			{
				_mediaPlayer = mediaPlayer;
				_interrupts = interrupts;
				_screenControl = screenControl;
				dfd.resolve();

			}, 50);

			return dfd;
		};

		self.triggerCancel = function(cancelKey, triggerType)
		{
			var dfd = $.Deferred();

			setTimeout(function() {

				if(_currentlyPlayingInterrupt != null && _currentlyPlayingInterrupt.trigger.cancelKey)
				{
					var currentInterruptCancelKey = _currentlyPlayingInterrupt.trigger.cancelKey.trim();
					cancelKey = cancelKey.trim();

					//Make sure both are lowercase for keyboard
					if(triggerType == "KeyboardEvent")
					{
						currentInterruptCancelKey = currentInterruptCancelKey.toLowerCase();
						cancelKey = cancelKey.toLowerCase();
					}

					if(currentInterruptCancelKey == cancelKey && triggerType == _currentlyPlayingInterrupt.trigger.type)
					{
						self.cancelInterrupt().always(function(){
							dfd.resolve();
						});
					}
				}else {
					dfd.reject("No interrupt is playing");
				}				
			},5);

			return dfd;
		}

		/**
		 * Attempts to trigger an interrupt associated with the provided trigger key and trigger type
		 * @param  triggerKey - The Trigger Key
		 * @param  triggerType - The Trigger Type (KeyboardEvent)
		 * @param  playIndefinitely - Whether or not to play the interrupt content until cancelled or until specified duration
		 * @param  highPriority - If true this interrupt will take precendence over low priority interrupts and will also trigger over other
		 *                        high priority interrupts
		 * @return JQuery Promise
		 */
		self.triggerInterrupt = function(triggerKey, triggerType, playIndefinitely, highPriority, triggerAdditionalData)
		{
			var dfd = $.Deferred();

			try
			{
				setTimeout(function() {
					
					// Only trigger the interrupt if:
					// 1: No current interrupt is playing
					// 2: An interrupt is playing but the new interrupt is high priority (high priority can only be overriden by other high priority interrupts)
					// 3: The current interrupt is low priority (low priority can be overridden by other low priorty and high priorty interrupts)
					// AND
					// Screen is scheduled to be on OR screen cannot be switched off
					readLastScreenStatus().then(function (screenInformation) {
						if ((!_isPlayingInterrupt || (_isPlayingInterrupt && highPriority) || (_isPlayingInterrupt && !_highPriorityInterruptPlaying)) && (!Settings.application.features.canSwitchScreenOnOff || (Settings.application.features.canSwitchScreenOnOff && (_screenControl == null || _screenControl.isScreenCurrentlyScheduledOn(screenInformation)))))
						{
							var interruptToPlay = getInterruptToPlay(triggerKey, triggerType)

							if (interruptToPlay !== null)
							{
								// clear the timeout
								clearTimeout(_scheduleTimerId);

								// Remember that an interrupt is playing
								_isPlayingInterrupt = true;

								if (highPriority == "undefined")
								{
									_highPriorityInterruptPlaying = true;
								}
								else
								{
									_highPriorityInterruptPlaying = highPriority;
								}

								// Start to play interrupt
								playInterrupt(interruptToPlay.action.layout, triggerAdditionalData);
								_currentlyPlayingInterrupt = interruptToPlay;

								// If not set to play indeinitely then play until for the duration of the interrupt content
								if (!playIndefinitely)
								{
									_scheduleTimerId = setTimeout(self.cancelInterrupt, interruptToPlay.action.duration * 1000)
								}

								dfd.resolve();
							}
							else
							{
								logger.warning("Unable to find requested interrupt. Interrupt Key: " + triggerKey + ", Type: " + triggerType);
								dfd.reject();
							}
						}
						else
						{
							logger.warning("Unable to trigger interrupt because an interrupt that cannot be cancelled is currently playing");
							dfd.reject();
						}
					});
				}, 50);
			}
			catch (e)
			{
				logger.error("Error trying to play interrupt.");
				dfd.reject();
			}
			
			return dfd;
		};

		/**
		 * Returns true if an interrupt is currently playing, false otherwise
		 * @return {Boolean}
		 */
		self.isPlayingInterrupt = function()
		{
			return _isPlayingInterrupt;
		}

		/**
		 * Returns true if the current Interrupt can be overriden, false otherwise
		 * @return {Boolean}
		 */
		self.currentInterruptIsHighPriority = function()
		{
			return _highPriorityInterruptPlaying;
		}

		/**
		 * Used to cancel the current interrupt
		 * @return {JQuery Promise}
		 */
		self.cancelInterrupt = function()
		{
			var dfd = $.Deferred();

			// Cancel Interrupt
			setTimeout(function() {
				_currentlyPlayingInterrupt = null;
				// clear the timeout
				clearTimeout(_scheduleTimerId);

				if (_isPlayingInterrupt) {
					playScheduledContent();
					_isPlayingInterrupt = false;
					_highPriorityInterruptPlaying = false;
					logger.debug("Interrupt Cancelled");
					dfd.resolve();
				}
				else {
					logger.warning("Attempted to cancel interrupt but no interrupt is currently playing or current interrupt cannot be cancelled");
					dfd.reject();
				}
				
			}, 50);

			return dfd;
		};
 
		/**
		 * Start playing a layout
		 * @param  {Layout Model of the interrupt to play}
		 */
		function playInterrupt(layoutToPlay, triggerAdditionalData)
		{
			_mediaPlayer.playInterrupt(layoutToPlay, triggerAdditionalData);
		}

		/**
		 * Screen Control - Read last screen state command sent by WBT API
		*/
		function readLastScreenStatus() {
			var dfd = $.Deferred();
			// Read from file and switch screen on or off based on persisted file state
			PowerInterface.readLastScreenStatus().then(function (screenInformation) {
				dfd.resolve(screenInformation);
			});
			return dfd;
		}

		/**
		 * Switch to the default scheduled content
		 */
		function playScheduledContent()
		{
			// Load content
			_mediaPlayer.playScheduledContent();
		}

		/**
		 * Find an interrupt matching trigger key and trigger type
		 * @param  {The trigger key}
		 * @param  {The trigger type}
		 * @return {The interrupt model of matching interrupt of null if none found}
		 */
		function getInterruptToPlay(triggerKey, triggerType)
		{
			var interruptToPlay = null;

			if(_interrupts)
			{
				for (var i = 0; i < _interrupts.length; i++)
				{
					var interrupt = _interrupts[i];
					if (interrupt.trigger.triggerKey.toLowerCase() === triggerKey.toLowerCase() && interrupt.trigger.type === triggerType)
					{
						interruptToPlay = interrupt;
						break;
					}
				}
			}
			return interruptToPlay;
		}
	}
	
	return new InterruptController();
}); 