define(["system/network/NetworkInterface"], function(NetworkInterface) {
	function MediaSyncController() {
		var self = this;
        var isMaster = false;
        var masterIPBroadcastTimer = null;
        //Is sync running?
        var syncRunning = false;
        var currentContentIndex = -1;
        //IP address of master
        var masterIPAddress = "";
        //Sync group ID
        var currentSyncGroupID = 0;
        //How often the master will send the masteralive event
        var masterAliveInterval = 17200;
        //Time in ms between IP broadcasts
        var masterIPBroadcastInterval = 30000;

        var playAsset = null;
        //Time after receiving the sync signal to play content
        var contentLoadingTime = 0.40; // Delay second for loading content time (espcially video, HTML)     
        var findIpTimer = null;

        //Time in seconds we allow for a socket connection before deciding it has failed
        var socketConnectTimeout = 15;
        var socketConnectTimer = null;
        var syncFailedTimer = null;
        //Time in seconds that the slave will expect a masteralive event within
        var syncFailedTimeout = 40;

        self.startMaster=function(syncGroupID,playAssetCB)
        {
            currentSyncGroupID = syncGroupID;
            isMaster = true;
            playAsset = playAssetCB;
            logger.debug("Getting IP");

            findIpTimer = setInterval(function()
            {
                NetworkInterface.getIP().done(function(ip){

                    if(ip)
                    {

                        //Cancel the interval timer
                        clearInterval(findIpTimer);

                        logger.debug("IP is " + ip);
                        masterIPAddress = ip;

                        //Start the time to send at intervals of 30 seconds
                        masterIPBroadcastTimer = setInterval(function()
                        {
                            broadcastIp(syncGroupID,ip);
                        }, masterIPBroadcastInterval);
        
                        //Broadcast now
                        broadcastIp(syncGroupID,ip);
                    } else {
                        log.info("Network interface not ready, wait another 5 seconds");                      
                    }
                }).fail(function()
                {
                    log.info("Network interface not ready, wait another 5 seconds");                    
                });
            },5000);

            //Start playback
            startSyncPlayback("localhost");          
        };
        
        self.startSlave=function(syncGroupID,playAssetCB)
        {
            isMaster = false;
            playAsset = playAssetCB;
            currentSyncGroupID = syncGroupID;

            findIpTimer = setInterval(function()
            {
                //Get IP from network interface. It is is ready we can try syncing
                NetworkInterface.getIP().done(function(ip){
                    //Cancel the interval timer
                    clearInterval(findIpTimer);
                    logger.debug("Network interface has an IP, listening for broadcasts");
                    startListeningForMaster(syncGroupID);                    
                }).fail(function()
                {
                    log.info("Network interface not ready, wait another 2 seconds")                    
                });
            },2000);
        };

        function onSyncServiceLoaded()
        {
            if(isMaster)
            {
                logger.debug("Starting playback with index of -1");
                self.sendPlayAsset(currentContentIndex);
            } else {
                logger.debug("Slave - wait for sync signal");
                syncFailedTimer = setTimeout(syncFailed,syncFailedTimeout * 1000);
            }
        }

        var masterUDPListeningRequest = null;
        function startListeningForMaster(syncGroupID)
        {
            logger.debug("Requesting service to start listening for IP broadcast");
          
            masterUDPListeningRequest = webOS.service.request('luna://com.lg.app.signage.signagelivesyncsservice', {
                method: 'receive',
                onSuccess: function (msg) {
                    var broadcastObject = JSON.parse(msg.data);
                    if(broadcastObject.syncGroup == syncGroupID)
                    {
                        logger.info("Master IP is " + broadcastObject.ipAddress);

                        masterIPAddress = broadcastObject.ipAddress;

                        stopListeningForMaster();
                        //Start playback
                        startSyncPlayback(masterIPAddress);                  
                    } else {
                        logger.debug("Group broadcast found for Group ID " + broadcastObject.syncGroup + ". My sync group ID is " + syncGroupID);
                    }     
                },
                onFailure: function (err) {
                    logger.error(JSON.stringify(err));
                    logger.debug("Will retry in 10 seconds");
                    setTimeout(function(){
                        startListeningForMaster(currentSyncGroupID);
                    },10000)               
                },
                subscribe: true
            });                      
        }

        function syncConnectionFailed()
        {
            self.stop();

            if(isMaster)            
                startSyncPlayback(masterIPAddress);
            else //Slave must start listening for the master again incase the IP changes, and we need to know if it is online.
                startListeningForMaster(currentSyncGroupID);

        }

        function syncFailed()
        {
            logger.error("Master failed");
            clearTimeout(syncFailedTimer);
            syncConnectionFailed();
        }

        var socket = null;
        function connectToMaster(ipAddr,service)
        {
            var opts = 
            {
                forceNew : true
            };
        
            socket = io('http://'+ ipAddr + ':9990/', opts);

            socketConnectTimer = setTimeout(function()
            {
                logger.error("Failed to connect to sync websocket, retry");

                syncConnectionFailed();
            },socketConnectTimeout*1000);

            socket.on('connect', function () 
            {              
                clearTimeout(socketConnectTimer);
                logger.debug("Connected to master");
       
                if(syncRunning==false)
                 {
                     syncRunning = true;
                     onSyncServiceLoaded();               
                 }
            });     

            self.sendPlayAsset = function(index)
            {
                logger.debug("sendPlayAsset called with index " + index);
                //Might do it without the timeout
                setTimeout(function () {
                    socket.emit('sync',JSON.stringify({
                        currentContentIndex: index,
                    }));
                }, contentLoadingTime * 1000);
            }; 
                
            function displayContent(index) {
                //Save the current index incase sync fails
                currentContentIndex = index;     
                playAsset(index);
            }

            socket.on('masteralive', function()
            {
                if(!isMaster)
                {
                    logger.info("Master is alive");
                    //Restart the sync failed timer
                    clearTimeout(syncFailedTimer);
                    syncFailedTimer = setTimeout(syncFailed,syncFailedTimeout * 1000);
                }
            });

            socket.on('syncevent', function (eventData) {
                              
                var msg = JSON.parse(eventData);
                var contentIndex = msg.currentContentIndex;
                        
                displayContent(contentIndex);                                 
             });            
                
            socket.on('disconnect', function (msg) 
            {
                logger.error("Disconnected: " + msg);
            });

            socket.on('error', function (error)
            {
                logger.error("Socket error: " + error.message);
            });

            socket.on('connect_error', function (error) 
            {      
                logger.error("Socket error: " + error.message);      
            });      

            socket.on('reconnect_error', function (error) 
            {      
                logger.error("Reconnect error: " + error.message);      
            });      

            socket.on('reconnect_failed', function (error) 
            {      
                logger.error("Reconnect failed");      
                //Here we will need to restart sync - TO-DO.
            });      


            self.stopSyncService = function()
            {
                logger.debug("Stopping the sync service");
     
                if(isMaster)
                {
                    try{
                     
                        socket.send("close");
                    } catch(err)
                    {
                       logger.error("Error sending close to service: " + err.message);
                    }
                }
     
                try{
                    socket.disconnect(true);
                } catch(err)
                {
                    logger.error("Error closing client ws connection: " + err.message);
                }

                try{
                    socket.destroy();
                } catch(err)
                {
                    logger.error("Error destroying socket: " + err.message);
                }


                try{
                    if(service != null)
                        service.cancel();
                } catch(err)
                {
                    logger.error("Error closing cancelling service request: " + err.message);
                }

                socket = null;
            };
                     
        }

        function startSyncPlayback(ipAddr)
        {
            logger.debug("Requesting service to start websocket server");

            if(isMaster)
            {
                var request = webOS.service.request('luna://com.lg.app.signage.signagelivesyncsservice', {
                    method: 'serverOn',
                    parameters: {
                        keepAliveInterval: masterAliveInterval,
                    },
                    onSuccess: function () {
                        logger.debug("Sync server on");

                        //Connect to websocket server
                        connectToMaster("localhost",request);
                    },
                    onFailure: function (err) {
                        //If this fails we could have a failure counter, first time it trys again after a few seconds, second time it go back to listening for master, if slave. If master something is wrong!
                        logger.error(JSON.stringify(err));
                        
                        //Try again in 15 seconds
                        setTimeout(function()
                        {
                            startSyncPlayback(ipAddr);
                        },15000);
                    },
                    subscribe: true
                });    
            } else {
                //Slave does not need to start its own server, just the client
                connectToMaster(ipAddr,null);
            } 
        }

        self.stop = function()
        {
            syncRunning = false;

            if(isMaster)
            {
                clearInterval(masterIPBroadcastTimer);
            } else {
                //Slave
                stopListeningForMaster();
                clearTimeout(syncFailedTimer);
            }

            if(self.stopSyncService)
                self.stopSyncService();
        };


        function broadcastIp(syncGroupID, ipAddr)
        {
            webOS.service.request('luna://com.lg.app.signage.signagelivesyncsservice', {
                method: 'broadcast',
                parameters: {
                    ipAddress: ipAddr,
                    syncGroup: syncGroupID
                },
                onSuccess: function()
                {
                    logger.debug("IP Broadcast sent");
                },	
                onFailure: function (err) {
                    logger.debug("Error sending broadcast: " + err.message);
                },
                subscribe: false
            });
        }

        function stopListeningForMaster()
        {
            try{
                if(masterUDPListeningRequest != null)
                {
                    masterUDPListeningRequest.cancel();
                    masterUDPListeningRequest = null;
                }
            } catch(err)
            {
                logger.error("Error cancelling master ipreceive request: " + err.message);
            }
        }
        
        self.isSyncRunning = function()
        {
            return syncRunning;
        };
    
    
	}

	return MediaSyncController;
});
