define(["controllers/InterruptController", "properties/App.Settings", "core/SLWebBasedInterruptsAPIInterface", "models/WebBasedInterruptMessageModel", "models/PusherSubscriptionModel", "models/ScreenControlModel", "system/power/PowerInterface"], function(InterruptController, Settings, SLWebBasedInterruptsAPIInterface, WebBasedInterruptMessageModel, PusherSubscriptionModel, ScreenControlModel, PowerInterface) {
	function WebBasedInterruptsController(){

		var self = this;

		var _started = false;
		var _initialized = false;
		var _isRunning = false;
		var _options = null;
		var _pusherConfiguration = null;
		var _pusher = null;
		var _apiInterface = null;
		var _screenControl = ScreenControlModel;
		var _switchScreenOn = null;
		var _switchScreenOff = null;

		/**
		 * [initialize description]
		 * @param  {[type]} options [description]
		 * @return {[type]}         [description]
		 */
		self.initialize = function(options){
			var dfd = $.Deferred();

			_options = options;
			_switchScreenOn = _options.switchScreenOn;
			_switchScreenOff = _options.switchScreenOff;

			logger.info("Initializing Web based interrupts controller");

			if (_options.serialNo)
			{
				// if we already have a configuration then we can initialize the listeners straight away
				if (_pusher && _pusherConfiguration)
				{
					logger.info("Existing pusher configuration found");
					cancelSubscriptions(_pusherConfiguration);
					configurePusherSubscriptions(_pusherConfiguration);
				}
				
				_initialized = true;
				_started = true;
				dfd.resolve();
			
			}
			else
			{
				setTimeout(function(){
					logger.warning("Error: Could not initialize Web Based Interrupts - Player Serial Number Not defined. Is the player registered?");
					dfd.reject();
				}, 50);
			}

			return dfd;
		};

		/**
		 * [configurationCallbackHandler description]
		 * @param  {[type]} pusherConfiguration [description]
		 * @return {[type]}                     [description]
		 */
		function configurationCallbackHandler(pusherConfiguration)
		{
			logger.info("Web Based Interrupts Controller: Configuration Callback Handler called");

			// unsubsribe from previous channels
			if (_pusherConfiguration != null)
			{
				cancelSubscriptions(_pusherConfiguration);
			}

			_pusherConfiguration = pusherConfiguration;

			// resubscribe based on new configuration
			configurePusherSubscriptions(pusherConfiguration);
		}

		/**
		 * [configurePusherSubscriptions description]
		 * @param  {[type]} pusherConfiguration [description]
		 * @return {[type]}                     [description]
		 */
		function configurePusherSubscriptions(pusherConfiguration)
		{
			logger.info("Web Based Interrupts Controller: Configuring Pusher Subscriptions");

			// We don't want to create a whole new pusher connection if we alredy have one.
			// Disable stats module as this violates CSP on chrome and we don't need it anyway
			if (_pusher === null)
			{
				logger.info("Pusher: Initializing new pusher connection");
				_pusher = new Pusher(pusherConfiguration.pusherKey, { disableStats: true });
			}

			$.each(pusherConfiguration.channels, function(index, channelId){
				var channel = _pusher.subscribe(channelId);
				channel.bind('WebBasedInterrupt', processPusherMessage);
				logger.info("Pusher: Subscribed to channel: " + channelId);
			});
		}

		/**
		 * [unbindPusherSubscriptions description]
		 * @param  {[type]} pusherConfiguration [description]
		 * @return {[type]}                     [description]
		 */
		function cancelSubscriptions(pusherConfiguration)
		{	
			logger.info("Web Based Interrupts Controller: Cancelling Subscriptions");

			$.each(pusherConfiguration.channels, function(index, channelId){
				try {
					_pusher.unsubscribe(channelId);
				}
				catch (error) {
					logger.error(error);
				}
			});
		}

		/**
		 * [processPusherMessage description]
		 * @param  {[type]} msg [description]
		 * @return {[type]}     [description]
		 */
		function processPusherMessage(msg)
		{
			if (_started)
			{
				logger.info("Pusher message received");
				//Cast message
				var message = new WebBasedInterruptMessageModel(msg);

				// Attempt to trigger an interrupt (if one exists) for each pin
				// For the time being play to set duration
				var triggerData = message.data.triggerData;
				switch (triggerData) {
					case "ScreenOn":
						PowerInterface.readLastScreenStatus().then(function(screenInformation) {
							if(!screenInformation || screenInformation && screenInformation.screenStatus !== "ScreenOn") {
								_switchScreenOn(true); //persistState to File
							}
						});
						break;
					case "ScreenOff":
						PowerInterface.readLastScreenStatus().then(function(screenInformation) {
							if(!screenInformation || screenInformation && screenInformation.screenStatus !== "ScreenOff") {
								_switchScreenOff(true); //persistState to File
							}
						});
						break;
					default:
						InterruptController.triggerInterrupt(message.data.triggerData, message.data.triggerType, false, true, message.additionalData)
							.done(function () {
								logger.info("Playing Web-based Interrupt for Key: " + message.data.triggerData);
								if (message.data.additionalData !== null) {
									logger.info("Additional Data included with Web-based Interrupt: " + message.additionalData);
								} else {
									logger.debug("No additional data included with Web-based Interrupt");
								}
							})
							.fail(function () {
								logger.warning("Attempted to play web-based interrupt but failed. No interrupt found for key: " + message.data.triggerData);
							});
						break;
				}
			}
		}

		self.checkForConfigurationUpdate = function()
		{
			var dfd = $.Deferred();

			if (self.isInitialized())
			{
				// API Options
				if (_apiInterface == null)
				{
					_apiInterface = new SLWebBasedInterruptsAPIInterface();			
					var apiOptions = {
						url : _options.url,
						//serialNo : "12345", // Test Serial Number
						serialNo : _options.serialNo,
						configurationCallback : configurationCallbackHandler
					};

					_apiInterface.initialize(_pusherConfiguration, apiOptions)
						.done(function(){
							// Success
							logger.debug("Interrupts API initialized successfully.");
							
							_apiInterface.checkForConfigurationUpdate();

							dfd.resolve();
						})
						.fail(function(){
							// Failed
							logger.error("Error: failed to initialize the Interrupts API..");
							dfd.reject();
						});			
				} else {
					_apiInterface.checkForConfigurationUpdate();
				}
			}

			return dfd;
		};

		
		
		/**
		 * Stops the module
		 * @return {Jquery Promise}
		 */
		self.stop = function()
		{
			var dfd = $.Deferred();

			// Only stop the ocntroller is it is already started
			if (self.isRunning())
			{
				logger.info("Stopping web-based interrupts");
				// Stop
				_started = false;
				_initialized = false;

				// unsubsribe from previous channels
				if (_pusherConfiguration != null)
				{
					cancelSubscriptions(_pusherConfiguration);
				}

				if (_apiInterface)
				{
					_apiInterface.stop();				
				}

				dfd.resolve();

			}
			else
			{
				// Controller not started so we can just resolve this
				logger.warning("Stop called on Web-based Interrupts Controller but controller not started.");
				dfd.resolve();
			}

			return dfd;
		};

		self.isRunning = function()
		{
			return _started;
		};


		/**
		 * Returns true if module is initizialized
		 * @return {Boolean}
		 */
		self.isInitialized = function()
		{
			return _initialized;
		};

		/**
		 * Returns true if sensor is available. In this case
		 * if initialization was successful we know the sensor was available.
		 * No other way to test this.
		 * @return {Boolean}
		 */
		self.isAvailable = function()
		{
			if (self.isInitialized())
				return true;
			else
				return false;
		};

		/**
		 * Returns true if supported in this client, false otherwise
		 * @return {Boolean}
		 */
		self.isSupported = function()
		{
			return Settings.application.modules.webBasedInterrupts.supported;
		};

		/**
		 * Returns true if the module is enabled, false otherwise
		 * @return {Boolean}
		 */
		self.isEnabled = function()
		{
			return Settings.application.modules.webBasedInterrupts.supported;
		};

		
	}
	
	return new WebBasedInterruptsController();
}); 