define(["properties/App.Settings", "system/DateTime", "system/io/File", "core/LogUploadInterfaceHelper"], function(Settings, DateTime, File, LogUploadInterfaceHelper) {   
    function LogUploadInterface() {
        var self = this;

        self.uploadsEnabled = false;

        var pendingConsoleMessages = "";
        var pendingNetworkRequests = "";

        var currentLogFiles = null;

        var playerSerialNumber = null;

        var writeTimeout = null;

        var logFilesList = [];

        var rollLogsOnNextWrite = false;

        // Public Interface
        // ----------------------------------------------------
        self.init = function(loggingConfig, clientDetails) {
            logger.info('Init LogUploadInterface');
            // Currently we are only interested in whether log uploads are enabled. In the enhanced version of log uploads
            // there are additional configuration options that are passed in the player config
            self.uploadsEnabled = loggingConfig && loggingConfig.application && loggingConfig.application.upload;

            // self.uploadsEnabled = false; // FOR TESTING ONLY
            
            // Store serialno as we need this for the S3 keyname
            playerSerialNumber = clientDetails.signageliveSerialNumber;

            // Load log files list from disk into memory
            logger.info('Log Uploads are: ' + self.uploadLogs ? 'Enabled' : 'Disabled');
            return loadLogFileList();
        };

        self.uploadLogs = function() {
            // force log roll on the write to disk which will also result in logs being uploadded.
            logger.info('Forcing log roll and upload.');
            rollLogsOnNextWrite = true;
            // Do the next roll in 5 seconds
            writeTimeout = setTimeout(writePendingLogsToDisk, (5 * 1000));
        };

        // Internal Methods
        // ----------------------------------------------------
        function queueConsoleMessagesForLogging(level, message) {
            // Generate log entry string
            // var paddedLevel = padLevel(level + "]: ");
            // Add to array in memory of console messages pending being written to file
            pendingConsoleMessages += message + "\r\n";
        }

        function queueNetworkRequestsForLogging(evt) {
            // Get important values
            var requestMethod = this.arguments[0];
            var requestUrl = this.arguments[1];
            var requestId = this.requestId;
            // Generate log entry string
            var logEntry = '';
            switch (evt.type) {
                case "loadstart":
                    logEntry = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime, 'STARTED', evt.loaded, 'NA', 'NA', 'NA');
                    addNetworkRequestLogToPendingRequestLogs(logEntry);
                    break;
                case "load":
                    // Ignore
                    break;
                case "loadend":
                    var currentStatus = (this.status >= 200 && this.status < 400)
                        ? 'COMPLETED'
                        : 'FAILED';
                    logEntry = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime.toLocaleTimeString(), currentStatus, evt.loaded, this.status, this.statusText, DateTime.now().toLocaleTimeString());
                    addNetworkRequestLogToPendingRequestLogs(logEntry);
                    break;
                case "progress":
                    // Do not log progress events as for long requests this may generate large logs that are not massively useful
                    // text = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime.toLocaleTimeString(), 'IN PROGRESS', evt.loaded, 'NA', 'NA', 'NA');
                    break;
                case "error":
                    logEntry = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime.toLocaleTimeString(), 'ERROR', evt.loaded, this.status, this.statusText, DateTime.now().toLocaleTimeString());
                    addNetworkRequestLogToPendingRequestLogs(logEntry);
                    break;
                case "abort":
                    logEntry = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime.toLocaleTimeString(), 'ABORTED', evt.loaded, 0, 'ABORTED' , DateTime.now().toLocaleTimeString());
                    addNetworkRequestLogToPendingRequestLogs(logEntry);
                    break;
                case "timeout":
                    logEntry = generateNetworkMessageText(requestId, requestMethod, requestUrl, this.requestStartTime.toLocaleTimeString(), 'TIMEOUT', evt.loaded, 0, 'TIMEOUT', DateTime.now().toLocaleTimeString());
                    addNetworkRequestLogToPendingRequestLogs(logEntry);
                    break;
            }             
        }

        function addNetworkRequestLogToPendingRequestLogs(logEntry) {
            pendingNetworkRequests += logEntry + "\r\n";
        }

        function generateNetworkMessageText(requestId, requestMethod, requestUrl, requestStartTime, currentStatus, bytesTransferred, responseStatus, responseStatusText, requestEndTime) {
            return "[" + DateTime.now().toLocaleString() + "]: " + "Request Id: " + requestId + " | Method: " + requestMethod + " | URL: " + requestUrl + " | Current Status: " + currentStatus + " | Response Status: " + responseStatus + " " + responseStatusText + " | Start Time: " + requestStartTime + " | End Time: " + requestEndTime + " | Bytes Transferred: " + bytesTransferred;
        }

        function loadLogFileList() {
            // If it doesn't exist then we have no logs stored locally to either upload or delete sp we can just leave the logFileList as an empty array
            if (File.exists(Settings.application.fileSystem.systemFiles.localLogFileListFilePath) === false) {
                logger.info('Local Log File List does not exist. Is this the first run?');
                return $.Deferred().resolve();
            }

            // The log file list exists so attempt to read and parse the file so we have a list of stored log files in memory
            // so we know what to upload and/or delete
            var dfd = $.Deferred();
            File.read(Settings.application.fileSystem.systemFiles.localLogFileListFilePath).then(function(logFiles) {
                try {
                    if (logFiles != null && logFiles != "") {
                        var parsedLogFilesArray = JSON.parse(logFiles);
                        logFilesList = parsedLogFilesArray;
                        logger.info('Log File List loaded. Stored log files: ' + logFiles);
                        dfd.resolve();
                    } 
                } catch (err) {
                    logger.error('Error loading log file list. Error: ' + err.message);
                    return dfd.resolve();
                }
            });
            return dfd;
        }

        function writePendingLogsToDisk() {
            var dfd = $.Deferred();
            // Copy logs and empty pending log arrays
            var consoleLogsToWrite = pendingConsoleMessages;
            var networkLogsToWrite = pendingNetworkRequests;
            pendingConsoleMessages = "";
            pendingNetworkRequests = "";

            // Get the current log file, creating a new log file if required
            getCurrentLogFilesAndRollNewLogsIfRequired().then(function(logsToUse, logsRolled) {
                currentLogFiles = logsToUse;
                consoleLog = currentLogFiles[0];
                networkLog = currentLogFiles[1];
                // Store logsToUse and Write pending console logs to current console log entry
                File.append(consoleLog.path, consoleLogsToWrite).then(function() {
                    // Write pending network logs to current network log entry
                    return File.append(networkLog.path, networkLogsToWrite);
                })
                // If uploads are enabled then do the upload
                .then(function() {
                    // Only attempt the log upload if we have rolled new logs and uploading is enabled
                    var attemptLogUpload = logsRolled == true && self.uploadsEnabled === true;
                    // Skip uploading logs if self.uploadsEnabled is false
                    if (attemptLogUpload === false) {
                        return $.Deferred().resolve();
                    }
                    // Otherwise upload and delete all logs
                    var allLogsUploaded = $.Deferred().resolve();
                    // Get all logs to upload - any logs that are not the current logs
                    var logsToUpload = logFilesList.filter(function(log) { // Note: if array.filter is not supported on older players we should be able to use $(array).filter instead
                        return !currentLogFiles.some(function(currentLog) {
                            return log.path === currentLog.path;
                        });
                    });
                    // Iterate over each log, upload and delete
                    $.each(logsToUpload, function(index, logToUpload) {
                        // Upload and delete log
                        allLogsUploaded = allLogsUploaded.then(function() {
                            return uploadAndDeleteLogFile(logToUpload.path);
                        });
                    });
                    return allLogsUploaded;
                })
                // Check if we need to delete any old log files that have not been uploaded either because uploads are not enabled or perhaps due to network issues
                .then(function() {
                    return deleteOldLogFiles();
                })
                // Save to disk again in 30 secs and Resolve as logs have been updated on the disk
                .then(function() {
                    writeTimeout = setTimeout(writePendingLogsToDisk, (Settings.application.modules.logging.writeToDiskFrequencyInSeconds * 1000));
                    dfd.resolve();
                });
            })           

            return dfd;
        }

        function uploadAndDeleteLogFile(logFilePath) {
            var dfd = $.Deferred();
            // Read Log File
            File.read(logFilePath).done(function(logData) {
                // After reading the file data check that we have log data to upload
                var hasLogDataToUpload = (logData != null && logData != "");

                if (hasLogDataToUpload) {
                    // Get key name
                    var s3KeyName = getS3KeynameForLogFile(logFilePath);
                    // Do upload
                    LogUploadInterfaceHelper.uploadLogFile(s3KeyName, logData).then(function() {
                        // Delete file
                        deleteLogFile(logFilePath).then(function() {
                            dfd.resolve();
                        });
                    });
                }
                else {
                    logger.info('No log data to upload.');
                    // Delete file
                    deleteLogFile(logFilePath).then(function() {
                        dfd.resolve();
                    });
                }
            }).fail(function() {
                logger.error('Error reading log file. Removing it from log list');
                // Delete file
                deleteLogFile(logFilePath).then(function() {
                    dfd.resolve();
                });
            })

            return dfd;
        }

        function deleteOldLogFiles() {
            var dfd = $.Deferred();
            // Filter logs that are not the current log and are older than configured max age
            var logsToDelete = logFilesList.filter(function(log) { // Note: if array.filter is not supported on older players we should be able to use $(array).filter instead
                return !currentLogFiles.some(function(currentLog) { return log.path === currentLog.path; }) && new Date(log.date) < new Date().addMinutes(-Settings.application.modules.logging.maxAgeOfstoredLogFileInMinutes);
            });

            var allOldLogsDeleted = $.Deferred().resolve();
            // Iterate over each log and delete
            $.each(logsToDelete, function(index, logToDelete) {
                // Upload and delete log
                allOldLogsDeleted = allOldLogsDeleted.then(function() {
                    return deleteLogFile(logToDelete.path);
                });
            });
            // Once all logs have been deleted then resolve the main function deferred object
            allOldLogsDeleted.then(function() {
                dfd.resolve();
            });

            return dfd;

        }

        function deleteLogFile(logFilePath) {
            var dfd = $.Deferred();
            // Delete file
            File.delete(logFilePath).then(function() {
                // remove file from logs file list
                logFilesList = logFilesList.filter(function(logFile) {
                    return logFile.path !== logFilePath;
                });
                // Resave local file list
                File.write(Settings.application.fileSystem.systemFiles.localLogFileListFilePath, JSON.stringify(logFilesList)).then(function() {
                    dfd.resolve();
                });
            });
            return dfd;
        }

        function getS3KeynameForLogFile(logFilePath) {
            var logFileName = logFilePath.split('/').pop();
            var logTimeString = "";
            try {
                var logTime = logFileName.split('_')[2];
                var logDate = new Date(parseInt(logTime));
                logTimeString = "_" + logDate.toISOString();
            }
            catch (err) {}      
            var uploadFileName = playerSerialNumber + '/' + logFileName + logTimeString + '.txt'; // always a text file
            return uploadFileName
        }

        function getCurrentLogFilesAndRollNewLogsIfRequired() {
            var dfd = $.Deferred();
            var currentDate = new Date();
            // Create new log files if we do not have a current log (app just started) or the log file is due to be rolled or was forcibly rolled (uploadLogs() called)
            if (currentLogFiles == null || rollLogsOnNextWrite === true || currentLogFiles[0].date <= currentDate.addMinutes(-Settings.application.modules.logging.logRollingfrequencyInMinutes)) {
                // Create the new log object
                var newLogs = [
                    {
                        date: currentDate,
                        path: Settings.application.fileSystem.systemFiles.consoleLogFileBasePath + "_" + currentDate.getTime(),
                    },
                    {
                        date: currentDate,
                        path: Settings.application.fileSystem.systemFiles.networkLogFileBasePath + "_" + currentDate.getTime(),
                    }
                ];
                // Add to logFileList and save
                logFilesList = logFilesList.concat(newLogs);
                File.write(Settings.application.fileSystem.systemFiles.localLogFileListFilePath, JSON.stringify(logFilesList)).then(function() {
                    // If logs were rolled due to forcing an upload then set this back to false
                    if (rollLogsOnNextWrite === true) {
                        rollLogsOnNextWrite = false;
                    }
                    // Resolve
                    dfd.resolve(newLogs, true);
                });
                return dfd;
            }
            else {
                 // Nothing to do - just continuing writing to the current log files
                return dfd.resolve(currentLogFiles, false);
            }
        }

        // NOTE: Doesn't work on LG so ignore for now?
        if(Settings.application.modules.logging && Settings.application.modules.logging.requestUploadUrlEndpoint) {
            // IIFE (Immediately Invoked Function Expression) functions that intercept console messages and network requests
            // ----------------------------------------------------

            // Override global console function to intercept console messages
            var originalConsole = window.console;
            var newConsole = {};
            ['log', 'debug', 'info', 'warn', 'error'].forEach(function(method) {
                newConsole[method] = function(text) {
                    originalConsole[method].apply(originalConsole, [text]);
                    queueConsoleMessagesForLogging(method.toUpperCase(), text);
                };
            });
            window.console = newConsole;
            
            // // Override XHR.open to Intercept and capture network requests
            (function() {
                var origOpen = XMLHttpRequest.prototype.open;
                var requestId = 0;
                XMLHttpRequest.prototype.open = function() {
                    requestId++;
                    this.arguments = arguments;
                    this.requestId = requestId;
                    this.requestStartTime = new Date();
                    this.addEventListener('loadstart', queueNetworkRequestsForLogging);
                    this.addEventListener('load', queueNetworkRequestsForLogging);
                    this.addEventListener('loadend', queueNetworkRequestsForLogging);
                    // this.addEventListener('progress', queueNetworkRequestsForLogging); Ignore Progress updates
                    this.addEventListener('error', queueNetworkRequestsForLogging);
                    this.addEventListener('abort', queueNetworkRequestsForLogging);
                    this.addEventListener('timeout', queueNetworkRequestsForLogging);
                    origOpen.apply(this, arguments);
                };
            })();

            // Start process to save logs to disk
            (function() {
                // Clear current interval is one is already set
                if (writeTimeout != null) {
                    clearTimeout(writeTimeout);
                    writeTimeout = null;
                }

                // Configure new write Interval
                writeTimeout = setTimeout(writePendingLogsToDisk, (Settings.application.modules.logging.writeToDiskFrequencyInSeconds * 1000));
            })();
        }

    }

    return new LogUploadInterface();
});