define(["properties/App.Settings", "properties/App.Properties", "system/io/File", "system/io/Directory", "models/ConfigurationModel", "models/ObjectProcessingQueueModel", "models/ConnectivityIntervalSchemeModel", "models/MediaTypeConstants", "system/DateTime", "models/SystemReportModel", "controllers/UserMessageController", "models/RssChannelCache", "models/RssChannelModel", "controllers/RealTimeEventsController", "models/DeviceInfoModel","core/SLAPIMethodsInterface","models/API","system/network/NetworkInterface", "core/DSSAPIInterface", "controllers/ZeroTouchConfigurationController", "system/power/PowerInterface", "models/RebootReasons"], function(Settings, Properties, File, Directory, ConfigurationModel, ObjectProcessingQueueModel, ConnectivityIntervalSchemeModel, MediaTypeConstants, DateTime, SystemReportModel, UserMessageController, RssChannelCache, RssChannel, RealTimeEventsController, DeviceInfoModel, APIMethodsInterface, API, NetworkInterface, DSSAPIInterface, ZeroTouchConfigurationController, PowerInterface, RebootReasons) {
    /*
    	SLAPIInterface - Provides an interface to communicate with the SL API
    	------------------------------------------------------------------------------
    	This module has been updated so that the authentication is now async rather than synchronous
     */
    function SLAPIInterface(apiEndpoint, apiCreds) {
        var self = this;

        // call back methods for registration and configuration
        var _registrationCallback;
        var _configurationCallback;
        var _alreadyRegisteredCallback;
        var _configurationNoChangeCallback;

        // default connection settings
        var _supportCheckFailureRetryIntervalMS = 10000;
        var _supportCheckUnsupportedRetryIntervalMS = 60000;
        var _registrationFailureRetryIntervalMS = 5000;
        var _authenticationFailureRetryIntervalMS = 5000;

        var _defaultConfigurationCheckIntervalMS = 60000;
        // 60 seconds
        var _defaultHealthCheckIntervalSecondMS = 60000;
        var _lastHealthCheckIntervalMS = _defaultHealthCheckIntervalSecondMS;

        // default to one hour
        var _defaultSystemReportIntervalMS = 3600000;

        var _registrationTimerId = 0;
        var _authenticationTimerId = 0;
        var _configurationTimerId = 0;
        var _healthCheckTimerId = 0;
        var _systemReportTimerId = 0;
        var _supportCheckTimerId = 0;

        self.lastConnectionTime = null;

        var _lastConfigurationModifiedDate = new Date(1970, 0, 1);
        var _lastConfigurationResponse;

        // don't permit configuration checks when content is being downloaded
        var _contentDownloadInProgress = false;

        var _externalAssets = [];

        var _APIMethodsInterface = new APIMethodsInterface();
  
        //
        // checkDeviceSupport()
        //
        function checkDeviceSupport() {
            logger.info('Checking API for device support');

            UserMessageController.postMessage('Confirming device support');

            var devInfo = new DeviceInfoModel();

            devInfo.generate().done(function()
            {
                var jqXHR = executeAPICall('device', {
                    type: 'POST',
                    contentType: 'application/json',
                    data: devInfo,
                    beforeSend: function(xhr, settings) {
                        xhr.setRequestHeader('X-SIGNAGELIVE-APP-ID', Settings.application.api.appId);
                        xhr.setRequestHeader('X-SIGNAGELIVE-CLIENT-ID', Settings.application.api.clientId);
                    }
                });

                if (jqXHR != null) {
                    jqXHR.done(function(data, textStatus, jqXHR) {
                        self.lastConnectionTime = new Date();
                        switch(data.result)
                        {
                            case 1:
                                //Supported
                                register();
                                break;
                            case 2:
                                UserMessageController.postMessage(data.message);
                                _supportCheckTimerId = setTimeout(checkDeviceSupport, _supportCheckUnsupportedRetryIntervalMS);
                                break;
                            case 3:
                                UserMessageController.postMessage("A client update is required.");
                                _supportCheckTimerId = setTimeout(checkDeviceSupport, _supportCheckUnsupportedRetryIntervalMS);
                                break;
                            case 0:
                                UserMessageController.postMessage('This device is not supported.\nManufacturer: ' + devInfo.manufacturer + '&nbsp;&nbsp;&nbsp;Model: ' + devInfo.model);
                                _supportCheckTimerId = setTimeout(checkDeviceSupport, _supportCheckUnsupportedRetryIntervalMS);
                            default:
                                break;
                        }
    
                    }).fail(function(jqXHR, textStatus, errorThrown) {
                        logger.error('Error connecting with Signagelive. Retrying...');
                    //_supportCheckUnsupportedRetryIntervalMS
                        _APIMethodsInterface.deviceSupportCheckAPICallFailed();
                        // call the support method again
                        _supportCheckTimerId = setTimeout(checkDeviceSupport, _supportCheckFailureRetryIntervalMS);
                    });
                } else {
                    logger.error('Error connecting with Signagelive. Retrying...');
                    
                    _APIMethodsInterface.deviceSupportCheckAPICallFailed();

                    // call the support method again
                    _supportCheckTimerId = setTimeout(checkDeviceSupport, _supportCheckFailureRetryIntervalMS);
                }

            });

        }

        //
        // register()
        //
        function register() {
            logger.info('Register with Signagelive API ' + Settings.application.api.clientId);

            UserMessageController.postMessage('Registering with Signagelive');

            var jqXHR = executeAPICall('register', {
                type: 'POST'
            });

            if (jqXHR != null) {
                jqXHR.done(function(data, textStatus, jqXHR) {
                    self.lastConnectionTime = new Date();
                    // set the client token
                    Settings.application.api.clientKey = data.client_key;

                    // execute the callback
                    if (_registrationCallback != undefined & _registrationCallback != null)
                        _registrationCallback(data.client_key);

                    // start the main configuration process
                    configuration();
                    postSystemReport();

                }).fail(function(jqXHR, textStatus, errorThrown) {

                    // conflict - this will indicate that the mac address is already registered
                    if (jqXHR.status == 409) {
                        _alreadyRegisteredCallback();
                    } else{
                        _APIMethodsInterface.registerAPICallFailed();
                    }

                    // log the error
                    logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);

                    // call the registration method again
                    _registrationTimerId = setTimeout(register, _registrationFailureRetryIntervalMS);
                });
            } else {

                // Log error
                logger.error('Error registing with Signagelive. Retrying...');

                _APIMethodsInterface.registerAPICallFailed();

                // call the registration method again
                _registrationTimerId = setTimeout(register, _registrationFailureRetryIntervalMS);
            }
        }

        //
        // authenticate()
        //
        function authenticate() {

            return _APIMethodsInterface.authenticate(API.PLAYER_API);
        }

        //
        // configuration()
        //
        function configuration() {
            if (_contentDownloadInProgress) {
                // reset the configuration timer for the default period before trying again
                _configurationTimerId = setTimeout(configuration, _defaultConfigurationCheckIntervalMS);
                return;
            }

            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                // Need to authenticate asynchronously
                authenticate().done(function(authenticated) {

                    // Client is not authenticated try again soon
                    if (!authenticated) {
                        setConfigurationCheckTimer();
                        return;
                    } else {
                        // Client is now authenticated - check for config update
                        checkForConfigurationUpdate();
                    }
                });
            } else {
                // Client is already authenticated - check for config update
                checkForConfigurationUpdate();
            }
        }

        //
        // checkForConfigurationUpdate()
        //
        function checkForConfigurationUpdate() {

            try {
                logger.info('Checking for a configuration update');

                // check to see if there is an update?
                var jqXHR = _APIMethodsInterface.executeAPICall(API.PLAYER_API,'configuration', {
                    type: 'HEAD',
		            contentType: 'application/json'
                });

                jqXHR.done(function(data, textStatus, jqXHR) {
                    self.lastConnectionTime = new Date();
                    // check the last modified date to see if there is an update
                    var lastModifiedDate = new Date(jqXHR.getResponseHeader('Last-Modified'));

                    if (lastModifiedDate.getTime() != _lastConfigurationModifiedDate.getTime()) {
                        var getConfigurationPromise = getConfiguration();

                        logger.debug("Configuration changed");

                        getConfigurationPromise.done(function() {
                            // remember the date of the last configuration response.
                            _lastConfigurationModifiedDate = lastModifiedDate;
                        }).fail(function() {

                        }).always(function() {
                            // reset the configuration timer
                            setConfigurationCheckTimer();
                        });
                    } else {
                        logger.debug("Configuration unchanged");
                        // No Change
                        if (_configurationNoChangeCallback != null || _configurationNoChangeCallback != undefined)
                            _configurationNoChangeCallback();

                        if (Settings.application.features.canDownloadFiles) {
                            // sync the external media assets
                            synchronizeExternalMediaAssets().always(function() {
                                // reset the configuration timer
                                setConfigurationCheckTimer();
                            });
                        } else {
                            // reset the configuration timer
                            setConfigurationCheckTimer();
                        }
                    }

                }).fail(function(jqXHR, textStatus, errorThrown) {

                    // reset auth token if it is not valid
                    if (jqXHR.status == 401) {
                        Settings.application.api.authTokenExpiry = null;

                        // try the configuration again in 30 seconds
                        _configurationTimerId = setTimeout(configuration, 30000);
                    } else {
                        // reset the configuration timer
                        setConfigurationCheckTimer();
                    }
                });
            } catch (e) {
                logger.error('Exception trying to perform content check.');

                // reset the configuration timer
                setConfigurationCheckTimer();
            }
        }

        //
        // synchronizeExternalMediaAssets()
        //
        function synchronizeExternalMediaAssets() {
            logger.debug('Sync External Media Assets');
            var extenalSyncDeferred = $.Deferred();

            // sync the externally available assets
            // set-up the asset download queue (for this playlist)
            var assetQueue = new ObjectProcessingQueueModel();
            $(_externalAssets).each(function(index, mediaAsset) {
                assetQueue.push(mediaAsset);
            });

            // download the queue
            downloadMediaAssetQueue(assetQueue, false).done(function() {
                extenalSyncDeferred.resolve();
            }).fail(function() {
                extenalSyncDeferred.reject();
            });

            return extenalSyncDeferred;
        }

        //
        // getConfiguration()
        //
        function getConfiguration() {

            var dfd = $.Deferred();

            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                authenticate().done(function(authenticated) {

                    // Client is not authenticated try again soon
                    if (!authenticated) {
                        dfd.reject();
                        return;
                    } else {
                        // Authenticated get new config
                        getUpdatedConfiguration()
                            .done(function() {
                                dfd.resolve();
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    }
                });
            } else {
                // Authenticated - get new config
                getUpdatedConfiguration()
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }

            return dfd;
        }

        //
        // getUpdatedConfiguration()
        //
        function getUpdatedConfiguration() {

            // Deffered object to use
            var getConfigurationDeferred = $.Deferred();

            UserMessageController.postMessage('Getting player configuration');

            var getConfigurationjqXHR = executeAPICall('configuration');

            var configurationJsonData = '';

            getConfigurationjqXHR.done(function(data, textStatus, jqXHR) {
                // remember the configuration response - we will need this once the configuration response has been
                // successfully processed
                //logger.debug("DOWNLOADED CONFIG: " + JSON.stringify(data));

                _lastConfigurationResponse = new ConfigurationModel(data);

                // remember the json response so that it can be saved to disk when required.
                configurationJsonData = data;

                // check the zero touch configuration against the player details and amend as appropriate
                ZeroTouchConfigurationController.configure(configurationJsonData).then(function (rebootRequired) {
                    // need to apply the new configuration without a reboot
                    if(configurationJsonData.settings && configurationJsonData.settings.usePMModeForScreenManagement && Settings.application.webOSVersion >= 4.1) {
                        PowerInterface.getPMMode().then(function(mode) {
                            if(mode !== "screenOffAlways") {
                                PowerInterface.setPMMode().then(function() {
                                    logger.info("PM Mode set to screenOffAlways");
                                });
                            }
                        });
                    };

                    // has the health check interval been updated?
                    if (_lastConfigurationResponse.connectivity.healthCheckInterval != _lastHealthCheckIntervalMS) {
                        logger.debug('Setting health check interval to ' + _lastConfigurationResponse.connectivity.healthCheckInterval);
                        // the health check interval has been updated - remember the new interval and reset the process
                        _lastHealthCheckIntervalMS = _lastConfigurationResponse.connectivity.healthCheckInterval;
                        clearTimeout(_healthCheckTimerId);

                        healthCheck();
                    }

                    // Notify Real Time Event Controller of any changes
                    RealTimeEventsController.updateConfiguration(_lastConfigurationResponse);

                    // process any content updates.
                    downloadContent(_lastConfigurationResponse.content)
                        .then(function () {
                            logger.info('All content downloaded.');
                            // If we have a config that includes data sync services config then start syncing
                            if (_lastConfigurationResponse && _lastConfigurationResponse.dataSyncServices && _lastConfigurationResponse.content.integrations && _lastConfigurationResponse.content.integrations.length > 0) {
                                logger.info('Restarting data sync services sync...');
                                return DSSAPIInterface.synchroniseIntegrationsForMediaAssets(_lastConfigurationResponse.dataSyncServices, _lastConfigurationResponse.content.integrations, _lastConfigurationResponse.content.mediaAssets);
                            } else {
                                logger.info('No last config or data sync services config object in configuration. Skipping DSS synchronisation...');
                                return $.Deferred().resolve();
                            }
                        })
                        .then(function () {
                            logger.info('Data Sync Services Synchronisation complete. Download Process Completed');
                            // execute the configuration complete callback
                            if (_configurationCallback != null || _configurationCallback != undefined)
                                _configurationCallback(_lastConfigurationResponse);
                            // save the configuration to disk
                            try {
                                logger.debug('Saving configuration to storage');
                                // save the file to storage
                                var configString = JSON.stringify(configurationJsonData);
                                File.write(Settings.application.fileSystem.systemFiles.lastConfigFilePath, configString)
                                    .done(function () {
                                        // get the assets that are available externally (e.g. RSS feeds) as SL will not know that
                                        // these have been updated.
                                        _externalAssets = $.grep(_lastConfigurationResponse.content.mediaAssets, function (mediaAsset, index) {
                                            return mediaAsset.isExternallyDelivered;
                                        });
                                        if (rebootRequired) {
                                            logger.debug("Performing instant reboot due to zero touch configuration changes.");
                                            PowerInterface.reboot(RebootReasons.ZERO_TOUCH);
                                        };
                                        getConfigurationDeferred.resolve();
                                    })
                                    .fail(function (e) {
                                        logger.error('Error whilst saving configuration file: ' + e.message);
                                        getConfigurationDeferred.reject();
                                    });
                            } catch (e) {
                                logger.error('Error whilst saving configuration file: ' + e.message);
                                getConfigurationDeferred.reject();
                            }
                        }).fail(function () {
                            logger.error('Download Process Failed');
                            getConfigurationDeferred.reject();
                        }).always(function () {
                            // always reset the download flag
                            _contentDownloadInProgress = false;
                        });
                    });
            }).fail(function (jqXHR, textStatus, errorThrown) {
                // log the error
                logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                getConfigurationDeferred.reject();
            });
            return getConfigurationDeferred.promise();
        }

        //
        // setConfigurationCheckTimer()
        //
        function setConfigurationCheckTimer() {
            // calculate when the next configuration check should occur based on settings in the configuration response.
            var interval = calculateNextConfigurationCheckInterval();

            // reset the configuration timer
            _configurationTimerId = setTimeout(configuration, interval);
        }

        //
        // calculateNextConfigurationCheckInterval()
        //
        function calculateNextConfigurationCheckInterval() {
            var interval = _defaultConfigurationCheckIntervalMS;

            if (_lastConfigurationResponse != null) {
                if (_lastConfigurationResponse.connectivity.configurationIntervalScheme == ConnectivityIntervalSchemeModel.REGULAR_INTERVAL) {
                    // set the configuration check interval to the one in the response
                    interval = _lastConfigurationResponse.connectivity.configurationCheckInterval;
                } else {
                    // calculate how long until the next time of day configuration check
                    var nextCheckInTime = DateTime.now().setTime(_lastConfigurationResponse.connectivity.configurationCheckTime);
                    if (nextCheckInTime < DateTime.now())
                        nextCheckInTime = nextCheckInTime.addDays(1);

                    interval = nextCheckInTime.subtract(DateTime.now());
                }
            }

            return interval;
        }

        //
        // calculateNextHealthCheckInterval()
        //
        function calculateNextHealthCheckInterval() {
            var interval = _defaultHealthCheckIntervalSecondMS;

            if (_lastConfigurationResponse != null) {
                // set the health check interval to the one in the response
                interval = _lastConfigurationResponse.connectivity.healthCheckInterval;
            }

            logger.debug('Health Check interval is : ' + interval);

            return interval;
        }

        //
        // downloadContent(content)
        //
        function downloadContent(content) {
            UserMessageController.postMessage('Downloading content');

            // prevent any content checks whilst content is being downloaded
            _contentDownloadInProgress = true;

            var updateSuccess = false;

            var dfd = $.Deferred();

            if (Settings.application.features.canDownloadFiles) {
                // get the layouts that require a download
                var layoutsToDownload = getLayoutsRequiringDownload(content);

                // add the layouts to a queue
                var layoutQueue = new ObjectProcessingQueueModel(layoutsToDownload);

                // process the layout download queue
                downloadLayoutQueue(layoutQueue).done(function(processedItems) {
                    logger.info('Layouts Downloaded');

                    updateSuccess = true;

                }).fail(function(processedItems, failedItems) {
                    logger.error('Layout Download Queue Failed');

                    updateSuccess = false;
                }).always(function() {
                    // after layouts have been processed process the default playlist
                    if (content.defaultPlaylist != null) {
                        logger.info('Downloading Default Playlist');

                        // add the playlist to a queue - so we can use promises
                        var playlistQueue = new ObjectProcessingQueueModel();
                        playlistQueue.items.push(content.defaultPlaylist);

                        // process the queue
                        downloadPlaylistQueue(playlistQueue).done(function(processedItems) {
                            logger.debug('Default Playlist Downloaded');

                            updateSuccess = updateSuccess & true;
                        }).fail(function(processedItems, failedItems) {
                            logger.error('Default Playlist Download Failed');

                            updateSuccess = false;
                        }).always(function() {
                            // return the overall success of the entire done load
                            $("#userMessageText").empty();
                            if (updateSuccess)
                                dfd.resolve();
                            else
                                dfd.reject();
                        });
                    } else {
                        dfd.resolve();
                    }
                });

            } else {
                // can't download - just fake it!
                setTimeout(function() {
                    dfd.resolve();
                }, 500);
            }

            return dfd.promise();
        }

        //
        // getLoyoutsRequiringDownlood(layouts)
        //
        function getLayoutsRequiringDownload(content) {
            // TODO - are there any layouts that require a download in the new configuration?
            // are there layouts where the content is not available locally...

            var layoutsRequiringDownload = [];

            $(content.schedules).each(function(i, schedule) {
                var addToSchedule = false;
                if (schedule.layout.isDownloaded == false) {
                    // find the media asset
                    var layouts = $.grep(layoutsRequiringDownload, function(layout, index) {
                        return layout.externalId == schedule.layout.externalId;
                    });

                    if (layouts.length == 0)
                        addToSchedule = true;
                } else {
                    // TODO - remove and enhance to check layout design, playlists etc
                    // find the media asset
                    var layouts = $.grep(layoutsRequiringDownload, function(layout, index) {
                        return layout.externalId == schedule.layout.externalId;
                    });
                    if (layouts.length == 0)
                        addToSchedule = true;
                }

                if (addToSchedule) {
                    layoutsRequiringDownload.push(schedule.layout);
                }
            });

            // Add interrupts
            $(content.interrupts).each(function(i, interrupt) {
                var addToSchedule = false;
                // find the media asset
                var layouts = $.grep(layoutsRequiringDownload, function(layout, index) {
                    return layout.externalId == interrupt.action.layout.externalId;
                });
                if (layouts.length == 0) {
                    layoutsRequiringDownload.push(interrupt.action.layout);
                }
            });

            return layoutsRequiringDownload;
        }

        //
        // downloadLayoutQueue(layoutQueue)
        //
        function downloadLayoutQueue(layoutQueue) {
            var queueDFD = $.Deferred();
            downloadNextLayoutInQueue(layoutQueue, queueDFD);
            return queueDFD.promise();
        }

        //
        // downloadNextLayoutInQueue(layoutQueue, queueDfd)
        //
        function downloadNextLayoutInQueue(layoutQueue, queueDfd) {
            // is the queue empty - return the success of the queue processing?
            if (layoutQueue.items.length == 0) {
                if (layoutQueue.failedItems.length == 0)
                    queueDfd.resolve(layoutQueue.processedItems);
                else
                    queueDfd.reject(layoutQueue.processedItems, layoutQueue.failedItems);
                return;
            }

            var layout = layoutQueue.currentItem();

            logger.info('Layout Download Start ' + layout.name);

            // notify the SL API the download is starting
            notifyDownloadStart(layout.externalId);

            // setup the playlist and asset download queues
            var playlistQueue = new ObjectProcessingQueueModel();

            // the assets that will be downloaded are misc objects
            // such as default zone assets, backgrounds etc
            var assetQueue = new ObjectProcessingQueueModel();

            if (layout.backgroundAsset != null) {
                assetQueue.push(layout.backgroundAsset);
            }

            // setup the queues
            $(layout.zones).each(function(index, zone) {
                if (zone.backgroundAsset != null)
                    assetQueue.push(zone.backgroundAsset);

                if (zone.seperatorAsset != null)
                    assetQueue.push(zone.seperatorAsset);

                if (zone.defaultMediaAsset != null)
                    assetQueue.push(zone.defaultMediaAsset);

                if (zone.defaultPlaylist != null)
                    playlistQueue.push(zone.defaultPlaylist);

                // add all of the scheduled playlists to the queue
                $(zone.scheduledPlaylists).each(function(index, scheduledPlaylist) {
                    // check the asset is not already in the queue
                    var matchingItemsInQueue = $.grep(playlistQueue.items, function(playlist, index) {
                        return playlist.externalId == scheduledPlaylist.playlist.externalId;
                    }).length;

                    if (matchingItemsInQueue == 0)
                        playlistQueue.push(scheduledPlaylist.playlist);
                });
            });

            // process the playlist queue
            downloadPlaylistQueue(playlistQueue).done(function(processedItems) {
                var layout = layoutQueue.shift();
                layoutQueue.processedItems.push(layout);

                logger.info('Layout Download Complete ' + layout.name);

                // notify the SL API the download is complete
                notifyDownloadComplete(layout.externalId);

            }).fail(function(processedItems, failedItems) {
                var layout = layoutQueue.shift();
                layoutQueue.failedItems.push(layout);

                logger.error('Layout Download Failed ' + layout.name);

                // notify the SL API the download has failed
                notifyDownloadFail(layout.externalId);

            }).always(function() {
                // TODO correctly handle success and failure

                downloadMediaAssetQueue(assetQueue, true).done(function() {

                }).fail(function() {

                }).always(function() {
                    // process the next item in the queue
                    setTimeout(function() {
                        downloadNextLayoutInQueue(layoutQueue, queueDfd);
                    }, 500);
                });
            });
        }

        //
        // downloadPlaylistQueue(playlistQueue)
        //
        function downloadPlaylistQueue(playlistQueue) {
            var queueDFD = $.Deferred();
            downloadNextPlaylistInQueue(playlistQueue, queueDFD);
            return queueDFD.promise();
        }

        //
        // downloadNextPlaylistInQueue(playlistQueue, queueDfd)
        //
        function downloadNextPlaylistInQueue(playlistQueue, queueDfd) {
            // is the queue empty - return the success of the queue processing
            if (playlistQueue.items.length == 0) {
                if (playlistQueue.failedItems.length == 0)
                    queueDfd.resolve(playlistQueue.processedItems);
                else
                    queueDfd.reject(playlistQueue.processedItems, playlistQueue.failedItems);
                return;
            }

            // process the next item
            var playlist = playlistQueue.currentItem();
            logger.info('Playlist Download Start ' + playlist.name);

            // notify the SL API the download is starting
            notifyDownloadStart(playlist.externalId);

            // set-up the asset download queue (for this playlist)
            var assetQueue = new ObjectProcessingQueueModel();
            $(playlist.playlistMediaAssets).each(function (index, playlistMediaAsset) {

                //If this is a feed, add the assets to the queue
                if (playlistMediaAsset.mediaAsset.contentType == MediaTypeConstants.FEED) {
                    $(playlistMediaAsset.feed.feedMediaAssets).each(function (index, feedMediaAsset) {
                        var matchingItemsInQueue = $.grep(assetQueue.items, function (mediaAsset, index) {
                            return mediaAsset.externalId == feedMediaAsset.mediaAssetId;
                        }).length;

                        if (matchingItemsInQueue == 0) {
                            // check the file is not already downloaded
                            assetQueue.push(feedMediaAsset.mediaAsset);
                        }
                    });
                } else {
                    // check the asset is not already in the queue
                    var matchingItemsInQueue = $.grep(assetQueue.items, function (mediaAsset, index) {
                        return mediaAsset.externalId == playlistMediaAsset.mediaAssetId;
                    }).length;

                    if (matchingItemsInQueue == 0) {
                        // check the file is not already downloaded
                        assetQueue.push(playlistMediaAsset.mediaAsset);
                    }
                }
            });
                

            logger.debug("Items in queue: " + assetQueue.items.length);

            // process the media asset queue
            downloadMediaAssetQueue(assetQueue, true).done(function(downloadedAssets) {
                // add the playlist to the successful items collection
                var playlist = playlistQueue.items.shift();
                playlistQueue.processedItems.push(playlist);

                logger.info('Playlist Download Complete ' + playlist.name);

                // notify the SL API the download is complete
                notifyDownloadComplete(playlist.externalId);

            }).fail(function(downloadedAssets, failedAssets) {
                // add the playlist to the successful items collection
                var playlist = playlistQueue.items.shift();
                playlistQueue.failedItems.push(playlist);

                logger.error('Playlist Download Fail ' + playlist.name + ' ' + failedAssets.length + ' assets have failed to download');

                // notify the SL API the download has failed
                notifyDownloadFail(playlist.externalId);

            }).always(function() {
                // process the next item in the queue
                setTimeout(function() {
                    downloadNextPlaylistInQueue(playlistQueue, queueDfd);
                }, 500);
            });
        }

        //
        // downloadMediaAssetQueue(assetQueue)
        //
        function downloadMediaAssetQueue(assetQueue, withDownloadNotifications) {
            var queueDFD = $.Deferred();
            downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
            return queueDFD.promise();
        }

        //
        // downloadNextMediaAssetInQueue(assetQueue, queueDFD)
        //
        function downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications) {
            // is the queue empty - return the success of the queue processing
            if (assetQueue.items.length == 0) {
                if (assetQueue.failedItems.length == 0)
                    queueDFD.resolve(assetQueue.processedItems);
                else
                    queueDFD.reject(assetQueue.processedItems, assetQueue.failedItems);
                return;
            }

            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                authenticate().done(function(authenticated) {
                    if (!authenticated) {
                        // Auth failed
                        queueDFD.reject(assetQueue.processedItems, assetQueue.failedItems);
                    } else {
                        // Authenticated now - do download
                        doDownloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                    }
                });
            } else {
                // Already authenticated - do download
                doDownloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
            }
        }

        /**
         * Attempts to download the next media asset in the download queue
         * @param  {[type]} assetQueue                [description]
         * @param  {[type]} queueDFD                  [description]
         * @param  {[type]} withDownloadNotifications [description]
         * @return {[type]}                           [description]
         */
        function doDownloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications) {
            // process the next item
            var asset = assetQueue.currentItem();

            logger.info('Checking if asset requires a download');

            // does the media asset require a download?
            var requiresDownload = doesAssetRequireDownload(asset);
            if (requiresDownload == false) {
                // remove this item from the queue
                // add the asset to the successful items collection
                var asset = assetQueue.items.shift();
                assetQueue.processedItems.push(asset);

                // process the next item in the queue and ignore this one (as it either is already downloaded or is not downloadable)
                setTimeout(function() {
                    downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                }, 500);
                return;
            }

            logger.info('Media Asset Download Start ' + asset.name);

            // notify the SL API the download has started
            if (withDownloadNotifications)
                notifyDownloadStart(asset.externalId);

            require(["system/plugins/DownloadPlugin"], function(DownloadPlugin) {
                // initialize the downloader
                var downloadPlugin = new DownloadPlugin();

                // SL Media Asset Downloads on the SSSP must append the authentication values to the url
                // as the FileDownloadPlugin does not support custom headers

                // Add Query String Params if the request is for content from the SL API
                var authString = '';
                if (asset.remoteUrl.lastIndexOf(Settings.application.api.url, 0) === 0)
                {
                    authString = "?sl_app_id=" + Settings.application.api.appId + "&sl_app_key=" + Settings.application.api.appKey + "&sl_client_id=" + Settings.application.api.clientId + "&sl_client_key=" + Settings.application.api.clientKey + "&sl_auth_token=" + encodeURIComponent(Settings.application.api.authToken);
                    if (Settings.application.api.brand !== 'Signagelive') {
                        authString += "&sl_install_id=" + Settings.application.api.brand;
                    }
                }
                var remoteUrl = asset.remoteUrl + authString;

                // Build Local File Url
                var localFile = "";
                if (Settings.application.fileSystem.root !== "") {
                    localFile = localFile + Settings.application.fileSystem.root + "/";
                }
                if (Settings.application.fileSystem.mediaDirectory !== "") {
                    localFile = localFile + Settings.application.fileSystem.mediaDirectory + '/';
                }
                localFile = localFile + asset.localFileName;

                logger.debug('Local file name: ' + localFile);

                // Start download
                downloadPlugin.downloadFile(remoteUrl, localFile, asset.contentType)
                    // On Download Success
                    .done(function(_remoteUrl, _localFile) {
                        // Handle RSS ContentType
                        if (asset.contentType == MediaTypeConstants.RSS) {
                            // Add to channel cache then add to process items and notify of success	
                            RssChannelCache.loadRssAssetIntoCache(asset);
                            handleAssetDownloadSuccess(assetQueue, withDownloadNotifications).then(function() {
                                downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                            });
                        }
                        // Handle Widget ContentType
                        else if (asset.contentType == MediaTypeConstants.WIDGET) {
                            // Need to extract the widget and handle failures appropriately
                            logger.info("Unzipping widget: " + asset.localFileName);

                            // Get widget folder
                            var widgetDestination = "";
                            if (Settings.application.fileSystem.root !== "") {
                                widgetDestination = widgetDestination + Settings.application.fileSystem.root + "/";
                            }
                            if (Settings.application.fileSystem.widgetDirectory !== "") {
                                widgetDestination = widgetDestination + Settings.application.fileSystem.widgetDirectory + '/';
                            }
                            widgetDestination = widgetDestination + asset.localFileName.replace('.wgt', '');

                            logger.debug("Creating new widget directory: " + widgetDestination);

                            // Create new directory - it will not exist at this point
                            Directory.createDirectory(widgetDestination)
                                // On directory creation success
                                .done(function() {
                                    // Extract Archive
                                    logger.debug("New Directory created. Unzipping " + _localFile + " to destination " + widgetDestination);

                                    File.unzip(_localFile, widgetDestination)
                                        // On unzip success
                                        .done(function() {
                                            logger.debug("Successfully extracted widget file: " + _localFile);
                                            handleAssetDownloadSuccess(assetQueue, withDownloadNotifications).then(function() {
                                                downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                                            });
                                        })
                                        // On unzip failure
                                        .fail(function() {
                                            logger.error("Failed to upzip widget pckage");
                                            handleAssetDownloadFailure(_localFile, assetQueue, withDownloadNotifications).then(function() {
                                                downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                                            });
                                        });
                                })
                                // On Directory creation failure
                                .fail(function() {
                                    handleAssetDownloadFailure(_localFile, assetQueue, withDownloadNotifications).then(function() {
                                        downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                                    });
                                });
                        }
                        // Handle All Other ContentTypes
                        else {
                            // Default case - nothing to do so just handle success
                            handleAssetDownloadSuccess(assetQueue, withDownloadNotifications).then(function() {
                                downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                            });
                        }
                    })
                    // On Download Failure
                    .fail(function(_remoteUrl, _localFile) {
                        handleAssetDownloadFailure(_localFile, assetQueue, withDownloadNotifications).then(function() {
                            downloadNextMediaAssetInQueue(assetQueue, queueDFD, withDownloadNotifications);
                        });
                    });
            });
        }

        /**
         * Updates queue and sends notification on download success
         * @param  {[type]} assetQueue                [description]
         * @param  {[type]} withDownloadNotifications [description]
         * @return {[type]}                           [description]
         */
        function handleAssetDownloadSuccess(assetQueue, withDownloadNotifications) {
            var dfd = $.Deferred();
            // add the asset to the successful items collection
            var asset = assetQueue.items.shift();
            assetQueue.processedItems.push(asset);

            // Log Successfull download
            logger.info('Media Asset Download Complete ' + asset.name);

            // notify the SL API the download is complete
            if (withDownloadNotifications) {

                RealTimeEventsController.sendMediaAssetDownloaded(asset);

                notifyDownloadComplete(asset.externalId).always(function() {
                    // This small wait seems important in order to ensure there are no IO errors when writing the next file
                    setTimeout(function() {
                        dfd.resolve();
                    }, 500);
                });
            } else {
                setTimeout(function() {
                    dfd.resolve();
                }, 500);
            }

            return dfd;
        }

        /**
         * Updates the queue, sends failure notification and attempts to delete partially downloaded file
         * @param  {[type]} localFile                 [description]
         * @param  {[type]} assetQueue                [description]
         * @param  {[type]} withDownloadNotifications [description]
         * @return {[type]}                           [description]
         */
        function handleAssetDownloadFailure(localFile, assetQueue, withDownloadNotifications) {
            var dfd = $.Deferred();
            // add the asset to the failed items collection
            var asset = assetQueue.items.shift();
            assetQueue.failedItems.push(asset);

            logger.error('Media Asset Download Fail ' + asset.name);

            // notify the SL API the download has failed
            if (withDownloadNotifications) {
                notifyDownloadFail(asset.externalId).always(function() {
                    try {
                        File.delete(localFile).always(function() {
                            setTimeout(function() {
                                dfd.resolve();
                            }, 500);
                        });
                    } catch (e) {
                        logger.error('Error whilst removing a failed download file ' + e.message);
                        setTimeout(function() {
                            dfd.resolve();
                        }, 500);
                    }
                });
            } else {
                setTimeout(function() {
                    dfd.resolve();
                }, 500);
            }

            // remove the partially downloaded file - this does not need to be done prior to resolving method


            return dfd;
        }

        function notifySystemStart() {
            var dfd = $.Deferred();

            notify({
                type: "SystemStart"
            }).done(function() {
                dfd.resolve();
            }).fail(function() {
                dfd.reject();
            });

            return dfd;
        }

        self.healthCheck = function() {
            return healthCheck();
        };

        function healthCheck() {
            logger.info('HEALTH CHECK');
            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                authenticate().done(function(authenticated) {

                    if (!authenticated) {
                        // reset the configuration timer
                        _healthCheckTimerId = setTimeout(healthCheck, calculateNextHealthCheckInterval());
                        return;
                    } else {
                        // Authenticated now - do health check
                        doHealthCheck();
                    }
                });
            } else {
                // Authenticated already - do health check
                doHealthCheck();
            }


        }

        function doHealthCheck() {
            var jqXHR = executeAPICall('notifications', {
                type: 'POST',
                data: {
                    type: 'HealthCheck'
                }
            });

            jqXHR.done(function(data, textStatus, jqXHR) {
                self.lastConnectionTime = new Date();
                logger.info('Health Check Code ' + jqXHR.status);

            }).fail(function(jqXHR, textStatus, errorThrown) {
                // log the error
                logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);

                // reset auth token if it is not valid
                if (jqXHR.status == 401) {
                    Settings.application.api.authTokenExpiry = null;
                }
            }).always(function() {
                // reset the configuration timer
                _healthCheckTimerId = setTimeout(healthCheck, calculateNextHealthCheckInterval());
            });
        }

        function notify(data) {
            var dfd = $.Deferred();

            logger.info('Notify ' + data.type + ', Object Id: ' + data.object_id);

            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                authenticate().done(function(authenticated) {

                    if (!authenticated) {
                        // Not authed - cant send notification
                        dfd.reject();
                    } else {
                        // Now authenticated do send notification
                        sendNotification(data).done(function() {
                            dfd.resolve();
                        }).fail(function() {
                            dfd.reject();
                        });
                    }
                });
            } else {
                // Already authenticated - send the notification request
                sendNotification(data).done(function() {
                    dfd.resolve();
                }).fail(function() {
                    dfd.reject();
                });
            }

            return dfd;
        }

        function sendNotification(data) {
            var dfd = $.Deferred();

            var jqXHR = executeAPICall('notifications', {
                type: 'POST',
                data: data
            });

            jqXHR.done(function(data, textStatus, jqXHR) {
                self.lastConnectionTime = new Date();
                logger.info('Notify Response Code ' + jqXHR.status);
                dfd.resolve();
            }).fail(function(jqXHR, textStatus, errorThrown) {
                // reset auth token if it is not valid
                if (jqXHR.status == 401) {
                    Settings.application.api.authTokenExpiry = null;
                }

                // log the error
                logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                dfd.reject();
            });

            return dfd;
        }

        function executeAPICall(operationName, settings) {

            return _APIMethodsInterface.executeAPICall(API.PLAYER_API,operationName,settings);
          
        }

        function isRegistrationRequired() {
            return (Settings.application.api.clientKey === undefined || Settings.application.api.clientKey === null || Settings.application.api.clientKey === "");
        }

        function isAuthenticationRequired() {
            logger.debug('isAuthenticationRequired()');

            if (Settings.application.api.authTokenExpiry == null)
                return true;
            if (Settings.application.api.authTokenExpiry < DateTime.now())
                return true;
            return false;
        }

        function notifyDownloadStart(objectId) {
            return notify({
                type: "DownloadStart",
                object_id: objectId
            });
        }

        function notifyDownloadComplete(objectId) {
            return notify({
                type: "DownloadComplete",
                object_id: objectId
            });
        }

        function notifyDownloadFail(objectId) {
            return notify({
                type: "DownloadFail",
                object_id: objectId
            });
        }

        function doesAssetRequireDownload(mediaAsset) {
            var requiresDownload = false;
            try {

                // is the asset downloadable
                var downloadable = mediaAsset.isDownloadable;

                // is it already downloaded?
                if (downloadable) {
                    if (mediaAsset.isExternallyDelivered)
                        requiresDownload = true;
                    else {
                        // Build Local File Url
                        var localFile = "";
                        if (Settings.application.fileSystem.root !== "") {
                            localFile = localFile + Settings.application.fileSystem.root + "/";
                        }
                        if (Settings.application.fileSystem.mediaDirectory !== "") {
                            localFile = localFile + Settings.application.fileSystem.mediaDirectory + '/';
                        }
                        localFile = localFile + mediaAsset.localFileName;
                        if (File.exists(localFile) == false) {
                            requiresDownload = true;
                        }
                    }
                }
            } catch (e) {
                logger.error('Error while checking if a download is required ' + e.message);
            }

            return requiresDownload;
        }

        function postSystemReport() {
            try {

                logger.info('Sending system report');

                // does the client need to authenticate again?
                if (isAuthenticationRequired()) {
                    if (authenticate() == false) {
                        // reset the system report timer
                        setSystemReportTimer(true);
                        return;
                    }
                }
                var sysReport = new SystemReportModel();
                sysReport.generate().done(function() {
                    var s = JSON.stringify(sysReport);
                    logger.debug(s);

                    var jqXHR = executeAPICall('systemreport', {
                        type: 'POST',
                        data: sysReport
                    });

                    jqXHR.done(function(data, textStatus, jqXHR) {
                        self.lastConnectionTime = new Date();
                        logger.info('System Report Done ' + jqXHR.status);
                        setSystemReportTimer(false);
                    }).fail(function(jqXHR, textStatus, errorThrown) {
                        // log the error
                        logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                        setSystemReportTimer(true);
                    })
                }).fail(function() {
                    logger.debug('System Report Generation Failed');
                    setSystemReportTimer(true);
                });
            } catch (e) {
                logger.debug('System Report Generation Failed ' + e.message);
                setSystemReportTimer(true);
            }
        }

        function setSystemReportTimer(retry) {
            // calculate when the next system report should occur based on settings in the configuration response.
            var interval = calculateNextSystemReportInterval(retry);

            // reset the configuration timer
            logger.info('Next System Report to be sent in ' + (interval / 1000) + ' seconds');
            _systemReportTimerId = setTimeout(postSystemReport, interval);
        }

        function calculateNextSystemReportInterval(retry) {
            var interval = _defaultSystemReportIntervalMS;

            // If we need to retry because auth failed for example then retry in 60 seconds
            if (retry === true) {
                return 60000;
            }

            if (_lastConfigurationResponse != null) {
                if (_lastConfigurationResponse.connectivity.diagnosticsUpdateIntervalScheme == ConnectivityIntervalSchemeModel.REGULAR_INTERVAL) {
                    // set the interval to the one in the response
                    interval = _lastConfigurationResponse.connectivity.diagnosticsUpdateInterval;
                } else {
                    // calculate how long until the next time of day configuration check
                    var nextCheckInTime = DateTime.now().setTime(_lastConfigurationResponse.connectivity.diagnosticsUpdateTime);
                    if (nextCheckInTime < DateTime.now())
                        nextCheckInTime = nextCheckInTime.addDays(1);

                    interval = nextCheckInTime.subtract(DateTime.now());
                }
            }

            return interval;
        }

        function logAPIError(requestUrl, requestData, responseStatus, responseMessage, responseText) {
            var logData = {
                url: requestUrl,
                requestData: requestData,
                responseStatus: responseStatus,
                responseMessage: responseMessage,
                responseBody: responseText
            };

            logger.error("API Error " + JSON.stringify(logData));
        }


        self.initialize = function(lastConfiguration, callbacks) {
            logger.info('Initializing SLAPIInterface');

            // register callbacks
            _registrationCallback = callbacks.deviceRegistered;
            _alreadyRegisteredCallback = callbacks.deviceAlreadyRegistered;
            _configurationCallback = callbacks.configuration;
            _configurationNoChangeCallback = callbacks.configurationNoChange;
            // Register self with Network Interface so that it can make health checks to check
            // internet connection
            NetworkInterface.registerAPIInterface(self);

            if (isRegistrationRequired()) {
                checkDeviceSupport();
            } else {

                // load the configuration - if there is one
                if (lastConfiguration != null) {
                    _lastConfigurationResponse = lastConfiguration;
                    _lastConfigurationModifiedDate = lastConfiguration.lastModified;

                    // get the assets that are available externally (e.g. RSS feeds) as SL will not know that
                    // these have been updated.
                    _externalAssets = $.grep(_lastConfigurationResponse.content.mediaAssets, function(mediaAsset, index) {
                        return mediaAsset.isExternallyDelivered;
                    });
                }

                // notify that the client has started
                notifySystemStart().always(function () {

                    if (_lastConfigurationResponse && _lastConfigurationResponse.dataSyncServices && _lastConfigurationResponse.content.integrations && _lastConfigurationResponse.content.integrations.length > 0) {
                        // Start Syncing Data Sync Services
                        DSSAPIInterface.synchroniseIntegrationsForMediaAssets(_lastConfigurationResponse.dataSyncServices, _lastConfigurationResponse.content.integrations, _lastConfigurationResponse.content.mediaAssets)
                            .always(function () {
                                // start the configuration process
                                configuration();
                                // start the health check process
                                _healthCheckTimerId = setTimeout(healthCheck, calculateNextHealthCheckInterval());
                                // start the system report process
                                postSystemReport();
                            });
                    }
                    // else just start the configuration process
                    else {
                        // start the configuration process
                        configuration();
                        // start the health check process
                        _healthCheckTimerId = setTimeout(healthCheck, calculateNextHealthCheckInterval());
                        // start the system report process
                        postSystemReport();
                    }
                });

            }
        };

        self.notifyFirmwareUpdateComplete = function() {
            notify({
                type: "FirmwareUpdateComplete"
            });
        };
        self.notifyFirmwareUpdateFail = function() {
            notify({
                type: "FirmwareUpdateFailed"
            });
        };
    }

    return SLAPIInterface;
});