define(["properties/App.Settings", "properties/App.Properties", "system/io/File", "system/io/Directory", "system/io/DirectoryInfo", "system/network/NetworkInterface", "system/power/PowerInterface", "system/DateTime", "models/ConfigurationModel", "core/SLAPIInterface", "controllers/UserMessageController", "gui/views/MediaPlayerView", "controllers/ApplicationUpdateController", "controllers/FirmwareUpdateController", "controllers/UserConfigurationController", "system/remotecontrol/RemoteControlInterface", "models/KeyboardShortcutsModel", "controllers/InterruptController", "models/PlaybackStartEndReasons", "controllers/ScreenCaptureController", "controllers/WebBasedInterruptsController", "controllers/TouchScreenController", "system/clock/ClockInterface", "system/sound/SoundInterface", "controllers/RealTimeEventsController", "controllers/LocalAPIController", "system/graphics/GraphicsInterface", "models/RebootReasons", "core/LogUploadInterface"], function(Settings, Properties, File, Directory, DirectoryInfo, NetworkInterface, PowerInterface, DateTime, ConfigurationModel, SLAPIInterface, UserMessageController, MediaPlayerView, ApplicationUpdateController, FirmwareUpdateController, UserConfigurationController, RemoteControlInterface, KeyboardShortcutsModel, InterruptController, PlaybackStartEndReasons, ScreenCaptureController, WebBasedInterruptsController, TouchScreenController, ClockInterface, SoundInterface, RealTimeEventsController, LocalAPIController, GraphicsInterface, RebootReasons, LogUploadInterface) {
    function SLClientEngine(callbacks) {
        var self = this;
        var _callbacks = callbacks;
        var _apiInterface = null;
        var _currentConfiguration = null;
        var _previousConfiguration = null;
        var _mediaPlayerView = new MediaPlayerView({ showSetup: showSetupView });
        var _appUpdateController = null;
        var _firmwareUpdateController = null;
        var _screenCapController = null;
        var _isFirstBoot = false;
        var _scheduledRebootTimer = null;
        var _touchScreenController = null;
        var _lastTouchReceived = null;
        var _localAPIServer = null;
        var _persistScreen = null;
        var _mediaPlayerStarted = false;
        var displayDetectioninProgress = false;

        // Used to manage Views - Initially both are not visible
        self.propertiesViewVisible = false;
        self.setupViewVisible = false;

        /**
         * Client initialization
         */
        self.initialize = function() {
            logger.info('Initializing SLClientEngine');

            _touchScreenController = new TouchScreenController();
            _touchScreenController.initialize({ showDeviceProperties: showDeviceInformationView, showSetupScreen: showSetupView });

            // Setup listening for remote
            if (Settings.application.features.canListenForRemoteKeyPresses)
                setupRemoteControlListener();

            UserConfigurationController.initialize().done(function() {
                // is this the first run
                if (File.exists(Settings.application.fileSystem.systemFiles.settingsFilePath) == false) {
                    logger.info('Settings file: ' + Settings.application.fileSystem.systemFiles.settingsFilePath + ' does not exist!');
                    firstBootHandler();
                } else {
                    // check the date time is correct - for the specified time-zone?
                    if (Settings.user.settingsLoaded) {

                        // Load module data from local storage
                        $.extend(true, Settings.application.modules, UserConfigurationController.modules());

                        // Initialize User Configuration Controller
                        if (Settings.application.features.canSyncDateTime) {

                            var userSettings = Settings.user.mySettings;
                            checkDateTimeIsCorrectFirstRun(userSettings);
                        } else {
                            initializeCoreServices();
                        }
                    } else {
                        // TODO critical error as there is no valid settings file
                        logger.error("Settings have not loaded");
                    }
                }

                // Setup Keyboard Shortcuts
                new KeyboardShortcutsModel().setup({
                    showSetup: showSetupView,
                    showInfo: showDeviceInformationView,
                    triggerKeyboardInterrupt: triggerKeyboardInterrupt
                });
            }).fail(function() {
                logger.error("SLClientEngine: Failed to initialize User Configuration Controller.");
            });
        };

        /**
         * Check date time on first run
         */
        function checkDateTimeIsCorrectFirstRun(userSettings) {
            try {

                if (Settings.application.features.canSyncDateTime) {
                    // the date time is incorrect - tell the user they need to connect the unit to the internet
                    UserMessageController.postMessage('Checking the date time is correct');
                    logger.info('Checking the date time is correct');

                    var jqXHR = $.ajax({
                        url: Settings.application.dateTimeServer + userSettings.timeZone.id,
                        contentType: 'application/json',
                        timeout: 10000,
                        cache: false
                    }).done(function(data, textStatus, jqXHR) {
                        logger.info('Local Time for ' + data.time_zone + ' is ' + data.local_time);

                        var isDateTimeCorrect = true;

                        var twoMinsForwards = DateTime.now().addMinutes(2);
                        var twoMinsBackwards = DateTime.now().addMinutes(-2);

                        var serverTime = DateTime.parse(data.local_time);

                        if (serverTime > twoMinsBackwards & serverTime < twoMinsForwards)
                            isDateTimeCorrect = true;
                        else
                            isDateTimeCorrect = false;

                        if (isDateTimeCorrect) {

                            // init time sync check
                            initializeTimeSync();

                            // init media player
                            initializeCoreServices();

                        } else {
                            logger.debug("Resyncing date time on the panel");
                            UserMessageController.postMessage('Resetting the date time, your panel may automatically restart in a moment');

                            DateTime.setDateTime(DateTime.parse(data.local_time))
                                .done(function() {
                                    logger.debug("Date time successfully updated.");

                                    // Everything worked so continue to load client

                                    // init time sync check
                                    initializeTimeSync();

                                    // init media player
                                    initializeCoreServices();
                                })
                                .fail(function() {
                                    logger.error("There was an error updating the date time on the panel. Retrying in 30 seconds...");

                                    // retry in 30 seconds
                                    initializeTimeSync();
                                });
                        }
                    }).fail(function(jqXHR, textStatus, errorThrown) {
                        logger.error("There was an error updating the date time on the panel. Retrying in 30 seconds...");
                        if (DateTime.now() <= DateTime.parse('2013-10-01T00:00:00')) {
                            // the date time is incorrect - tell the user they need to connect the unit to the internet
                            UserMessageController.postMessage('The date time is incorrect, please re-connect this panel to the internet to automatically re-sync the date time');

                            checkDateTimeIsCorrect();
                        } else {
                            // attempt to initalize the media player
                            initializeCoreServices();
                            initializeTimeSync();
                        }
                    });
                } else {
                    // init media player
                    logger.warning("Cannot sync date time for this player type. Starting Core services...");
                    initializeCoreServices();
                    initializeTimeSync();
                }
            } catch (e) {
                logger.error('Date time could not be checked');
                // attempt to initalize the media player
                initializeCoreServices();
                initializeTimeSync();
            }
        }

        /**
         * Starts the time sync process
         */
        function initializeTimeSync() {
            setTimeout(function() {
                checkDateTimeIsCorrect();
            }, 30000);
        }

        /**
         * Starts the player initialization after first boot
         */
        self.firstBootSequenceComplete = function() {
            initializeCoreServices();
        };

        /**
         * Initializes the player
         */
        function initializeCoreServices() {

            rotateCanvas();

            // by default delay the API interface by 10 seconds
            // to give the media player an opportunity to start playing content
            var apiInterfaceStartDelay = 10000;

            // check if the configuration file exists - if it does load it and start the media player
            if (File.exists(Settings.application.fileSystem.systemFiles.lastConfigFilePath)) {
                logger.debug('Loading last configuration');
                UserMessageController.postMessage('Loading saved configuration');
                File.read(Settings.application.fileSystem.systemFiles.lastConfigFilePath).done(function(configurationData) {
                    try {
                        // Store configuration from disk in memory
                        _currentConfiguration = new ConfigurationModel(JSON.parse(configurationData));

                        // Delete old files on startup - This only does anything if Settings.application.features.deleteOldFilesOnStartup = true in the
                        // client-specific configuration.
                        // NOTE: we always continue even if the file deletion fails as we do not want to prevent media playback from starting
                        deleteOldFilesOnStartUp().always(function() {

                            // If interrupts are supported then load then initialize the interrupt controller, check whether screen control is enabled and atart media playback
                            // Else just check whether screen control is enabled and start media playback
                            if (Settings.application.features.canPlayInterrupts) {
                                InterruptController.initialize(_mediaPlayerView, _currentConfiguration.content.interrupts, _currentConfiguration.screenControl)
                                    .always(function() {
                                        // Setup Screen Control
                                        // Only Setup screen control if it is a supported feature of this app
                                        // If screen control is enabled the media player will be started in the method that switches the screen on.
                                        if (Settings.application.features.canSwitchScreenOnOff) {
                                            setupScreenOnOffController();
                                        }
                                        // otherwise just start the media player
                                        else {
                                            startMediaPlayer();
                                        }
                                    });
                            } else {
                                // Setup Screen Control
                                // Only Setup screen control if it is a supported feature of this app
                                // If screen control is enabled the media player will be started in the method that switches the screen on.
                                if (Settings.application.features.canSwitchScreenOnOff) {
                                    setupScreenOnOffController();
                                }
                                // Otherwise start the media player in the standard way
                                else {
                                    startMediaPlayer();
                                }
                            }

                            // Remote log uploads
                            configureRemoteLogUploads(false);

                            // Start Scheduled reboots process
                            setupScheduledReboots();

                            //Scheduled screenshots
                            setupScheduledScreenshots();

                            // Start API Interface
                            startAPIInterface(apiInterfaceStartDelay);

                            // Start real Time Event Controller
                            startRealTimeEventsController();

                            // start the local API
                            startLocalAPI();
                        });
                    }
                    catch (e) {
                        logger.error('An error occurred loading configuration data at startup ' + e.message + '. Starting the API interface...');
                        UserMessageController.postMessage('Unable to load saved configuration. Connecting to Signagelive...');
                        // start the API interface immediately as there is nothing configured on the display
                        apiInterfaceStartDelay = 0;
                        startAPIInterface(apiInterfaceStartDelay);
                    }
                }).fail(function() {
                    logger.error('An error occurred reading the stored configuration data at startup. Starting the API interface...');
                    UserMessageController.postMessage('Unable to load saved configuration. Connecting to Signagelive...');
                    // start the API interface immediately as there is nothing configured on the display
                    apiInterfaceStartDelay = 0;
                    startAPIInterface(apiInterfaceStartDelay);
                });
            } else {
                logger.debug('No configuration to load');

                // start the API interface immediately as there is nothing configured on the display
                apiInterfaceStartDelay = 0;
                startAPIInterface(apiInterfaceStartDelay);

            }

            // Start Application Update Controller
            if (Settings.application.features.canPerformApplicationUpdates) {
                _appUpdateController = new ApplicationUpdateController();
                _appUpdateController.start();
            }

            // Start Firmware Update Controller
            if (Settings.application.features.canPerformFirmwareUpdates) {
                _firmwareUpdateController = new FirmwareUpdateController();
                _firmwareUpdateController.start();
            }

            // Setup Screen Capture Controller
            if (Settings.application.features.canTakeScreenshots) {
                if (_screenCapController == null) {
                    _screenCapController = new ScreenCaptureController();
                }
            }

            if (Settings.application.features.canListenForWebBasedInterrupts) {
                //Web Based Interrupts
                setupWebBasedInterrupts();
            }
        }

        function deleteOldFilesOnStartUp() {
            var dfd = $.Deferred();

            if (Settings.application.features.deleteOldFilesOnStartup) {
                logger.debug('Removing old files...');
                removeOldMediaFilesFromStorage(true).done(function() {
                    dfd.resolve();
                }).fail(function() {
                    logger.error('Error removing old files.');
                    dfd.reject();
                });
            } else {
                setTimeout(function() {
                    logger.debug('Player not configured to clear storage on startup');
                    dfd.resolve();
                }, 50);
            }

            return dfd;
        }

        /**
         * Starts the API Interface
         */
        function startAPIInterface(apiInterfaceStartDelay) {
            // Start Api Interface
            _apiInterface = new SLAPIInterface();
            setTimeout(function() {
                _apiInterface.initialize(_currentConfiguration, {
                    deviceRegistered: deviceRegisteredCallbackHandler,
                    deviceAlreadyRegistered: deviceAlreadyRegisteredCallbackHandler,
                    configuration: configurationCallbackHandler,
                    configurationNoChange: configurationNoChangeCallbackHandler
                });
            }, apiInterfaceStartDelay);

            // Initialize the Proof of Play Logger
            if (_currentConfiguration != null && _currentConfiguration.proofOfPlayConfig != null) {
                proofOfPlayLogger.init(_currentConfiguration.proofOfPlayConfig, Settings.application.api);
            }
        }

        function configureRemoteLogUploads(configUpdated) {
            LogUploadInterface.init(_currentConfiguration.logging, _currentConfiguration.clientDetails).then(function() {
                // Do an immediate upload of logs
                if (configUpdated && _currentConfiguration.logging.application && _currentConfiguration.logging.application.upload) {
                    LogUploadInterface.uploadLogs();
                }
            })
        }

        function startRealTimeEventsController() {
            RealTimeEventsController.updateConfiguration(_currentConfiguration);
        }

        function startLocalAPI() {
            if (_localAPIServer != null)
                return;

            _localAPIServer = new LocalAPIController(_currentConfiguration);
            _localAPIServer.initialize();

            _localAPIServer.onEvent = function(type, data) {

                switch (type) {
                    case "SoftTrigger":
                        _mediaPlayerView.softTrigger(data);
                        break;
                }
            }
        }


        /**
         * [deviceRegisteredCallbackHandler description]
         */
        function deviceRegisteredCallbackHandler(clientKey) {
            logger.debug('Registration Callback ' + clientKey);

            var d = JSON.stringify({
                clientKey: clientKey
            });
            File.write(Settings.application.fileSystem.systemFiles.clientKeyFilePath, d);

            // remember the client key
            Settings.application.api.clientKey = clientKey;
        }

        /**
         * Callback that is called when the device is already registered. Shows a message.
         */
        function deviceAlreadyRegisteredCallbackHandler() {
            UserMessageController.postMessage('This player has been previously connected to Signagelive, please deactivate it to continue with re-activation.\nID: ' + Settings.application.api.clientId);
        }

        /**
         * This will run after every configuration head request IF there are no changes
         */
        function configurationNoChangeCallbackHandler() {
            logger.info("Configuration checked - no update");

            uploadStoredScreenshots();

            if (Settings.application.features.canListenForWebBasedInterrupts) {
                setupWebBasedInterrupts();

                //If running, check for config update
                if (WebBasedInterruptsController.isRunning()) {
                    WebBasedInterruptsController.checkForConfigurationUpdate();
                }
            }
        }


        /**
         * Callback method called after a configuration update
         */
        function configurationCallbackHandler(configuration) {
            logger.info("Configuration checked - updating configuration");
            //logger.info("New Configuration: " + JSON.stringify(configuration, undefined, 2));
            // If this client is in extended desktop or window mode and max number of outputs changes
            // we will need to restart to implement the restriction
            var restartRequired = false;
            var displayModule = UserConfigurationController.modules().display;
            if (displayModule && (displayModule.multiMonitorModeEnabled === true || displayModule.windowModeEnabled === true)) {
                var currentMaxNumberOfOutputs = (_currentConfiguration) ? _currentConfiguration.clientDetails.maxNumberOfOutputs : 2; // If this is the first activition then _currentConfiguration has not been set so default to 2
                logger.info('Current max outputs: ' + currentMaxNumberOfOutputs);
                logger.info('Updated max outputs: ' + configuration.clientDetails.maxNumberOfOutputs)
                if (currentMaxNumberOfOutputs !== configuration.clientDetails.maxNumberOfOutputs) {
                    logger.info("Max number of outputs has changed. Restart required.");
                    restartRequired = true;
                }
            }
            // remember the previous config as we need this to handle any asset delete events
            _previousConfiguration = _currentConfiguration;
            // Stored updated configuration in memory
            _currentConfiguration = configuration;
            // If interrupts are supported then load then initialize the interrupt controller, check whether screen control is enabled and start media playback
            // Else just check whether screen control is enabled and start media playback
            if (Settings.application.features.canPlayInterrupts) {
                InterruptController.initialize(_mediaPlayerView, _currentConfiguration.content.interrupts, _currentConfiguration.screenControl)
                    .always(function() {
                        // Setup Screen Control
                        // Only Setup screen control if it is a supported feature of this app
                        // If screen control is enabled the media player will be started in the method that switches the screen on.
                        if (Settings.application.features.canSwitchScreenOnOff) {
                            setupScreenOnOffController();
                        }
                        // otherwise just start the media player
                        else {
                            startMediaPlayer();
                        }
                    });
            } else {
                // Setup Screen Control
                // Only Setup screen control if it is a supported feature of this app
                // If screen control is enabled the media player will be started in the method that switches the screen on.
                if (Settings.application.features.canSwitchScreenOnOff) {
                    setupScreenOnOffController();
                }
                // Otherwise start the media player in the standard way
                else {
                    startMediaPlayer();
                }
            }
            // Update config for remote logging
            configureRemoteLogUploads(true);
            // Update Scheduled reboots process
            setupScheduledReboots();
            // Scheduled screenshots
            setupScheduledScreenshots();
            // Schedule cleanup of the local storage
            setTimeout(function() {
                removeOldMediaFilesFromStorage(Settings.application.features.deleteOldFilesOnContentUpdate);
            }, 15000);
            // Taking an on demand screenshot in 20 seconds
            if (_currentConfiguration.logging.screencapture.screenshotAtNextContentCheck && _screenCapController != null) {
                setTimeout(function() {
                    logger.info("Taking on demand screenshot");
                    _screenCapController.takeOnDemandScreenshot();
                }, 20000);
            }
            // If an instant reboot is required then shedule it 30 seconds from now
            if (_currentConfiguration.connectivity.instantReboot) {
                logger.debug("Performing instant reboot in 30 seconds.");
                setTimeout(function() {
                    PowerInterface.reboot(RebootReasons.INSTANT_REBOOT);
                }, 30000);
            } else {
                // No instant reboot
                // Upload screenshots
                uploadStoredScreenshots();
            }
            // Initialize the Proof of Play Logger
            if (_currentConfiguration.proofOfPlayConfig != null) {
                logger.debug("Config Update: Updating Proof of Play config");
                proofOfPlayLogger.updateConfig(_currentConfiguration.proofOfPlayConfig, Settings.application.api);
            } else {
                logger.debug("Config Update: Proof of Play Configuration is null");
            }
            // Check firmware updates
            if (_currentConfiguration.firmwareUpdate != null) {
                doFirmwareUpdate(_currentConfiguration.firmwareUpdate);
            }

            if (Settings.application.features.canListenForWebBasedInterrupts) {
                setupWebBasedInterrupts();
                // If running, check for config update
                if (WebBasedInterruptsController.isRunning()) {
                    WebBasedInterruptsController.checkForConfigurationUpdate();
                }
            }
            // If restart is required due to max outputs changing on a device that supports multi-display/window mode setups
            if (restartRequired) {
                logger.info("Max Number of outputs has changed. Restarting application in 30 seconds...");
                setTimeout(function() {
                    PowerInterface.restartApplication();
                }, 30000);
            }
            // Notify Real Time Event Controller of any changes
            RealTimeEventsController.updateConfiguration(_currentConfiguration);
            if (_localAPIServer == null)
                startLocalAPI();
            // Notify local API of Config Change
            _localAPIServer.updateConfiguration(_currentConfiguration);
            // Update any remotely configurable player settings
            configureRemotePlayerSettings(_currentConfiguration);
        }

        function configureRemotePlayerSettings(config) {
            if (config.remoteSettings) {
                var remoteSettings = config.remoteSettings;
                // Native Portrait Rotation
                if (remoteSettings.nativePortraitRotation) {
                    logger.info('Native Protrait Rotation included in Remote Settings.');
                    GraphicsInterface.hasNativePortraitModeSettingChanged(remoteSettings.nativePortraitRotation)
                        .done(function(nativePortraitModeHasChanged) {
                            if (nativePortraitModeHasChanged) {
                                logger.info('Native Protrait Rotation setting has changed... Applying setting...');
                                // Note: timeout here to ensure that the media player has fully started before we attempt to stop it
                                setTimeout(function() {
                                    stopMediaPlayer(PlaybackStartEndReasons.REMOTE_SETTINGS_UPDATE);
                                    GraphicsInterface.setNativePortraitMode(remoteSettings.nativePortraitRotation)
                                        .done(function(updated) {
                                            logger.info('Native Portrait Settings Updated: ' + updated);
                                            // Adjust Layout Dimensions
                                            _mediaPlayerView.adjustLayoutViewDimensions(document.documentElement.clientWidth, document.documentElement.clientHeight);
                                            // Start player
                                            startMediaPlayer();
                                        });
                                }, 5000);
                            }
                        });
                }
                // Additional remote settings to be implemented here
            }
        } 

        function doFirmwareUpdate(firmwareUpdate) {
            var updateRequired = _firmwareUpdateController.isFirmwareUpdateRequired(firmwareUpdate);
            if (updateRequired) {
                _firmwareUpdateController.updateFirmware(firmwareUpdate);
            }
            else {
                _firmwareUpdateController.cancelUpdate();
            }
        }

        function setupWebBasedInterrupts() {

            var interruptsAvailable = false; // _currentConfiguration.content.interrupts.length > 0;

                //Always start the WBI due to the ability to trigger screen control
                if (WebBasedInterruptsController.isRunning() == false) {
                    logger.info("Interrupts available - start web based interrupts controller");
                    var options = {
                        url: Settings.application.webBasedInterrupts.url,
                        configurationCheckInterval: Settings.application.webBasedInterrupts.configurationCheckInterval,
                        serialNo: (_currentConfiguration) ? _currentConfiguration.clientDetails.signageliveSerialNumber : null,
                        switchScreenOn: switchScreenOn,
                        switchScreenOff: switchScreenOff
                    };

                    WebBasedInterruptsController.initialize(options)
                        .done(function() {
                            logger.info("Web Triggers inialized")
                        })
                        .fail(function() {
                            logger.error("Error: Failed to initialize web based interrupts controller");
                        });
                } else {
                    logger.info("Web based interrupts controller is already running");
                }

        }

        function stopWebBasedInterrupts() {
            if (WebBasedInterruptsController.isInitialized() == true) {
                WebBasedInterruptsController.stop();
            }
        }


        /**
         *  Upload screenshots
         */
        function uploadStoredScreenshots() {
            if (_screenCapController != null) {
                setTimeout(function() {
                    _screenCapController.uploadAllScreenshots();
                }, 30000);
            }
        }

        /**
         * Setup scheduled screenshots
         */
        function setupScheduledScreenshots() {
            logger.info("Setting up scheduled screenshots");
            if (Settings.application.features.canTakeScreenshots) {
                if (_screenCapController == null)
                    _screenCapController = new ScreenCaptureController();

                // Wait 5 seconds before restarting screenshots
                setTimeout(function() {
                    logger.info("Stopping and starting screenshots");
                    try {
                        // Stop the controller regardless
                        _screenCapController.stop();

                        // Supply the new config
                        _screenCapController.initialize(_currentConfiguration);

                        // Restart the controller - the controller will do nothing if it is not enabled
                        _screenCapController.start();
                    } catch (e) {
                        logger.error("Error configuring screenshots: " + e.message);
                    }
                }, 5000);
            }
        }

        var _screenOnOffTimer = 0;

        /**
         * [setupScreenOnOffController description]
         */
        function setupScreenOnOffController() {
            // clear any previous screen control timer
            clearTimeout(_screenOnOffTimer);

            // Read from file and switch screen on or off based on persisted file state
            readLastScreenStatus().then(function (screenInformation) {
                // switch the screen on/off according to the current setting and stored last screen state
                if (_currentConfiguration.screenControl.isScreenCurrentlyScheduledOn(screenInformation)) {
                    switchScreenOn(_currentConfiguration.screenControl.persistedState); // ...and either set or clear the persisted state
                } else {
                    switchScreenOff(_currentConfiguration.screenControl.persistedState); // ...and either set or clear the persisted state
                    $("#userMessage").empty();
                }

                // are there any schedules?
                if (_currentConfiguration.screenControl.nextAction() != null) {
                    setUpScreenOnOffTimer();
                }
            });
        }

        /**
         * Setup Screen Control timer
         */
        function setUpScreenOnOffTimer() {
            var nextAction = _currentConfiguration.screenControl.nextAction();

            if (nextAction == null)
                return;

            // calculate the time to the next action
            var interval = nextAction.time.subtract(DateTime.now());

            if (nextAction.action == 'on') {
                // set the on timer
                _screenOnOffTimer = setTimeout(function() {
                    switchScreenOn();
                    setUpScreenOnOffTimer();
                }, interval);
            } else {
                // set the off timer
                _screenOnOffTimer = setTimeout(function() {
                    switchScreenOff();
                    setUpScreenOnOffTimer();
                }, interval);
            }
        }

        /**
         * Screen Control - Turn Screen On
         */
        function switchScreenOn(persistState) {
            logger.info('Switching screen on');

            if(persistState) {
                _persistScreen = "ScreenOn";
            }

            // Use PowerInterface to switch the screen on
            PowerInterface.switchScreenOn(persistState)
                .done(function() {
                    $("#userMessageText").empty();
                    // Restart the media player
                    startMediaPlayer();
                });
        }

        /**
         * Screen Control - Turn screen off
         */
        function switchScreenOff(persistState) {
            logger.info('Switching screen off');

            if(persistState) {
                _persistScreen = "ScreenOff";
            }

            // Stop the media player
            stopMediaPlayer(PlaybackStartEndReasons.SCREEN_OFF);

            // Use PowerInterface to switch the screen off
            PowerInterface.switchScreenOff(persistState).done(function() {
                $("#userMessageText").empty();
                // Do nothing
            });
        }

        /**
         * Screen Control - Read last screen state command sent by WBT API
         */
        function readLastScreenStatus() {
            var dfd = $.Deferred();
            // Read from file and switch screen on or off based on persisted file state
            PowerInterface.readLastScreenStatus().then(function (screenInformation) {
                _persistScreen = (screenInformation) ? screenInformation.screenStatus : null;
                dfd.resolve(screenInformation);
            });
            return dfd;
        }

        /**
         * Checks the date time is correct
         */
        function checkDateTimeIsCorrect() {
            try {
                logger.info('Checking date time is correct');
                // check the date time is correct - for the specified time-zone?
                File.read(Settings.application.fileSystem.systemFiles.settingsFilePath)
                    .done(function(settingsData) {
                        var settings = JSON.parse(settingsData);
                        logger.info('URL for checking time is: ' + Settings.application.dateTimeServer + settings.timeZone.id);
                        $.ajax({
                            url: Settings.application.dateTimeServer + settings.timeZone.id,
                            contentType: 'application/json',
                            timeout: 10000,
                            cache: false
                        })
                        .done(function(data, textStatus, jqXHR) {
                            logger.info('Local Time for ' + data.time_zone + ' is ' + data.local_time);
                            var isDateTimeCorrect = true;
                            var twoMinsForwards = DateTime.now().addMinutes(2);
                            var twoMinsBackwards = DateTime.now().addMinutes(-2);
                            var serverTime = DateTime.parse(data.local_time);
                            if (serverTime > twoMinsBackwards & serverTime < twoMinsForwards)
                                isDateTimeCorrect = true;
                            else
                                isDateTimeCorrect = false;

                            if (!isDateTimeCorrect) {
                                logger.info('DateTime is incorrect... resetting it...');
                                // Set DateTime
                                DateTime.setDateTime(DateTime.parse(data.local_time))
                                    .done(function() {
                                        logger.info("Date time successfully updated.");
                                        // check in specified frequency
                                        setTimeout(function() {
                                            checkDateTimeIsCorrect();
                                        }, Settings.application.features.syncDateTimeFrequency * 1000);
                                    })
                                    .fail(function() {
                                        logger.error("There was an error updating the date time on the panel.");
                                        // check in 30 seconds
                                        setTimeout(function() {
                                            checkDateTimeIsCorrect();
                                        }, 30000);
                                    });

                            } else {
                                logger.info('DateTime is correct');
                                // check in specified frequency
                                setTimeout(function() {
                                    checkDateTimeIsCorrect();
                                }, Settings.application.features.syncDateTimeFrequency * 1000);

                            }
                        }).fail(function() {
                            logger.error("Error calling out to timezone server while checking datetime");
                            // check in 30 seconds
                            setTimeout(function() {
                                checkDateTimeIsCorrect();
                            }, 30000);
                        });
                    })
                    .fail(function(e) {
                        logger.error("Error reading settings file whilst checking datetime");
                        // check in 30 seconds
                        setTimeout(function() {
                            checkDateTimeIsCorrect();
                        }, 30000);
                    });
            } catch (e) {
                logger.error("Unexpected error checking datetime: " + e);
                // check in 30 seconds
                setTimeout(function() {
                    checkDateTimeIsCorrect();
                }, 30000);
            }
        }

        /**
         * Removes old media files
         * !! Note that this method needs to implement an interface so that each app can have their own implementation
         * as there is platform specific code here.
         * 
         * If doDelete is false it will work out what files to remove and will send RTE asset removed notifications for them but will
         * not actually delete the file. This solves an issue with ChromeOS where files are deleted after a restart and not straight after a config update like most players.
         * The issue is that we require the mediaAsset details from the previous config in order to send the RTE but do not have this after a reboot. There is also
         * a question about what exactly is meant by asset removed - is it when it is removed from the config or entirely removed from the player? In most cases this would be the same thing except for Chrome.
         * 
         * To solve sthis problem, when removeOldMediaFilesFromStorage() is called on start up doDelete is always true. When it is called after a config update it is based on the valye of Settings.application.features.deleteOldFilesOnContentUpdate, which will be true for all cases except Chrome. 
         */
        function removeOldMediaFilesFromStorage(doDelete) {

            var processDfd = $.Deferred();

            try {
                if (_currentConfiguration == null || Settings.application.features.canDownloadFiles == false)
                    return;

                logger.info('Checking directory: ' + Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.mediaDirectory);
                // get a list of all the files in the media directory
                var mediaDirPath = "";
                if (Settings.application.fileSystem.root !== "") {
                    mediaDirPath = mediaDirPath + Settings.application.fileSystem.root + "/";
                }
                if (Settings.application.fileSystem.mediaDirectory !== "") {
                    mediaDirPath = mediaDirPath + Settings.application.fileSystem.mediaDirectory;
                }
                // Get DirectoryInfo for media dir and init
                var mediaDirectory = new DirectoryInfo(mediaDirPath);
                mediaDirectory.initialize()
                    .done(function() {
                        logger.info('There are ' + mediaDirectory.files.length + ' files in the media directory');
                        var filesToDeleteCount = 0;
                        // see if each file is in the list of media assets we need and if not add them to list to be deleted
                        var filesToDelete = [];
                        $(mediaDirectory.files).each(function(i, file) {
                            if (file.type != 'folder') {
                                logger.info('Checking if ' + file.name + ' is required');
                                var mediaAssets = $.grep(_currentConfiguration.content.mediaAssets, function(mediaAsset, index) {
                                    return mediaAsset.localFileName === file.name;
                                });
                                if (mediaAssets.length != 1) {
                                    var mediaAssetToDelete = null;
                                    // find the media asset in the previous config
                                    if (_previousConfiguration != null) {
                                        var mediaAssetsInPreviousConfig = $.grep(_previousConfiguration.content.mediaAssets, function(ma, index) {
                                            return ma.localFileName === file.name;
                                        });
                                        if (mediaAssetsInPreviousConfig.length > 0)
                                            mediaAssetToDelete = mediaAssetsInPreviousConfig[0];
                                    }
                                    filesToDelete.push({ mediaAsset: mediaAssetToDelete, file: file });
                                    filesToDeleteCount++;
                                }
                            }
                        });
                        // Delete each file synchronously
                        logger.info('There are ' + filesToDeleteCount + ' to delete');
                        var dfd = $.Deferred().resolve();
                        filesToDelete.forEach(function(fileToDelete) {
                            dfd = dfd.then(function() {
                                return deleteOldFile(fileToDelete, doDelete);
                            });
                        });
                        // Report that deletion process has completed and resolve
                        dfd.done(function() {
                            logger.info('Media files that are no longer required have been removed.');
                            processDfd.resolve();
                        });
                    })
                    .fail(function() {
                        logger.error('Error loading directory information for ' + mediaDirPath);
                        processDfd.resolve();
                    });
            } catch (e) {
                logger.error('Error whilst removing files that are not required from the media directory: ' + e.message);
                processDfd.reject();
            }
            return processDfd;
        }

        // Delete a single media file
        function deleteOldFile(fileToDelete, doDelete) {
            var dfd = $.Deferred();
            var file = fileToDelete.file;
            try {
                
                // The file has been removed from the config but we aren't going to delete the physical file yet - this will happen after the next reboot (Currently Chrome ONLY).
                // But for now we still want to send the asset removed notification.
                if (doDelete === false) {
                    
                    if (fileToDelete.mediaAsset != null) {
                        logger.info('Media asset: ' + file.name + ' removed from config but not player. File will be removed on next reboot');
                        RealTimeEventsController.sendMediaAssetRemoved(fileToDelete.mediaAsset);
                    }
                    
                    return $.Deferred().resolve();
                }
                // Otherwise just follow the standard delete process
                else {
                    // This is a widget - need to delete the file and the extracted widget directory
                    if (file.path.indexOf('.wgt') != -1) {
                        // Delete .wgt file
                        logger.info('Widget File ' + file.name + ' is to be deleted.');
                        // !!!!! Temp solution for LG gapless files which we do not want to delete
                        if (file.name.lastIndexOf("gaplessPlaylist_", 0) === 0) {
                            logger.debug("Not deleting: " + file.name + " as this is a gapless playlist file.");
                            dfd.resolve();
                        } else {
                            file.delete().always(function(deleteSuccess) {
                                // Report file delete success
                                if (deleteSuccess) {
                                    logger.info(file.name + ' deleted.');
                                    // Log external systems this has been deleted
                                    if (fileToDelete.mediaAsset != null)
                                        RealTimeEventsController.sendMediaAssetRemoved(fileToDelete.mediaAsset);
                                } else {
                                    logger.warning('Warning: ' + file.name + ' was not deleted. It may have already been deleted.');
                                }
                                // Delete extracted .wgt directory
                                logger.info("Deleting widget directory and files...");
                                var start = file.path.lastIndexOf('/') + 1;
                                var widgetName = file.path.substr(start).replace('.wgt', '');
                                var widgetDir = "";
                                if (Settings.application.fileSystem.root !== "") {
                                    widgetDir = widgetDir + Settings.application.fileSystem.root + "/";
                                }
                                if (Settings.application.fileSystem.widgetDirectory !== "") {
                                    widgetDir = widgetDir + Settings.application.fileSystem.widgetDirectory + '/';
                                }
                                widgetDir = widgetDir + widgetName;
                                logger.info("Deleting widget directory and files: " + widgetDir);
                                Directory.deleteDirectory(widgetDir)
                                    .done(function() {
                                        logger.info("Widget directory deleted");
                                    })
                                    .fail(function(error) {
                                        logger.info("An error occurred when trying to delete the widget directory: " + error);
                                    })
                                    .always(function() {
                                        // resolve delete
                                        dfd.resolve();
                                    });
                            });
                        }
                    }
                    // Standard assets- just delete the file
                    else {
                        // !!!!! Temp solution for LG gapless files, which we do not want to delete
                        if (file.name.lastIndexOf("gaplessPlaylist_", 0) === 0) {
                            logger.debug("Not deleting: " + file.name + " as this is a gapless playlist file.");
                            dfd.resolve();
                        } else {
                            logger.info('File ' + file.name + ' is to be deleted.');
                            file.delete().always(function(deleteSuccess) {
                                // Report file delete success
                                if (deleteSuccess) {
                                    logger.info(file.name + ' deleted.');
                                    // Log external systems this has been deleted
                                    if (fileToDelete.mediaAsset != null)
                                        RealTimeEventsController.sendMediaAssetRemoved(fileToDelete.mediaAsset);
                                } else {
                                    logger.warning('Warning: ' + file.name + ' not deleted.');
                                }
                                dfd.resolve();
                            });
                        }
                    }
                }
            } catch (e) {
                logger.error('Error deleting ' + file.name + ' error: ' + e.message);
                dfd.resolve();
            }
            return dfd.promise();
        }

        /**
         * Configures remote control listener
         */
        function setupRemoteControlListener() {

            // give focus to the remote focus anchor to capture events from the remote control
            $('#remoteFocus').focus();

            RemoteControlInterface.initialize();

            // Handle Key Events
            $('#remoteFocus').keydown(function(event) {
                var keyCode = event.keyCode;
                logger.info('KEY PRESSED: ' + keyCode);
                var key = "";

                switch (keyCode) {
                    case RemoteControlInterface.keys.KEY_UP:
                        key = "UP";
                        break;
                    case RemoteControlInterface.keys.KEY_DOWN:
                        key = "DOWN";
                        break;
                    case RemoteControlInterface.keys.KEY_LEFT:
                        key = "LEFT";
                        break;
                    case RemoteControlInterface.keys.KEY_RIGHT:
                        key = "RIGHT";
                        break;
                    case RemoteControlInterface.keys.KEY_RW:
                        logger.info('RW key pressed');
                        _mediaPlayerView.back();
                        break;
                    case RemoteControlInterface.keys.KEY_FF:
                        logger.info('FF key pressed');
                        _mediaPlayerView.skip();
                        break;
                    case RemoteControlInterface.keys.KEY_PLAY:
                        logger.info('PLAY key pressed');
                        _mediaPlayerView.play();
                        break;
                    case RemoteControlInterface.keys.KEY_PAUSE:
                        logger.info('PAUSE key pressed');
                        key = "PAUSE";
                        break;
                    case RemoteControlInterface.keys.KEY_MUTE:
                        logger.info('MUTE key pressed');
                        UserConfigurationController.toggleMute();
                        UserConfigurationController.saveSettings();
                        break;
                    case RemoteControlInterface.keys.KEY_VOL_UP:
                        logger.info('VOL_UP key pressed');
                        UserConfigurationController.volumeUp();
                        try {
                            SoundInterface.getVolume().done(function(level) {
                                level++;
                                SoundInterface.setVolume(level);
                            });
                        } catch (error) {
                            logger.error(error);
                        }
                        break;
                    case RemoteControlInterface.keys.KEY_VOL_DOWN:
                        logger.info('VOL_DOWN key pressed');
                        UserConfigurationController.volumeDown();
                        try {
                            SoundInterface.getVolume().done(function(level) {
                                level--;
                                SoundInterface.setVolume(level);
                            });
                        } catch (error) {
                            logger.error(error);
                        }
                        break;
                    case RemoteControlInterface.keys.KEY_INFO:
                        logger.info('INFO key pressed');
                        // If it isnt already being shown then show the device Information View
                        if (!self.propertiesViewVisible)
                            showDeviceInformationView();
                        break;
                    case RemoteControlInterface.keys.KEY_SETUP:
                        logger.info('SETUP key pressed');
                        // Show setup view
                        if (!self.setupViewVisible)
                            showSetupView();
                        break;
                    case RemoteControlInterface.keys.KEY_ENTER:
                        logger.info('ENTER key pressed');
                        key = "ENTER";
                        break;
                    case RemoteControlInterface.keys.KEY_0:
                    case RemoteControlInterface.keys.KEY_1:
                    case RemoteControlInterface.keys.KEY_2:
                    case RemoteControlInterface.keys.KEY_3:
                    case RemoteControlInterface.keys.KEY_4:
                    case RemoteControlInterface.keys.KEY_5:
                    case RemoteControlInterface.keys.KEY_6:
                    case RemoteControlInterface.keys.KEY_7:
                    case RemoteControlInterface.keys.KEY_8:
                    case RemoteControlInterface.keys.KEY_9:
                        logger.info('Remote key pressed: ' + keyCode);
                        if (RemoteControlInterface.keyCodesToInterruptKeys) {
                            var interruptKey = RemoteControlInterface.keyCodesToInterruptKeys[keyCode];
                            if (interruptKey != null) {
                                logger.info('Attempting to trigegr interrupt for key: ' + interruptKey);
                                InterruptController.triggerInterrupt(interruptKey, 'KeyboardEvent', false, true).fail(function() {
                                    _mediaPlayerView.onKeyPress(interruptKey);
                                });
                            }
                        }

                        break;
                    case RemoteControlInterface.keys.KEY_POWER:
                        // If screen currently scheduled to be off through screen control then we need to be able to override this with the PWR button, else we need the PWR button to be tracked and take priority like the WBT API
                        if(_mediaPlayerStarted) {
                            switchScreenOff(true);
                        } else {
                            switchScreenOn(true);
                        }
                    default:
                        logger.info("Key pressed: " + keyCode);
                        // Do Nothing
                        break;
                }

                if (key != "") {
                    _mediaPlayerView.onKeyPress(key);
                }
            });
        }

        /**
         * Shows setup view
         */
        function showSetupView() {

            if (self.setupViewVisible)
                return;

            if (_currentConfiguration.screenControl.isScreenCurrentlyScheduledOff() && _persistScreen !== "ScreenOn"  || _persistScreen == "ScreenOff" || !_mediaPlayerStarted)
                return;

            require(["gui/views/SetupView"], function(SetupView) {
                // Stop the media player
                stopMediaPlayer(PlaybackStartEndReasons.SETUP_SCREEN);
                // Remove the box that opens the settings menu
                //removeSettingsSelectionSquare();
                // Configure new Setup View
                var setupView = new SetupView({
                    complete: function(updatedSettings) {
                        // Hide the Setup View
                        $("#" + setupView.id).remove();
                        self.setupViewVisible = false;
                        // Call the setupCompleteHandler and pass updated settings
                        setupCompleteHandler(updatedSettings);
                    },
                    cancel: function() {
                        // Hide the Setup View
                        $("#" + setupView.id).remove();
                        self.setupViewVisible = false;
                        readLastScreenStatus().then(function (screenInformation) {
                            if (!screenInformation || (screenInformation && screenInformation.screenStatus !== "ScreenOff")) {
                                // restart media player
                                startMediaPlayer();
                            }
                        });
                    }
                }, true, _currentConfiguration.remoteSettings);
                // Add to Canvas
                self.setupViewVisible = true;
                $('#canvas').append(setupView.html);
                setupView.addedToCanvas();
            });
        }

        /**
         * Shows transparent device information overlay
         */
        function showDeviceInformationView() {

            if (self.propertiesViewVisible)
                return;

            if (_currentConfiguration.screenControl.isScreenCurrentlyScheduledOff() && _persistScreen !== "ScreenOn" || _persistScreen == "ScreenOff" || !_mediaPlayerStarted)
                return;

            require(["gui/views/DevicePropertiesView"], function(DevicePropertiesView) {

                // Instantiate new DevicePropertiesView and initialize it (load the model and HTML)
                var propertiesView = new DevicePropertiesView();
                propertiesView.initialize().done(function() {
                    // For BH5C we need to stop the media player
                    if (Settings.application.model === '86BH5C') {
                        stopMediaPlayer(PlaybackStartEndReasons.SETUP_SCREEN);
                    }

                    // initialized so show it
                    self.propertiesViewVisible = true;
                    $('#canvas').append(propertiesView.html);

                    // And then hide it automatically after 15000 millisecodnds
                    setTimeout(function() {
                        self.propertiesViewVisible = false;
                        $('#' + propertiesView.id).remove();
                        if (Settings.application.model === '86BH5C') {
                            startMediaPlayer();
                        }
                    }, 15000);
                });
            });
        }

        /**
         * Called only on firstboot. Shows the SetupView.
         */
        function firstBootHandler() {

            logger.debug("First boot Handled Called");
            _isFirstBoot = true;

            // Show Setup View
            require(["gui/views/SetupView"], function(SetupView) {
                var setupView = new SetupView({
                    complete: function(updatedSettings) {
                        $("#" + setupView.id).remove();
                        setupCompleteHandler(updatedSettings);
                    }
                }, false, true);

                $('#canvas').append(setupView.html);
                setupView.addedToCanvas();
            });
        }

        /**
         * Callback that gets called after setup is completed
         */
        function setupCompleteHandler(updatedSettings) {
            logger.info('Settings complete ' + JSON.stringify(updatedSettings));

            try {
                
                // Configure the way layouts are resized
                if (updatedSettings.layoutFit && updatedSettings.layoutFit.id != UserConfigurationController.layoutFit().id) {
                    UserConfigurationController.layoutFit(updatedSettings.layoutFit);
                }

                // usePMModeForScreenManagement - webOS 4.1+
                usePMModeForScreenManagement(updatedSettings).then(function() {
                    // Content Rotation
                    if (updatedSettings.rotation.angle != UserConfigurationController.rotation().angle) {
                        // Update Rotation
                        UserConfigurationController.rotation(updatedSettings.rotation);
                        // If this isn't the first boot and timezones havent changed then rotate the canvas and restart playback
                        // Else this will happen automatically when the firstBootSequenceComplete() mthos is called
                        if (!_isFirstBoot && updatedSettings.timeZone.id == UserConfigurationController.timeZone().id) {
                            rotateCanvas();
                        }
                    }
                    // Module settings
                    UserConfigurationController.modules(updatedSettings.modules);
                    // Rotate Videos (Tizen only)
                    if (updatedSettings.rotateVideos !== undefined) {
                        UserConfigurationController.rotateVideos(updatedSettings.rotateVideos);
                    }
                    // Firmware Updates enabled?
                    if (Settings.application.features.canPerformFirmwareUpdates) {
                        // Save firmware preferences
                        UserConfigurationController.remoteFirmwareUpdates(updatedSettings.remoteFirmwareUpdates);
                        // Start firmware update process if it isn't already started
                        if (updatedSettings.remoteFirmwareUpdates.enabled == true && (_firmwareUpdateController == null || _firmwareUpdateController.started == false)) {
                            _firmwareUpdateController = new FirmwareUpdateController();
                            _firmwareUpdateController.start();
                        }
                        // Stop the process
                        if (updatedSettings.remoteFirmwareUpdates.enabled == false && _firmwareUpdateController != null && _firmwareUpdateController.started === true) {
                            _firmwareUpdateController.stop();
                        }
                    }
                    // NTP Setting (LG only)
                    var ntpSettingUpdated = false;
                    if (typeof updatedSettings.useNTP !== "undefined" && updatedSettings.useNTP != UserConfigurationController.useNTP()) {
                        UserConfigurationController.useNTP(updatedSettings.useNTP);
                        ntpSettingUpdated = true;
                    }
                    // Timezome
                    var timezoneSettingUpdated = false;
                    if (updatedSettings.timeZone.id != UserConfigurationController.timeZone().id) {
                        UserConfigurationController.timeZone(updatedSettings.timeZone);
                        timezoneSettingUpdated = true;
                    }
                    //Kiosk mode
                    var kioskSettingUpdated = false;
                    if(updatedSettings.modules.interactiveKiosk && updatedSettings.modules.interactiveKiosk.supported && (updatedSettings.modules.interactiveKiosk.enabled != UserConfigurationController.interactiveKiosk().enabled || updatedSettings.modules.interactiveKiosk.idleIntervalInSeconds != UserConfigurationController.interactiveKiosk().idleIntervalInSeconds))
                    {
                        UserConfigurationController.interactiveKiosk(updatedSettings.modules.interactiveKiosk);
                        kioskSettingUpdated = true;
                    }
                    
                    // Save settings
                    UserConfigurationController.saveSettings()
                        .always(function() {
                            // If it is the first boot OR Timezone/NTP settings have changed we need to update the panel and may require a restart
                            if (_isFirstBoot || ntpSettingUpdated || timezoneSettingUpdated) {
                                logger.info('Either first boot, NTP settings updated or Timezone settings updated. Setting Timezone and Time as appropriate...');
                                // Update the timezone if we can
                                resyncDateTimeAfterSettingsChange(updatedSettings)
                                    .always(function(timeUpdated) {
                                        updateTimeZone(updatedSettings)
                                            .always(function(timezoneUpdated) {
                                                // Restart?
                                                logger.info('timezoneUpdated: ' + timezoneUpdated + ', NTP Updated: ' +  'Time updated: ' + timeUpdated + 'Use NTP: ' + updatedSettings.useNTP)
                                                var restartRequired = timezoneUpdated === true || ntpSettingUpdated === true || (timeUpdated === true && updatedSettings.useNTP === true);
                                                if (restartRequired) {
                                                    logger.info("Restart Required. Rebooting in 5 seconds...");
                                                    UserMessageController.postMessage("Settings applied. Restarting device...");
                                                    setTimeout(function() {
                                                        PowerInterface.reboot(RebootReasons.TIMEZONE_UPDATE);
                                                    }, 3000);
                                                }
                                                else {
                                                    readLastScreenStatus().then(function (screenInformation) {
                                                        if (!screenInformation || (screenInformation && screenInformation.screenStatus !== "ScreenOff")) {
                                                            if (_isFirstBoot) {
                                                                _isFirstBoot = false;
                                                                self.firstBootSequenceComplete();
                                                            } else {
                                                                startMediaPlayer();
                                                            }
                                                        }
                                                    });
                                                }
                                            });
                                    });
                            }
                            // Otherwise just restart playback (as we now stop playback when the setup screen is showing)
                            else {
                                readLastScreenStatus().then(function (screenInformation) {
                                    if (!screenInformation || (screenInformation && screenInformation.screenStatus !== "ScreenOff")) {
                                        startMediaPlayer();
                                    }
                                });                  
                            } 
                        });
                });
            }
            catch (e) {
                logger.error('Error whilst saving settings : ' + e.message);
                startMediaPlayer();
            }
        }

        function usePMModeForScreenManagement(updatedSettings) {
            var dfd = $.Deferred();
            if (updatedSettings.usePMModeForScreenManagement != UserConfigurationController.usePMModeForScreenManagement()) {
                if(updatedSettings.usePMModeForScreenManagement) {
                    try {
                        PowerInterface.getPMMode().then(function(mode) {
                            if(mode !== "screenOffAlways") {
                                PowerInterface.setPMMode().then(function() {
                                    logger.info("PM Mode set to screenOffAlways");
                                    dfd.resolve();
                                });
                            } else {
                                dfd.resolve();
                            }
                        });
                        UserConfigurationController.usePMModeForScreenManagement(updatedSettings.usePMModeForScreenManagement);
                    } catch(e) {
                        logger.warning("Issue getting/setting PM Mode - reverting change");
                        updatedSettings.usePMModeForScreenManagement = UserConfigurationController.usePMModeForScreenManagement();
                        dfd.resolve();
                    }
                } else {
                    UserConfigurationController.usePMModeForScreenManagement(updatedSettings.usePMModeForScreenManagement);
                    dfd.resolve();
                }
            } else {
                dfd.resolve();
            }
            return dfd;
        };

        function updateTimeZone(updatedSettings) {
            logger.info("Updating timezone...");
            var dfd = $.Deferred();
            var restartRequired = false;
            if (Settings.application.features.canSetTimeZone === true && updatedSettings.useNTP === true) {
                ClockInterface.setTimeZone(updatedSettings.timeZone.id, updatedSettings.timeZone)
                    .done(function() {
                        logger.info("Timezone updated to " + updatedSettings.timeZone.id);
                        restartRequired = true;
                    })
                    .fail(function() {
                        logger.error("Error updating the timezone");
                    })
                    .always(function() {
                        dfd.resolve(restartRequired);
                    });
                return dfd;
            }
            else {
                return dfd.resolve(false);
            }
        }

        function resyncDateTimeAfterSettingsChange(updatedSettings) {
            logger.info("Resyncing DateTime...");
            var dfd = $.Deferred();
            if (Settings.application.features.canSyncDateTime === true) {
                $.ajax({
                    url: Settings.application.dateTimeServer + updatedSettings.timeZone.id,
                    contentType: 'application/json'
                }).done(function(data) {
                    logger.info('Local Time for ' + data.time_zone + ' is ' + data.local_time);
                    DateTime.setDateTime(DateTime.parse(data.local_time), updatedSettings.useNTP)
                        .done(function() {
                            logger.info("Date time successfully updated.");
                            dfd.resolve(true);
                        })
                        .fail(function() {
                            logger.error("There was an error updating the date time on the panel.");
                            dfd.resolve(false);
                        });
                    })
                    .fail(function() {
                        logger.error('Failed to retrieve remote date time');
                        dfd.resolve(false);
                    });
                return dfd;
            } else {
                return dfd.resolve(false);
            }
        }

        /**
         * Rotates the canvas - used to enable portrait support
         */
        function rotateCanvas() {
            if (UserConfigurationController.isRotated()) {
                switch (UserConfigurationController.rotationAngle()) {
                    case 90:

                        $('#canvas').width(document.documentElement.clientHeight);
                        $('#canvas').height(document.documentElement.clientWidth);
                        $('#canvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#canvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#canvas').removeClass('rotate180');
                        $('#canvas').removeClass('rotate270');
                        $('#canvas').addClass('rotate90');

                        $('#layoutBackgroundCanvas').attr('width', document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').attr('height', document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').width(document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').height(document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#layoutBackgroundCanvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#layoutBackgroundCanvas').removeClass('rotate180');
                        $('#layoutBackgroundCanvas').removeClass('rotate270');
                        $('#layoutBackgroundCanvas').addClass('rotate90');

                        $('#layoutBackgroundImage').attr('width', document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').attr('height', document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').width(document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').height(document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#layoutBackgroundImage').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#layoutBackgroundImage').removeClass('rotate180');
                        $('#layoutBackgroundImage').removeClass('rotate270');
                        $('#layoutBackgroundImage').addClass('rotate90');

                        _mediaPlayerView.adjustLayoutViewDimensions(document.documentElement.clientHeight, document.documentElement.clientWidth);

                        break;
                    case 180:

                        $('#canvas').css('left', '0px');
                        $('#canvas').css('top', '0px');
                        $('#canvas').width(document.documentElement.clientWidth);
                        $('#canvas').height(document.documentElement.clientHeight);
                        $('#canvas').removeClass('rotate90');
                        $('#canvas').removeClass('rotate270');
                        $('#canvas').addClass('rotate180');

                        $('#layoutBackgroundCanvas').css('left', '0px');
                        $('#layoutBackgroundCanvas').css('top', '0px');
                        $('#layoutBackgroundCanvas').attr('width', document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').attr('height', document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').width(document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').height(document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').removeClass('rotate90');
                        $('#layoutBackgroundCanvas').removeClass('rotate270');
                        $('#layoutBackgroundCanvas').addClass('rotate180');

                        $('#layoutBackgroundImage').css('left', '0px');
                        $('#layoutBackgroundImage').css('top', '0px');
                        $('#layoutBackgroundImage').attr('width', document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').attr('height', document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').width(document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').height(document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').removeClass('rotate90');
                        $('#layoutBackgroundImage').removeClass('rotate270');
                        $('#layoutBackgroundImage').addClass('rotate180');

                        _mediaPlayerView.adjustLayoutViewDimensions(document.documentElement.clientWidth, document.documentElement.clientHeight);

                        break;
                    case 270:
                        $('#canvas').width(document.documentElement.clientHeight);
                        $('#canvas').height(document.documentElement.clientWidth);
                        $('#canvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#canvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#canvas').removeClass('rotate90');
                        $('#canvas').removeClass('rotate180');
                        $('#canvas').addClass('rotate270');

                        $('#layoutBackgroundCanvas').attr('width', document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').attr('height', document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').width(document.documentElement.clientHeight);
                        $('#layoutBackgroundCanvas').height(document.documentElement.clientWidth);
                        $('#layoutBackgroundCanvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#layoutBackgroundCanvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#layoutBackgroundCanvas').removeClass('rotate90');
                        $('#layoutBackgroundCanvas').removeClass('rotate180');
                        $('#layoutBackgroundCanvas').addClass('rotate270');

                        $('#layoutBackgroundImage').attr('width', document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').attr('height', document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').width(document.documentElement.clientHeight);
                        $('#layoutBackgroundImage').height(document.documentElement.clientWidth);
                        $('#layoutBackgroundImage').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#layoutBackgroundImage').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');
                        $('#layoutBackgroundImage').removeClass('rotate90');
                        $('#layoutBackgroundImage').removeClass('rotate180');
                        $('#layoutBackgroundImage').addClass('rotate270');

                        _mediaPlayerView.adjustLayoutViewDimensions(document.documentElement.clientHeight, document.documentElement.clientWidth);

                        break;
                }


            } else {
                // remove rotation

                $('#canvas').removeClass('rotate90');
                $('#canvas').removeClass('rotate180');
                $('#canvas').removeClass('rotate270');
                $('#canvas').css('left', '0px');
                $('#canvas').css('top', '0px');
                $('#canvas').width(document.documentElement.clientWidth);
                $('#canvas').height(document.documentElement.clientHeight);

                $('#layoutBackgroundCanvas').removeClass('rotate90');
                $('#layoutBackgroundCanvas').removeClass('rotate180');
                $('#layoutBackgroundCanvas').removeClass('rotate270');
                $('#layoutBackgroundCanvas').css('left', '0px');
                $('#layoutBackgroundCanvas').css('top', '0px');
                $('#layoutBackgroundCanvas').attr('width', document.documentElement.clientWidth);
                $('#layoutBackgroundCanvas').attr('height', document.documentElement.clientHeight);
                $('#layoutBackgroundCanvas').width(document.documentElement.clientWidth);
                $('#layoutBackgroundCanvas').height(document.documentElement.clientHeight);

                $('#layoutBackgroundImage').removeClass('rotate90');
                $('#layoutBackgroundImage').removeClass('rotate180');
                $('#layoutBackgroundImage').removeClass('rotate270');
                $('#layoutBackgroundImage').css('left', '0px');
                $('#layoutBackgroundImage').css('top', '0px');
                $('#layoutBackgroundImage').attr('width', document.documentElement.clientWidth);
                $('#layoutBackgroundImage').attr('height', document.documentElement.clientHeight);
                $('#layoutBackgroundImage').width(document.documentElement.clientWidth);
                $('#layoutBackgroundImage').height(document.documentElement.clientHeight);

                _mediaPlayerView.adjustLayoutViewDimensions(document.documentElement.clientWidth, document.documentElement.clientHeight);
            }
        }

        /**
         * Setup Scheduled Reboots
         */
        function setupScheduledReboots() {

            logger.debug("Configuring Scheduled Reboots");

            try {

                // clear any previous scheduledReboots timer
                if (_scheduledRebootTimer != null)
                    clearTimeout(_scheduledRebootTimer);

                if (_currentConfiguration.connectivity.scheduledRebootsEnabled) {

                    logger.debug("Scheduled Reboots Enabled");

                    if (_currentConfiguration.connectivity.systemRebootTime == null) {
                        logger.error("System Reboot Time is Null. Cannot Configure Scheduled Reboots.");
                        return;
                    }

                    // The timeout till the next reboot
                    var rebootTimeout = 0;

                    // work out timeout
                    var now = DateTime.now();
                    var todaysRebootTime = DateTime.now().setTime(_currentConfiguration.connectivity.systemRebootTime);
                    var tomorrowsRebootTime = DateTime.now().setTime(_currentConfiguration.connectivity.systemRebootTime).addDays(1);

                    // Reboot Later today
                    if (now.getTime() < todaysRebootTime.getTime()) {
                        rebootTimeout = todaysRebootTime.getTime() - now.getTime();
                        logger.debug("Next Scheduled Reboot Today. Reboot Time: " + todaysRebootTime.toLocaleString());
                    }
                    // Reboot Tomorrow
                    else {
                        rebootTimeout = tomorrowsRebootTime.getTime() - now.getTime();
                        logger.debug("Next Scheduled Reboot Tomorrow. Reboot Time: " + tomorrowsRebootTime.toLocaleString());
                    }

                    // Only reboot if reboot Time out in the future
                    if (rebootTimeout > 0) {

                        _scheduledRebootTimer = setTimeout(function() {
                            PowerInterface.reboot(RebootReasons.SCHEDULED_REBOOT);
                            logger.debug("Reboot Scheduled. Timeout (ms): " + rebootTimeout);
                        }, rebootTimeout + 30000);

                    }
                } else
                    logger.debug("Scheduled Reboots Not Enabled.");
            } catch (e) {
                logger.error(e.message);
            }
        }

        function triggerKeyboardInterrupt(key) {
            logger.info('Interrupt Triggered: ' + key);
            InterruptController.triggerInterrupt(key, 'KeyboardEvent', false, true);
        }

        /**
         * Legacy - no longer used.
         */
        function removeSettingsSelectionSquare() {
            $('#settingsOptionBox').remove();
        }


        /**
         * Method that starts the media player.
         */
        function startMediaPlayer() {
            // Start the media player
            logger.info("Starting Media Player");
            _mediaPlayerStarted = true;

            _mediaPlayerView.loadContent(_currentConfiguration.content, _currentConfiguration.synchronization, _currentConfiguration.clientDetails);

            // Load Generic Modules
            logger.info("Starting Web-based Interrupts");
            setupWebBasedInterrupts();

            $.each(Settings.application.modules, function(moduleName, module) {

                // We are only interested in loading modules that are supported and enabled
                if (module.supported && module.enabled) {
                    // Attempt to load the module.
                    //logger.info(JSON.stringify(UserConfigurationController.modules().interactiveKiosk));
                    // Module Controller is located in controllers/[moduleName]Controller.js
                    var moduleController = "controllers/" + moduleName.charAt(0).toUpperCase() + moduleName.slice(1) + "Controller";
                    require([moduleController], function(controller) {

                        // Controller loaded via RequireJS. Initialize it passing the module options
                        controller.initialize(module.options)
                            // Successfully initialized
                            .done(function() {
                                // Attempt to start the controller
                                controller.start()
                                    // Successfully started
                                    .done(function() {
                                        logger.info("Module: " + moduleName + " started.");
                                    })
                                    // Failed to start
                                    .fail(function() {
                                        logger.error("Error: Failed to start module: " + moduleName);
                                    });
                            })
                            // Failed to initialize
                            .fail(function() {
                                logger.error("Error: Failed to initialize module: " + moduleName);
                            });

                    });
                }

            });
        }

        function stopMediaPlayer(endReason) {
            // Stop the media player
            logger.info("Stopping Media Player");
            _mediaPlayerStarted = false;

            _mediaPlayerView.stop(endReason);
            // Web-based interrupts
            // logger.info("Stopping web-based interrupts");
            // stopWebBasedInterrupts();
            // Stop all modules
            logger.info("Stopping Modules");
            $.each(Settings.application.modules, function(moduleName, module) {

                // We are only interested in loading modules that are supported and enabled
                if (module.supported && module.enabled) {
                    var moduleController = "controllers/" + moduleName.charAt(0).toUpperCase() + moduleName.slice(1) + "Controller";
                    require([moduleController], function(controller) {

                        // Stop the module
                        controller.stop()
                            // Success
                            .done(function() {
                                logger.info("Module: " + moduleName + " stopped");
                            })
                            // Failed
                            .fail(function() {
                                logger.warning("Error: Failed to stop module" + moduleName);
                            });
                    });
                }
            });
        }

        /* Only required for ChromeOS devices
        * When a device is booted and no displays are connected - it will not show the correct resolution once a display is added.
        * Therefore when a display is first connected we need to check the resolutions and update the DOM accordingly
        */
        try {
            if(chrome && chrome.system && chrome.system.display && GraphicsInterface.getDisplayInfo && GraphicsInterface.checkLastDisplayConfiguration && GraphicsInterface.setDOMProperties && GraphicsInterface.setLastDisplayConfiguration) {
                    /* Before we complete this check we want to see what the DOM currently reports as, as well as the number of displays connected.
                    * For example - if 2 displays are already connected upon bootup then we dont want to change the DOM.
                    * Although in its default state - ChromeOS will report that there is 1 display connected even if nothing is being used - and the resolution of that is 1366x768                    
                    */
                    GraphicsInterface.getDisplayInfo().then(function(displays) {
                        GraphicsInterface.setLastDisplayConfiguration($('body').height(), $('body').width(), displays.length);
                    });
                    /* Add the event listener to be able to determine when a display is added/removed
                    * This event can fire multiple times quickly so we only want to complete the callback once per actual event firing
                    */ 
                    chrome.system.display.onDisplayChanged.addListener(function() {
                        if(!displayDetectioninProgress) {
                            displayDetectioninProgress = true;
                            // Get the new set of display information with singleUnified false so we know the correct amount of displays connected
                            GraphicsInterface.getDisplayInfo().then(function(displays) {
                                logger.info("Displays connected: " + displays.length);

                                var height = displays[0].bounds.height;
                                var width = displays[0].bounds.width;

                                if(GraphicsInterface.checkLastDisplayConfiguration(height, width, displays.length) == false) {            
                                    logger.info("Last display configuration does not match current setup  --- Restarting media player ...");
                                    // Update the DOM with the correct dimensions
                                    GraphicsInterface.setDOMProperties(height, width);
                                    // Update the last display configuration with the new details
                                    GraphicsInterface.setLastDisplayConfiguration(height, width, displays.length);
                                    // Restart the media player so the content scales correctly
                                    stopMediaPlayer();
                                    rotateCanvas();
                                    startMediaPlayer();
                                } else {
                                    logger.info("Last display configuration matches current setup");
                                }

                                // We reapply the saved display settings whenever there is a display state change due to a previous issue where rotation settings were incorrectly being applied at 0/landscape
                                if(GraphicsInterface.loadDisplaySettings) {
                                    logger.info("Reapplying display settings...");
                                    GraphicsInterface.loadDisplaySettings();
                                }

                                displayDetectioninProgress = false;
                            });
                        }
                    });
            };
        } catch (e) {
            // Not Chrome OS device
        }
    }

    return SLClientEngine;
});