define(["system/DateTime", 'system/SystemInitializer', "system/io/File", "system/io/Directory", "properties/App.Properties", "properties/App.Settings", "core/SLClientEngine", "gui/views/LayoutView", "controllers/UserConfigurationController", "controllers/UserMessageController", "system/SystemInterface", "system/power/PowerInterface", "system/ClientIDGenerator", "core/LogUploadInterface"], function(DateTime, SystemInitializer, File, Directory, Properties, Settings, SLClientEngine, LayoutView, UserConfigurationController, UserMessageController, SystemInterface, PowerInterface, ClientIDGenerator, LogUploadInterface) {
    /**
     * [SLMediaPlayerApp description]
     * @param {[type]} options [description]
     */
    function SLMediaPlayerApp(options) {
        var self = this;
        var _engine = null;
        self.systemInitializer = null;

        // update the default options with the user config
        $.extend(true, Settings, options);

        /**
         * [initialize description]
         * @return {[type]} [description]
         */
        self.initialize = function() {
            // setup debug tracer
            setupDebugTracer();
            // Setup PoP logger
            if (proofOfPlayLogger == null) {
                logger.debug("Proof of Play Logger not created.");
            } else {
                logger.debug("Proof of Play Logger created.");
            }
            logger.info('Initializing SL MediaPlayer App');
            // perform any setup of the hardware on first boot
            setupSystem().done(function() {
                // setup core UI
                setupCoreUI();
                // initialize file system
                initializeFileSystem().done(function() {
                    logger.debug('File system Initalised');
                    // initialize application settings
                    initializeAppSettings().done(function() {
                        self.systemInitializer.checkPMMode().done(function() {
                            setupMessagePump();
                            startTheEngine();
                        });
                    });
                }).fail(function() {
                    logger.error("Directory initalization failed");
                });
            }).fail(function() {
                logger.error("Error whilst initializing system");
            });
        };

        /**
         * [setupSystem description]
         * @return {[type]} [description]
         */
        function setupSystem() {

            var dfd = $.Deferred();

            // Attempt to switch Off OSD - Not critical so continue even if this fails
            SystemInterface.switchOffOSD().always(function() {

                // Call system Inializer and resolve system setup
                self.systemInitializer = new SystemInitializer();
                self.systemInitializer.initialize().done(function() {
                    dfd.resolve();
                }).fail(function() {
                    dfd.reject();
                });

            });

            return dfd;
        }

        /**
         * [setupCoreUI description]
         * @return {[type]} [description]
         */
        function setupCoreUI() {
            // set the dimensions explicitly to the size of the window
            $('body').width(document.documentElement.clientWidth);
            $('body').height(document.documentElement.clientHeight);

            // add the canvas
            $('body').append("<a href='javascript:void(0);' id='remoteFocus'></a><canvas id='layoutBackgroundCanvas' class='viewer' style='position:absolute;'></canvas><img id='layoutBackgroundImage' class='viewer' src='' style='position;absolute;top:0px;left:0px;visibility:hidden;'></img><div id='canvas' style='position: absolute;'></div>");

            $('#canvas').css('left', '0px');
            $('#canvas').css('top', '0px');
            $('#canvas').width(document.documentElement.clientWidth);
            $('#canvas').height(document.documentElement.clientHeight);

            $('#layoutBackgroundCanvas').css('left', '0px');
            $('#layoutBackgroundCanvas').css('top', '0px');
            $('#layoutBackgroundCanvas').attr('width', document.documentElement.clientWidth);
            $('#layoutBackgroundCanvas').attr('height', document.documentElement.clientHeight);

            $('#layoutBackgroundImage').css('left', '0px');
            $('#layoutBackgroundImage').css('top', '0px');
            $('#layoutBackgroundImage').css('width', document.documentElement.clientWidth + 'px');
            $('#layoutBackgroundImage').css('height', document.documentElement.clientHeight + 'px');
            $('#layoutBackgroundImage').css('position', 'relative');

            // add the user message view
            $('#canvas').append("<div id='userMessage' style='position: absolute;top: 50%;width: 100%;height: 200px;margin-top: -50px;'><div id='userMessageText' style='position: absolute;top: 0px;width: 100%;height: 100%;color: #ffffff;font-size: 20px;line-height: 100px;font-weight: bold;text-align: center;vertical-align: middle;z-index: 1;'></div></div>");
        }

        /**
         * [setupDebugTracer description]
         * @return {[type]} [description]
         */
        function setupDebugTracer() {
            // set the logging level
            logger.level = Settings.user.loggingLevel;

            // setup the debug trace output
            if (Settings.user != null | Settings.user != undefined) {
                // enable the trace output
                if (Settings.user.showTraceOutput) {
                    $('body').append("<div id='trace' style='z-index:999;color:#ffffff;position:absolute;width:100%;height:30%;left:0px;top:0px;background-color:rgba(0,0,0,0.8); word-wrap:break-word;font-size:12px'></div>");
                    logger.enableTrace = true;
                } else {
                    logger.enableTrace = false;
                }
            }
        }

        /**
         * [initializeFileSystem description]
         * @return {[type]} [description]
         */
        function initializeFileSystem() {
            var dfd = $.Deferred();

            // create the directories
            $.when(
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.appDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.mediaDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.logDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.widgetDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.dataDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.tempDirectory),
                Directory.createDirectory(Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.screenshotDirectory)
            ).done(function() {
                dfd.resolve();
            }).fail(function() {
                dfd.reject();
            });

            return dfd;
        }

        /**
         * [initializeAppSettings description]
         * @return {[type]} [description]
         */
        function initializeAppSettings() {
            var dfd = $.Deferred();

            $.when(loadClientID(), loadClientKey(), loadUserSettings()).done(function() {
                dfd.resolve();
            }).fail(function() {
                dfd.reject();
            });

            return dfd;
        }

        /**
         * [loadClientID description]
         * @return {[type]} [description]
         */
        function loadClientID() {
           return new ClientIDGenerator().loadClientID();
        }

        /**
         * [loadClientKey description]
         * @return {[type]} [description]
         */
        function loadClientKey() {
            var dfd = $.Deferred();

            if (File.exists(Settings.application.fileSystem.systemFiles.clientKeyFilePath)) {
                // load the client Id
                File.read(Settings.application.fileSystem.systemFiles.clientKeyFilePath).done(function(data) {
                    if (data != null && data != "") {
                        var clientKeyData = JSON.parse(data);
                        logger.info('Stored Client Key is ' + clientKeyData.clientKey);
                        Settings.application.api.clientKey = clientKeyData.clientKey;
                    }

                    dfd.resolve();
                }).fail(function() {
                    dfd.reject();
                });
            } else {
                setTimeout(function() {
                    // if the file does not exist this becuase it has not yet been generated (by registering with the SL cloud)
                    // mark this method as successfull
                    dfd.resolve();
                }, 50);
            }

            return dfd;
        }

        /**
         * [loadUserSettings description]
         * @return {[type]} [description]
         */
        function loadUserSettings() {
            var dfd = $.Deferred();

            if (File.exists(Settings.application.fileSystem.systemFiles.settingsFilePath)) {
                // load the users settings
                File.read(Settings.application.fileSystem.systemFiles.settingsFilePath).done(function(data) {
                    if (data != null && data != "") {
                        Settings.user.settingsLoaded = true;
                        var userSettings = JSON.parse(data);

                        $.extend(true, Settings.user.mySettings, userSettings);
                    }
                    dfd.resolve();
                }).fail(function() {
                    dfd.reject();
                });

            } else {
                setTimeout(function() {
                    // if the file does not exist this becuase it has not yet been generated (by install)
                    // mark this method as successfull
                    dfd.resolve();
                }, 50);
            }

            return dfd;
        }

        /**
         * [setUpCanvas description]
         */
        function setUpCanvas() {

            if (UserConfigurationController.isRotated()) {

                switch (UserConfigurationController.rotationAngle()) {
                    case 90:
                        $('#canvas').width(document.documentElement.clientHeight);
                        $('#canvas').height(document.documentElement.clientWidth);

                        $('#canvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#canvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');

                        $('#canvas').addClass('rotate90');

                        //layoutView = new LayoutView(document.documentElement.clientHeight, document.documentElement.clientWidth);
                        break;
                    case 270:
                        $('#canvas').width(document.documentElement.clientHeight);
                        $('#canvas').height(document.documentElement.clientWidth);

                        $('#canvas').css('left', (document.documentElement.clientWidth - document.documentElement.clientHeight) / 2 + 'px');
                        $('#canvas').css('top', (document.documentElement.clientHeight - document.documentElement.clientWidth) / 2 + 'px');

                        $('#canvas').addClass('rotate270');

                        //layoutView = new LayoutView(document.documentElement.clientHeight, document.documentElement.clientWidth);
                        break;
                }
            } else {

                //layoutView = new LayoutView(document.documentElement.clientWidth, document.documentElement.clientHeight);
            }
        }

        /**
         * [setupMessagePump description]
         * @return {[type]} [description]
         */
        function setupMessagePump() {
            // setup user message pump
            UserMessageController.addEventListener(function(message) {
                logger.debug('User Message: ' + message);
                message = message.replace(/\n/g,'<br/>');
                // If the player supports screen control only show this message if the screen is on
                if (Settings.application.features.canSwitchScreenOnOff) {
                    PowerInterface.getScreenOnOffStatus().done(function(screenOn) {
                        if (screenOn == "on") {
                            $('#userMessageText').html(message);
                        }
                    });
                }
                // If the player does not have screen control then always show the message
                else {
                    $('#userMessageText').html(message);
                }

            });
        }

        /**
         * [startTheEngine description]
         * @return {[type]} [description]
         */
        function startTheEngine() {
            // setup the SL client engine - this class starts all processes
            _engine = new SLClientEngine({});
            _engine.initialize();
        }

    }

    return SLMediaPlayerApp;
});