define(["properties/App.Settings", "system/io/File", "system/io/FileInfo", "models/ConfigurationModel", "models/ObjectProcessingQueueModel", "models/ConnectivityIntervalSchemeModel", "models/MediaTypeConstants", "system/DateTime", "models/SystemReportModel", "controllers/UserMessageController", "models/RssChannelCache", "models/RssChannelModel", "models/LocalFileList", "controllers/UserConfigurationController","core/SLAPIMethodsInterface","models/API"], function(Settings, File, FileInfo, ConfigurationModel, ObjectProcessingQueueModel, ConnectivityIntervalSchemeModel, MediaTypeConstants, DateTime, SystemReportModel, UserMessageController, RssChannelCache, RssChannel, LocalFileList, UserConfigurationController, APIMethodsInterface, API) {
    function SLPlayerDataAPIInterface(config) {
        logger.debug('PLAYER DATA API: Building Player Data API');
        var self = this;

        // call back methods for registration and configuration
        var _configurationCallback;

        // default connection settings
        var _authenticationFailureRetryIntervalMS = 5000;

        var _defaultUploadCheckIntervalMS = 100000;
        // 60 minutes

        var _authenticationTimerId = 0;
        var _uploadTimerId = 0;

        var _popConfig = config;

        var _APIMethodsInterface = new APIMethodsInterface();

        function authenticate() {
            return _APIMethodsInterface.authenticate(API.PLAYER_DATA); 
        }

        function setUploadTimer(forceRetry) {
            // calculate when the next configuration check should occur based on settings in the configuration response.
            var interval = forceRetry === true
                ? 30000
                : calculateNextUploadInterval();
            // reset the configuration timer
            _uploadTimerId = setTimeout(postProofOfPlayLog, interval);
        }

        function calculateNextUploadInterval() {
            var interval = _defaultUploadCheckIntervalMS;

            if (_popConfig != null) {
                interval = _popConfig.uploadfrequency * 1000;
            }

            logger.info('PoP: Next upload in ' + interval + ' ms');

            return interval;
        }

        function isAuthenticationRequired() {
            logger.debug('PLAYER DATA API: isAuthenticationRequired()');

            if (Settings.application.api.authTokenExpiry == null)
                return true;
            if (Settings.application.api.authTokenExpiry < DateTime.now())
                return true;
            return false;
        }

        function postProofOfPlayLog() {
            try {
                logger.debug('PLAYER DATA API: Sending Proof of Play Log');

                // does the client need to authenticate again?
                if (isAuthenticationRequired()) {
                    logger.debug('PLAYER DATA API: Auth Required');
                    // Need to authenticate asynchronously
                    authenticate()
                        .done(function(authenticated) {
                            // Client is not authenticated try again soon
                            if (!authenticated) {
                                // reset the system report timer
                                setUploadTimer(true);
                                return;
                            } else {
                                uploadLogs();
                            }
                        });
                } else {
                    uploadLogs();
                }
            } catch (e) {
                logger.debug('PLAYER DATA API: Proof of Play Upload Failed ' + e.message);
                setUploadTimer();
            }
        }

        function uploadLogs() {
            logger.info("PoP: Starting log upload");
            // Get the file to upload
            var logurl = Settings.application.fileSystem.systemFiles.proofOfPlayFilePath;
            var uploadFileUrl = Settings.application.fileSystem.systemFiles.proofOfPlayUploadFilePath;
            logger.info("PoP: STARTING TO ROLL LOGS");
            File.resetProofOfPlay(logurl, uploadFileUrl)
                .done(function() {
                    logger.info("PoP: FINISHED ROLLING LOGS");
                    File.read(uploadFileUrl, true)
                        .done(function(popReport) {
                            logger.info("PLAYER DATA API: Uploading Proof of Play");
                            //logger.info("Data: " + popReport);

                            if (popReport != null && popReport != '') {
                                var jqXHR = executeAPICall('proofofplay', {
                                    type: 'POST',
                                    data: popReport
                                });

                                jqXHR.done(function(data, textStatus, jqXHR) {
                                    logger.info('PLAYER DATA API: Proof of Play Upload Done ' + jqXHR.status);

                                    try {
                                        // Delete Proof of Play Upload File
                                        File.deleteLocal(uploadFileUrl);
                                        setUploadTimer();
                                    } catch (e) {
                                        logger.error("Error deleting Proof of play upload file: " + e.message);
                                        setUploadTimer();
                                    }
                                }).fail(function(jqXHR, textStatus, errorThrown) {
                                    // log the error
                                    logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                                    if (jqXHR.status === 401) {
                                        logger.warning("PoP Upload failed: 401 Unauthorized. Will reauthenticate and retry");
                                        setTimeout(function() {
                                            Settings.application.api.authTokenExpiry = null;
                                            postProofOfPlayLog();
                                        }, 500);
                                    } else {
                                        setUploadTimer();
                                    }
                                });
                            } else {
                                logger.debug('Proof of Play: No data to upload');
                                setUploadTimer();
                            }
                        }).fail(function() {
                            // log the error
                            logger.debug('Proof of Play: No data to upload');
                            setUploadTimer();

                        });
                }).fail(function() {
                    // log the error
                    logger.error('PoP: Error: Failed to roll PoP Logs');
                    setUploadTimer();
                });
        }

        function executeAPICall(operationName, settings) 
        {
            return _APIMethodsInterface.executeAPICall(API.PLAYER_DATA, operationName,settings);
        }

        function logAPIError(requestUrl, requestData, responseStatus, responseMessage, responseText) {
            var logData = {
                url: requestUrl,
                requestData: requestData,
                responseStatus: responseStatus,
                responseMessage: responseMessage,
                responseBody: responseText
            };

            logger.error("PLAYER DATA API: API Error " + JSON.stringify(logData));
        }

        self.initialize = function() {
            logger.info('PLAYER DATA API: Initializing SLPlayerDataAPIInterface, API Creds: ' + JSON.stringify(Settings.application.playerdataapi));

            if (Settings.application.api != null && Settings.application.api.appId != null && Settings.application.api.appKey != null && Settings.application.api.clientId != null && Settings.application.api.clientKey != null) {
                // Start the upload process in 20 seconds (allow time for internet connection to be established)
                setTimeout(function() {
                    postProofOfPlayLog();
                }, 20000);
            }
        };
    }

    return SLPlayerDataAPIInterface;
});