define(["properties/App.Settings","core/SLAPIMethodsInterface","models/API"], function(Settings, APIMethodsInterface, API) {
    /*
    	SLScreenshotUploadAPIInterface - Provides an interface to communicate with the SL Screenshot Upload API
    	------------------------------------------------------------------------------
    	This module has been updated so that the authentication is now async rather than synchronous
     */
    function SLScreenshotUploadAPIInterface() {
        var self = this;

        // default connection settings
        var _authenticationFailureRetryIntervalMS = 5000;

        var _UploadComplete;
        var _UploadFailed;

        var _HasUploadFailed = false;
        var _HasUploadStarted = false;
        var _HasUploadFinished = false;

        var _APIMethodsInterface = new APIMethodsInterface();

        //
        // authenticate()
        //
        function authenticate() {

            return _APIMethodsInterface.authenticate(API.SCREENSHOT_UPLOAD); 
        }


        function notify(data) {
            var dfd = $.Deferred();

            logger.info('Notify ' + data.type + ', Object Id: ' + data.object_id);

            // does the client need to authenticate again?
            if (isAuthenticationRequired()) {

                authenticate().done(function(authenticated) {

                    if (!authenticated) {
                        // Not authed - cant send notification
                        dfd.reject();
                    } else {
                        // Now authenticated do send notification
                        sendNotification(data).done(function() {
                            dfd.resolve();
                        }).fail(function() {
                            dfd.reject();
                        });
                    }
                });
            } else {
                // Already authenticated - send the notification request
                sendNotification(data).done(function() {
                    dfd.resolve();
                }).fail(function() {
                    dfd.reject();
                });
            }

            return dfd;
        }

        function sendNotification(data) {
            var dfd = $.Deferred();

            var jqXHR = executeAPICall('notifications', {
                type: 'POST',
                data: data
            });

            jqXHR.done(function(data, textStatus, jqXHR) {
                logger.info('Notify Response Code ' + jqXHR.status);
                dfd.resolve();
            }).fail(function(jqXHR, textStatus, errorThrown) {
                // reset auth token if it is not valid
                if (jqXHR.status == 401) {
                    Settings.application.api.authTokenExpiry = null;
                }

                // log the error
                logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                dfd.reject();
            });

            return dfd;
        }

        function executeAPICall(operationName, settings) {
            return _APIMethodsInterface.executeAPICall(API.SCREENSHOT_UPLOAD,operationName,settings);
        }

        self.uploadMultipleScreenshots = function(screenshotArray) {
            uploadScreenshotAPICall(screenshotArray).done(function(data, textStatus, jqXHR) {
                _UploadComplete();
            }).fail(function(jqXHR, textStatus, errorThrown) {
                // reset auth token if it is not valid
                if (jqXHR.status == 401) {
                    Settings.application.api.authTokenExpiry = null;
                }
                // log the error
                // logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                _UploadFailed();
            });
        };

        self.uploadSingleScreenshot = function(screenshotUploadObject) {
            var screenshotArray = [screenshotUploadObject];
            uploadScreenshotAPICall(screenshotArray).done(function(data, textStatus, jqXHR) {
                _UploadComplete();
            }).fail(function(jqXHR, textStatus, errorThrown) {
                // reset auth token if it is not valid
                if (jqXHR.status == 401) {
                    Settings.application.api.authTokenExpiry = null;
                }
                // log the error
                //   logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
                _UploadFailed();
            });
        };

        function uploadScreenshotAPICall(screenshotArray) {
            try {
                logger.info('PLAYER API - Client ID: ' + Settings.application.api.clientId + ', Token: ' + Settings.application.api.authToken);
                var screenshotUploadRequestBody = {};
                screenshotUploadRequestBody.screenshots = screenshotArray;
                return executeAPICall('Upload', {
                    type: 'POST',
                    dataType: 'text',
                    contentType: 'application/json',
                    data: screenshotUploadRequestBody,
                    timeout: 0
                });
            } catch (e) {
                // logger.error('Error executing api call ' + e.message);
                return null;
            }
        }

        function isAuthenticationRequired() {
            logger.debug('isAuthenticationRequired()');
            if (Settings.application.api.authTokenExpiry == null)
                return true;
            if (Settings.application.api.authTokenExpiry < DateTime.now())
                return true;
            return false;
        }

        function logAPIError(requestUrl, requestData, responseStatus, responseMessage, responseText) {
            var logData = {
                url: requestUrl,
                requestData: requestData,
                responseStatus: responseStatus,
                responseMessage: responseMessage,
                responseBody: responseText
            };

            logger.error("API Error " + JSON.stringify(logData));
        }


        self.initialize = function(onUploadComplete, onUploadFailed) {
            logger.info('Initializing SLScreenshotUploadAPIInterface');

            _UploadComplete = onUploadComplete;
            _UploadFailed = onUploadFailed;
        };

    }

    return SLScreenshotUploadAPIInterface;
})