define(["properties/App.Settings", "system/DateTime", "controllers/UserConfigurationController", "models/PusherSubscriptionModel","core/SLAPIMethodsInterface","models/API"], function(Settings, DateTime, UserConfigurationController, PusherSubscriptionModel, APIMethodsInterface, API)
{
	/**
	 * This modules checks the web-based interrupts API for configuration updates.
	 * The configuration includes the pusher key and details of the channels the player should
	 * listen for updates against.
	 */
	function SLWebBasedInterruptsAPIInterface()
	{
		var self = this;

		var _options = null;
		var _lastModifiedDate = null;
		var _lastConfiguration = null;
		var _configurationCheckTimer = null;

		var _APIMethodsInterface = new APIMethodsInterface();

		/**
		 * Initializes the The API Interface. Starts the configuration check process
		 * @param  {Object} options any initialization options
		 * @return {Jquery Promise} Resolved on success or rejected on failure
		 */
		self.initialize = function(lastConfiguration, options)
		{
			logger.debug("Initializing SLWebBasedInterruptsAPIInterface");
			var dfd = $.Deferred();
			
			setTimeout(function()
			{
				// Initialize basic settings
				if (lastConfiguration != null) {
					_lastConfiguration = lastConfiguration;
					_lastModifiedDate = lastConfiguration.lastModifiedDate;
					
				}
				_options = options;

				// API started so resolve initialisation
				dfd.resolve();

			}, 50);

			return dfd;
		};

		self.stop = function()
		{
			_lastModifiedDate = null;
			_lastConfiguration = null;
		}

		/**
		 * Checks for a new configuration
		 * @return {[type]} [description]
		 */
		self.checkForConfigurationUpdate = function()
		{

			getConfigurationLastModifiedDate()
			.done(function(lastModifiedDate) {

				// Do we need to perform an update?
				if (_lastModifiedDate == null || lastModifiedDate.getTime() != _lastModifiedDate.getTime())
				{
					getConfiguration()
					.done(function(pusherConfiguration) {

						// Update last modified date and last configuration
						_lastConfiguration = pusherConfiguration;
						_lastModifiedDate = pusherConfiguration.lastModifiedDate;

						// Store updated pusher settings locally
						//Settings.application.webBasedInterrupts.pusherSettings = pusherConfiguration;n;
						//Settings.application.webBasedInterrupts.pusherSettings = _lastModifiedDate;

						// Do callback
						_options.configurationCallback(pusherConfiguration);

					})
					.fail(function(){
						logger.error("SL Web Based Interrupts API Interface: Error performing configuration update.");

					});
				}
				else
				{
					// No update available - schedule next check
					
				}
			})
			.fail(function() {
				logger.error("SL Web Based Interrupts API Interface: Error checking for configuration update.");
			});
		};

		/**
		 * Calls HEAD /Configuration on the API and returns the Last-Modified Header
		 * @param  {number} Player serial number
		 * @return {Jquery  Promise(DateTime)} The last modified date of the web based interrupt configuration
		 */
		function getConfigurationLastModifiedDate() {
			var dfd = $.Deferred();

			// Do the call
			var apiCall = executeAPICall('playerconfiguration', { type : "HEAD" });

			// On Success
			apiCall.done(function(data, textStatus, jqXHR) {
				var lastModifiedDate = new Date(jqXHR.getResponseHeader('Last-Modified'));
				dfd.resolve(lastModifiedDate);
			});

			// On Failure
			apiCall.fail(function(jqXHR, textStatus, errorThrown) {
				// log the error
				logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
				dfd.reject();
			});

			return dfd;
		}

		/**
		 * Gets the Web-based Interrupt Configuration. Returns the configuration object.
		 * @param  {number} Player Serial Number
		 * @return {Jquery Promise(PusherSubscrptionModel)}
		 */
		function getConfiguration()
		{
			var dfd = $.Deferred();

			// Do the api call
			var apiCall = executeAPICall('playerconfiguration');
			
			// On Success
			apiCall.done(function(data, textStatus, jqXHR) {
				var lastModifiedDate = new Date(jqXHR.getResponseHeader('Last-Modified'));
				data.lastModifiedDate = lastModifiedDate;
				dfd.resolve(new PusherSubscriptionModel(data));
			});
			
			// On Failure
			apiCall.fail(function(jqXHR, textStatus, errorThrown) {
				// log the error
				logAPIError(this.url, this.data, jqXHR.status, errorThrown, jqXHR.responseText);
				dfd.reject();
			});

			return dfd;
		}


		/**
		 * Executes API call using provided operation name. Optionally takes a settings object that overrides default options.
		 * @param  {String} operationName [description]
		 * @param  {number} serialNo	  [description]
		 * @param  {Object} settings      [description]
		 * @return {[type]}               [description]
		 */
		function executeAPICall(operationName, settings) {
			return _APIMethodsInterface.executeAPICall(API.WEB_TRIGGERS,operationName + "/" + _options.serialNo ,settings);
		}

		/**
		 * Logs as much detail as possible about an API Error
		 * @param  {[type]} requestUrl      [description]
		 * @param  {[type]} requestData     [description]
		 * @param  {[type]} responseStatus  [description]
		 * @param  {[type]} responseMessage [description]
		 * @param  {[type]} responseText    [description]
		 * @return {[type]}                 [description]
		 */
		function logAPIError(requestUrl, requestData, responseStatus, responseMessage, responseText) {
			var logData = {
				url : requestUrl,
				requestData : requestData,
				responseStatus : responseStatus,
				responseMessage : responseMessage,
				responseBody : responseText
			};

			logger.error("SL Web-Based Interrupts API: " + JSON.stringify(logData));
		}

		
	}

	return SLWebBasedInterruptsAPIInterface;
});

