define(["properties/App.Settings", "gui/PlayUrlHelper", "gui/viewmodels/LayoutViewModel", "gui/views/MediaZoneView", "gui/views/TextZoneView", "system/io/File", "controllers/UserConfigurationController", "models/PlaybackStartEndReasons", "controllers/WidgetEventController"], function(Settings, PlayUrlHelper, LayoutViewModel, MediaZoneView, TextZoneView, File, UserConfigurationController, PlaybackStartEndReasons, WidgetEventController) {

    function LayoutView(screenWidth, screenHeight) {
        var self = this;
        self.zoneViews = [];
        var _isFullscreen = true;
        var _backgroundImage = null;
        var _backgroundFillColour = "#000000";
        var _mainZone = null;
        var _initialized = false;
        var _syncConfig = null;
        var _clientDetails = null;

        self.setLayout = function(layout, playReason, triggerAdditionalData, syncConfig, clientDetails) {
            if (_initialized == false) {
                logger.debug('Adding event listeners to User Configuration Controller');
                // add user configuration change event listener
                UserConfigurationController.addEventListener(function(property, data) {
                    switch (property) {
                        case "mute":
                            $(self.zoneViews).each(function(index, zone) {
                                if (zone.viewModel.type == 'media')
                                    zone.setMute(data);
                            });
                            break;
                        case "volumeUp":
                            $(self.zoneViews).each(function(index, zone) {
                                if (zone.viewModel.type == 'media')
                                    zone.volumeUp();
                            });
                            break;
                        case "volumeDown":
                            $(self.zoneViews).each(function(index, zone) {
                                if (zone.viewModel.type == 'media')
                                    zone.volumeDown();
                            });
                            break;
                    }
                });
                _initialized = true;
            }
            _syncConfig = syncConfig;
            _clientDetails = clientDetails;
            // stop the current layout playing. Reason for stopping the previous layout is the play reason for this layout
            stop(playReason);
            // Initialize Widget Event Controller
            WidgetEventController.initialize();
            // enable video playback
            Settings.application.criticalSections.videoAvailable = true;
            self.layout = new LayoutViewModel(layout, screenWidth, screenHeight);
            _isFullscreen = (self.layout.zones.length == 1 && self.layout.zones[0].width == self.layout.width && self.layout.zones[0].height == self.layout.height);
            var layoutBackgroundCanvas = document.getElementById('layoutBackgroundCanvas');
            var backgroundCanvasContext = layoutBackgroundCanvas.getContext("2d");
            backgroundCanvasContext.globalAlpha = 1;
            // reset the stored background image
            _backgroundImage = null;
            // add the background
            if (layout.backgroundAsset != null) {
                logger.debug('Loading background asset ' + layout.backgroundAsset.name);

                PlayUrlHelper.getMediaAssetPlayUrl(layout.backgroundAsset).done(function(url) {
                    _backgroundImage = new Image();
                    _backgroundImage.addEventListener('load', function(eventData) {
                        logger.debug('Backgroud image loaded - draw it');

                        try {

                            // On WebOS we use the #layoutBackgroundImage <img> element as drawCanvas is just resulting in a black background being draw?!!!
                            logger.debug('** Layout background image loaded');
                            $('#layoutBackgroundCanvas').css('visibility', 'hidden');
                            $('#layoutBackgroundImage').attr('src', url);
                            $('#layoutBackgroundImage').css('visibility', 'visible');

                            // Load Zones
                            loadZones(playReason, triggerAdditionalData);

                        } catch (e) {
                            logger.error('Error whilst drawing background: ' + e.message);
                        }
                    }, false);

                    _backgroundImage.onerror = function() {
                        logger.error('Error loading background image');
                    };

                    logger.debug('Loading background image file ' + url);
                    _backgroundImage.src = url;
                });

            } else {

                if (_isFullscreen) {
                    // set to black
                    _backgroundFillColour = "#000000";

                } else {
                    // set to white
                    _backgroundFillColour = "#ffffff";
                }

                logger.debug('** No background image. Painting background canvas: ' + _backgroundFillColour);
                $('#layoutBackgroundCanvas').css('visibility', 'visible');
                backgroundCanvasContext.fillStyle = _backgroundFillColour;
                backgroundCanvasContext.fillRect(0, 0, self.layout.width, self.layout.height);

                $('#layoutBackgroundImage').attr('src', '');
                $('#layoutBackgroundImage').css('visibility', 'hidden');

                loadZones(playReason, triggerAdditionalData, layout);
            }

        };

        /**
         * Loads each media zone in the layout view
         * @param {PlaybackStartEndReasons} The reason for playback starting
         * @return {[type]} [description]
         */
        function loadZones(playReason, triggerAdditionalData, layout) {
            // add the zones to the presentation
            $(self.layout.zones).each(function(index, zoneViewModel) {
                var zoneView = null;
                switch (zoneViewModel.type) {
                    case "media":
                        zoneView = new MediaZoneView(zoneViewModel, playReason, triggerAdditionalData, layout);
                        zoneView.setMute(UserConfigurationController.isMuted());
                        zoneView.videoStart = function(zone) {
                            try {

                                // Cut out canvas - only required if canvas is in front of the layout
                                if (!Settings.application.features.canvasBehindLayout) {
                                    var layoutBackgroundCanvas = document.getElementById('layoutBackgroundCanvas');
                                    var backgroundCanvasContext = layoutBackgroundCanvas.getContext("2d");
                                    backgroundCanvasContext.globalAlpha = 0;
                                    backgroundCanvasContext.clearRect(zone.x, zone.y, zone.width, zone.height);
                                }
                            } catch (e) {
                                logger.error('Error whilst clearing background rectangle for video to play: ' + e.message);
                            }
                        };
                        zoneView.videoComplete = function(zone) {
                            try {

                                // Redraw canvas - only required if canvas is in front of the layout
                                if (!Settings.application.features.canvasBehindLayout) {
                                    var layoutBackgroundCanvas = document.getElementById('layoutBackgroundCanvas');
                                    var backgroundCanvasContext = layoutBackgroundCanvas.getContext("2d");
                                    backgroundCanvasContext.globalAlpha = 1;

                                    if (_backgroundImage != null) {
                                        var bgImageToScreenXScalar = _backgroundImage.width / screenWidth;
                                        var bgImageToScreenYScalar = _backgroundImage.height / screenHeight;

                                        // ensure that the image is not clipped outside of it bounds
                                        var clippingX = Math.max(bgImageToScreenXScalar * zone.x, 0);
                                        var clippingY = Math.max(bgImageToScreenYScalar * zone.y, 0);
                                        var clippingWidth = Math.min(bgImageToScreenXScalar * zone.width, _backgroundImage.width);
                                        var clippingHeight = Math.min(bgImageToScreenYScalar * zone.height, _backgroundImage.height);

                                        var placementX = zone.x;
                                        var placementY = zone.y;
                                        var placementWidth = zone.width;
                                        var placementHeight = zone.height;

                                        // TODO: This may need to be changed to use drawBackgroundImage() when the SSSP code is merged in
                                        backgroundCanvasContext.drawImage(_backgroundImage, clippingX, clippingY, clippingWidth, clippingHeight, placementX, placementY, placementWidth, placementHeight);

                                    } else {
                                        backgroundCanvasContext.fillStyle = _backgroundFillColour;
                                        backgroundCanvasContext.fillRect(zone.x, zone.y, zone.width, zone.height);
                                    }
                                }
                            } catch (e) {
                                logger.error('Error whilst re-drawing background ' + e.message);
                            }
                        };
                        break;
                    case "text":
                        zoneView = new TextZoneView(zoneViewModel);
                        break;
                }

                if (zoneView != null) {
                    // add to the collection of layers
                    self.zoneViews.push(zoneView);
                }
            });

            _mainZone = null;
            // find the 'main' zone - the 'main' zone is the largest zone, or the one closest to the origin
            $(self.zoneViews).each(function(index, zone) {
                if (zone.viewModel.type == 'media') {
                    if (_mainZone == null) {
                        logger.debug('First Zone selected');
                        _mainZone = zone;
                    } else {
                        // size of main zone
                        var mainZoneSize = _mainZone.width * _mainZone.height;

                        logger.debug('Main Zone Size: ' + mainZoneSize);

                        // size of the current zone
                        var zoneSize = zone.width * zone.height;

                        // select the biggest zone
                        if (zoneSize > mainZoneSize) {
                            logger.debug('1 Zone selected at position ' + zone.x);
                            _mainZone = zone;
                        }

                        // if the zones are the same size - select the one closest to the origin
                        if (zoneSize == mainZoneSize) {
                            var mainZoneDistanceFromOrigin = Math.sqrt((_mainZone.x * _mainZone.x) + (_mainZone.y * _mainZone.y));
                            var zoneDistanceFromOrigin = Math.sqrt((zone.x * zone.x) + (zone.y * zone.y));

                            if (zoneDistanceFromOrigin < mainZoneDistanceFromOrigin) {
                                logger.debug('2 Zone selected at position ' + zone.x);
                                _mainZone = zone;
                            }

                            // if the origins are the same select the one highest in the z-order
                            if (zoneDistanceFromOrigin == mainZoneDistanceFromOrigin) {
                                logger.debug('3 Zone selected at position ' + zone.x);
                                _mainZone = zone;
                            }
                        }
                    }
                }
            });

            //Enable sync on the main zone
            if (_syncConfig.enabled) {
                //Sync
                _mainZone.enableSync(_syncConfig, _clientDetails);
            }

            // play the layout once the background has been drawn
            setTimeout(function() {
                play();
            }, 250);
        }

        // due to the fact that the video is below the HTML layer, to display video
        // through the background each media zone must have a transparent section cut out
        function addZonesTransparency() {
            var layoutBackgroundCanvas = document.getElementById('layoutBackgroundCanvas');
            var backgroundCanvasContext = layoutBackgroundCanvas.getContext("2d");

            // add the zones to the presentation
            $(self.layout.zones).each(function(index, zoneViewModel) {
                switch (zoneViewModel.type) {
                    case "media":
                        // make the background transparent - so that videos are displayed as they are at the bottom of the z-order
                        if (backgroundCanvasContext != null) {
                            logger.debug('Drawing transparent section');
                            backgroundCanvasContext.globalAlpha = 0;
                            backgroundCanvasContext.clearRect(zoneViewModel.x, zoneViewModel.y, zoneViewModel.width, zoneViewModel.height);
                        }
                        break;
                }
            });
        }

        /**
         * Startings playing each media zone in the layout
         * @return {[type]} [description]
         */
        function play() {
            $(self.zoneViews).each(function(index, zone) {
                setTimeout(function() {
                    zone.play();
                }, 10);
            });
        }

        /**
         * Stops all media zones in the layout view
         * @param {PlaybackStartEndReasons} The reason for stopping playback
         * @return {[type]} [description]
         */
        function stop(endReason) {
            logger.debug('Stopping Layout View');

            // is there a layout already playing?
            if (self.layout != null & self.layout != undefined) {
                // stop each zone
                $(self.zoneViews).each(function(i, zoneView) {
                    zoneView.stop(endReason);
                });

                // remove all of the zone views from this layout view
                self.zoneViews = [];

                // Hide background image
                $('#layoutBackgroundImage').attr('src', '');
                $('#layoutBackgroundImage').css('visibility', 'hidden');

                // Playback has stopped so set canvas to black
                // layoutBackgroundCanvas = document.getElementById('layoutBackgroundCanvas');
                // backgroundCanvasContext = layoutBackgroundCanvas.getContext("2d");
                // backgroundCanvasContext.fillStyle = '#000000';
                // backgroundCanvasContext.fillRect(0, 0, self.layout.width, self.layout.height);

                // Stop WidgetEventController
                WidgetEventController.stop();

                // empty the canvas ready for the next layout
                $('#canvas').empty();
                $("div[id$='_videoViewer_container']").remove();
                $("div[id$='_tvViewer_container']").remove();
            }
        }

        /**
         * Stops playback of the current layout
         * @return {[type]} [description]
         */
        self.stop = function(endReason) {
            stop(endReason);
        };

        /**
         * Skip to the next asset in the main zone
         * @return {[type]} [description]
         */
        self.skip = function() {
            logger.debug('Layout - skip');

            _mainZone.skip();
        };

        /**
         * Go back to the previous asset in the main zone
         * @return {[type]} [description]
         */
        self.back = function() {
            logger.debug('Layout - back');

            _mainZone.back();
        };

        self.onKeyPress = function(interruptKey) {
            $(self.zoneViews).each(function(index, zone) {
                setTimeout(function() {
                    zone.onKeyPress(interruptKey);
                }, 10);
            });
        };
    }

    return LayoutView;
});