define(["system/DateTime", "gui/views/LayoutView", "models/PlaybackStartEndReasons", "system/power/PowerInterface"], function(DateTime, LayoutView, PlaybackStartEndReasons, PowerInterface) {
	/**
	 * MediaPlayerView
	 * --------------------
	 * Exposes the Media Player to the rest of the application
	 * @param {Object} Function callbacks
	 */
	function MediaPlayerView(callbacks) {
		var self = this;
		var _callbacks = callbacks;
		var _layoutView = new LayoutView(document.documentElement.clientWidth, document.documentElement.clientHeight);
		var _scheduleTimerId = 0;
		var _content = null;
		var _syncConfig = null;
		var _clientDetails = null;

		/**
		 * Takes a content model and starts playing the currently scheduled content.
		 * @param  {[type]} content [description]
		 * @return {[type]}         [description]
		 */
		self.loadContent = function(content, syncConfig, clientDetails) {
			_content = content;	
			_syncConfig = syncConfig;
			_clientDetails = clientDetails;
			setCurrentSchedule();
		};

		/**
		 * Takes a layout model representing the interrupt to play
		 * @param  {[type]} layout [description]
		 * @param {[type]} triggerAdditionalData [description]
		 * @return {[type]}        [description]
		 */
		self.playInterrupt = function(layout, triggerAdditionalData) {
			if (PowerInterface.screenOn === false) {
                logger.warning("playInterrupt() called but screen is off. Ignoring request");
                return;
            }
			_layoutView.setLayout(layout, PlaybackStartEndReasons.INTERRUPT, triggerAdditionalData, _syncConfig, _clientDetails);
		};

		/**
		 * Plays the current scheduled layout based on the existing content model.
		 * @return {[type]} [description]
		 */
		self.playScheduledContent = function() {
			setCurrentSchedule();
		};

		/**
		 * Adjust the dimensions of the currently playing layout view
		 * @param  {number} width  
		 * @param  {number} height 
		 * @return {[type]}        [description]
		 */
		self.adjustLayoutViewDimensions = function(width, height) {
			_layoutView.stop(PlaybackStartEndReasons.ADJUST_LAYOUT_DIMENSIONS);
			_layoutView = new LayoutView(width, height);
		};

		/**
		 * Play all media zones in the layout
		 * @return {[type]} [description]
		 */
		self.play = function() {
			_layoutView.play();
		};

		/**
		 * Pause the main media zone
		 * @return {[type]} [description]
		 */
		self.pause = function() {
			// TODO: Implement a pause?
			logger.warning("Media Player View: Pause called but not implemented.");
		};

		/**
		 * Stops playback of the layout and all zones within in it
		 * @param  {[type]} reason [description]
		 * @return {[type]}        [description]
		 */
		self.stop = function(reason) {
			_layoutView.stop(reason);
		};

		/**
		 * Skips to the next asset in the mainz one in the current layout
		 * @return {[type]} [description]
		 */
		self.skip = function() {
			_layoutView.skip();
		};

		/**
		 * Plays the previous asset in the main zone in the current layout
		 * @return {[type]} [description]
		 */
		self.back = function() {
			_layoutView.back();
		};

		/**
		 * Rotates the layout view - not implemented
		 * @return {[type]} [description]
		 */
		self.rotate = function() {
			// TODO: Implement rotation using this method?
			logger.warning("Media Player View: Rotate called but not implemented.");
		};

		/**
		 * From the content module it iterates over
		 * the schedules and determines the scheduled
		 * layout that should be playing now.
		 */
		function setCurrentSchedule() {

			clearTimeout(_scheduleTimerId);

			if (PowerInterface.screenOn === false) {
                logger.warning("setCurrentSchedule() called but screen is off. Ignoring request");
                return;
            }

			// find the current schedule
			var currentSchedule = null;
			$(_content.schedules).each(function(i, schedule) {
				try {
					if (schedule.start <= DateTime.now() && schedule.end >= DateTime.now()) {
						currentSchedule = schedule;
					}
				} catch(e) {

				}
			});

			if (currentSchedule != null) {
				try {
					$('#userMessage').css('visibility', 'hidden');

					// load the current scheduled layout
					_layoutView.setLayout(currentSchedule.layout, PlaybackStartEndReasons.SCHEDULE,null, _syncConfig, _clientDetails);

					// Add the settings selection square - Not Used
					//addSettingsSelectionSquare();

				} catch(e) {
					logger.error(e.message);
				}
			} else {
				$('#userMessage').css('visibility', 'visible');

				//UserMessageController.postMessage('A valid schedule could not be found please check the date time is correct');
			}

			// give focus to the remote focus anchor to capture events from the remote control
			$('#remoteFocus').focus();

			// calculate when the next layout should play
			var timespanMS = currentSchedule.end.subtract(DateTime.now());

			// set the timer to play the schedule at the appropriate time
			_scheduleTimerId = setTimeout(setCurrentSchedule, timespanMS);
		}

		self.onKeyPress = function(interruptKey)
		{
			_layoutView.onKeyPress(interruptKey);	
		};

		// Used with NEC ALP Soft Triggers and LocalAPIController (Local Triggers API)
		self.softTrigger = function(softTriggerConfig) {
            // soft trigger
            if (softTriggerConfig != null)
                _layoutView.softTrigger(softTriggerConfig);
        };
	}

	return MediaPlayerView;
});
