define(["system/DateTime", "system/io/File", "system/io/Directory", "gui/PlayUrlHelper", "gui/views/LayerBase", "models/MediaTypeConstants", "models/PlayDirectionConstants", "models/PlayStatusConstants", "properties/App.Settings", "gui/controls/VideoPlayer", "gui/controls/WidgetViewer", "gui/controls/TVPlayer", "controllers/UserConfigurationController", "models/PlaybackStartEndReasons", "system/sensors/NFCInterface", "models/PlaylistMediaAssetActionType", "controllers/MediaSyncController", "system/network/NetworkInterface"], function(DateTime, File, Directory, PlayUrlHelper, LayerBase, MediaTypeConstants, PlayDirection, PlayStatus, Settings, VideoPlayer, WidgetViewer, TVPlayer, UserConfig, PlaybackStartEndReasons, NFCInterface, PlaylistMediaAssetActionType, MediaSyncController, NetworkInterface) {
    /**
     * MediaZoneView
     * @param {ZoneViewModel} The MediaZoneView model
     * @param {PlaybackStartEndReasons} The reason that playback has started (Schedule/Interrupt)
     */
    function MediaZoneView(zoneViewModel, playReason, triggerAdditionalData, layout) {

        var _MediaSyncController = null;
        this.parent = new LayerBase(zoneViewModel);

        for (var property in this.parent) {

            if (property != 'prototype' & property != 'parent')
                this[property] = this.parent[property];
        }

        var self = this;
        self.viewModel = zoneViewModel;
        self.imageViewerIdSuffix = '_imgViewer';
        self.svgViewerIdSuffix = '_svgViewer';
        self.activeWebViewerIdSuffix = '_webViewer0';
        self.inactiveWebViewerIdSuffix = '_webViewer1';
        self.layout = layout;

        self.playReason = playReason;
        self.triggerAdditionalData = triggerAdditionalData;

        self.syncEnabled = false;
        self.isSyncMaster = false;
        self.syncMasterFailed = false;
        self.syncConfig = null;
        self.assetCompleteDone = false;
        self.currentAssetIsVideo = false;
        self.syncFailedTimerId = null;
        self.syncFailWaitTime = 10;
        self.secondsToWaitForMasterOnStartup = 45;

        // Polyfil for String.includes() (used by widget URL builder method)
        if (!String.prototype.includes) {
            String.prototype.includes = function(search, start) {
                'use strict';
                if (typeof start !== 'number') {
                    start = 0;
                }

                if (start + search.length > this.length) {
                    return false;
                } else {
                    return this.indexOf(search, start) !== -1;
                }
            };
        }

        // setup the Video plugin
        if (UserConfig.isRotated()) {

            // If rotation set then setup video and TV elements

            // 90 Degrees
            if (UserConfig.rotationAngle() == 90) {

                // Calculate position for 90 Degrees
                var x = document.documentElement.clientWidth - self.y - self.height;
                var y = self.x;

                // switch width and height?
                var modifiedHeight = self.width;
                var modifiedWidth = self.height;

                // Create video and TV Players
                self.videoViewer = new VideoPlayer(self.id, x, y, modifiedWidth, modifiedHeight, self.viewModel.zIndex, {
                    complete: onVideoComplete,
                    fail: onVideoFail,
                    playing: onVideoPlaying,
                    gaplessComplete: onGaplessVideoComplete,
                    gaplessFail: onGaplessVideoFail
                });

                self.tvViewer = new TVPlayer(self.id, x, y, modifiedWidth, modifiedHeight);
            }

            // 180 Degrees - same as if there is no rotation // but inverse x/y co-ordinates
            if (UserConfig.rotationAngle() == 180) {

                var x = document.documentElement.clientWidth - self.x - self.width;
                var y = document.documentElement.clientHeight - self.y - self.height;

                // Create video and TV Players
                self.videoViewer = new VideoPlayer(self.id, x, y, self.width, self.height, self.viewModel.zIndex, {
                    complete: onVideoComplete,
                    fail: onVideoFail,
                    playing: onVideoPlaying,
                    gaplessComplete: onGaplessVideoComplete,
                    gaplessFail: onGaplessVideoFail
                });

                self.tvViewer = new TVPlayer(self.id, x, y, self.width, self.height);
            }

            if (UserConfig.rotationAngle() == 270) {

                // Calculate position for 90 Degrees
                var x = self.y;
                var y = document.documentElement.clientHeight - self.x - self.width;

                // switch width and height?
                var modifiedHeight = self.width;
                var modifiedWidth = self.height;

                // Create video and TV Players
                self.videoViewer = new VideoPlayer(self.id, x, y, modifiedWidth, modifiedHeight, self.viewModel.zIndex, {
                    complete: onVideoComplete,
                    fail: onVideoFail,
                    playing: onVideoPlaying,
                    gaplessComplete: onGaplessVideoComplete,
                    gaplessFail: onGaplessVideoFail
                });

                self.tvViewer = new TVPlayer(self.id, x, y, modifiedWidth, modifiedHeight);
            }

            // No Rotation so position of video element is the same as the zone
        } else {

            // Create video and TV Players
            self.videoViewer = new VideoPlayer(self.id, self.x, self.y, self.width, self.height, self.viewModel.zIndex, {
                complete: onVideoComplete,
                fail: onVideoFail,
                playing: onVideoPlaying,
                gaplessComplete: onGaplessVideoComplete,
                gaplessFail: onGaplessVideoFail
            });

            self.tvViewer = new TVPlayer(self.id, self.x, self.y, self.width, self.height);
        }

        self.assetTimerId = 0;
        self.playlistTimerId = 0;
        self.webZoneLoadTimer = 0;

        var _currentStatus = PlayStatus.STOPPED;
        var _direction = PlayDirection.FORWARDS;

        self.currentPlaylist = null;
        self.currentAssetIndex = -1;
        self.currentScheduledPlaylistMediaAsset = null;
        self.previousScheduledPlaylistMediaAsset = null;

        self.currentAssetStartTime = DateTime.now();

        // Gapless
        self.gaplessPlaybackFailed = false;
        self.gaplessFileLocalFilePath = null;
        self.gaplessFilePlaybackPath = null;

        // setup the zone
        // add a cached element for images
        $('#' + this.id).append("<img class='viewer' id='" + self.id + self.imageViewerIdSuffix + "' src='' style='position:absolute;left:0px;top:0px;width:" + this.width + "px;height:" + this.height + "px;visibility:hidden; z-index:" + self.viewModel.zIndex + ";'></img>");
        //$('#' + this.id).append("<object type='image/svg+xml' class='viewer' id='" + self.id + self.svgViewerIdSuffix + "' data='' style='position:absolute;left:0px;top:0px;width:" + this.width + "px;height:" + this.height + "px;visibility:hidden; z-index:" + self.viewModel.zIndex + ";'></object>");

        // add two cached elements for web pages
        $('#' + this.id).append("<iframe class='viewer' id='" + this.id + self.activeWebViewerIdSuffix + "' frameborder='0' scrollbars='no' src='' style='position:absolute;left:0px;top:0px;width:" + this.width + "px;height:" + this.height + "px;visibility:hidden; z-index:" + self.viewModel.zIndex + ";'></iframe>");
        $('#' + this.id).append("<iframe class='viewer' id='" + this.id + self.inactiveWebViewerIdSuffix + "' frameborder='0' scrollbars='no' src='' style='position:absolute;left:0px;top:0px;width:" + this.width + "px;height:" + this.height + "px;visibility:hidden; z-index:" + self.viewModel.zIndex + ";'></iframe>");

        // Do not position video within the DOM as we do not want it to get rotated!
        $("body").append(self.videoViewer.html);
        self.videoViewer.addedToCanvas();

        // event handlers for starting and stopping video
        self.videoStart = null;
        self.videoComplete = null;

        self.videoViewerContainer = $('#' + self.id + '_videoViewer_container');
        self.tvViewerContainer = $('#' + self.id + '_tvViewer_container');

        // WidgetViewer IFrame
        try {
            self.widgetViewer = new WidgetViewer(self.id, self.x, self.y, self.width, self.height, {
                ready: onWidgetReady,
                moveNext: onWidgetMoveNextRequested
            });
            self.widgetViewer.appendToCanvas();
        } catch (e) {
            logger.error(e);
        }

        function zoneScreenElementSelector() {
            return '#' + self.id;
        }

        function imageViewerElementSelector() {
            return '#' + self.id + self.imageViewerIdSuffix;
        }

        function svgViewerElementSelector() {
            return '#' + self.id + self.svgViewerIdSuffix;
        }

        function activeWebViewerElementSelector() {
            return '#' + self.id + self.activeWebViewerIdSuffix;
        }

        function inActiveWebViewerElementSelector() {
            return '#' + self.id + self.inactiveWebViewerIdSuffix;
        }

        function swapWebViewers() {
            var temp = self.activeWebViewerIdSuffix;
            self.activeWebViewerIdSuffix = self.inactiveWebViewerIdSuffix;
            self.inactiveWebViewerIdSuffix = temp;
        }

        function onWidgetMoveNextRequested() {
            logger.info('Moving next on request from widget asset');
            clearAssetTimer();
            self.moveNext(self);
        }

        self.enableSync = function(syncConfig, clientDetails) {
            logger.debug("Enabling Sync");

            if (syncConfig.enabled) {
                _MediaSyncController = new MediaSyncController();

                self.syncConfig = syncConfig;
                self.syncEnabled = true;
                //var syncPort = parseInt(syncConfig.configurationParameters.receive_port);

                if (syncConfig.defaultMaster == clientDetails.signageliveSerialNumber) {
                    logger.debug("Player is sync master");
                    self.isSyncMaster = true;
                } else {
                    logger.debug("Player is sync slave");
                    self.isSyncMaster = false;
                }
            }
        }

        var onSyncEvent = function(index) {
            //Sync event received so we have started syncing
            if (!self.isSyncMaster && syncStartTimeoutTimer) {
                clearTimeout(syncStartTimeoutTimer);
                syncStartTimeoutTimer = null;
            }
            //function for changing the asset due to sync
            clearAssetTimer();
            clearSyncFailedTimer();
            //CHANGE NEXT LINE
            var assetIndex = parseInt(index);
            logger.debug('Sync event received - index is: ' + assetIndex);
            if ((self.assetCompleteDone == false) && (self.currentAssetIndex != -1)) {
                self.assetCompleteDone = true;
                logger.debug('Asset complete not done, logging proof of play');
                logProofOfPlayOnAssetComplete(self.playReason);

                if (self.currentAssetIsVideo) {
                    self.videoViewer.stop();
                    self.tvViewer.stop();
                    performVideoCompleteActions();

                    if (self.videoComplete != null)
                        self.videoComplete(self);
                }
            }

            if (self.syncMasterFailed) {
                self.syncMasterFailed = false;
            }
            self.syncPlayAsset(assetIndex);
        };

        self.syncPlayAsset = function(index) {
            logger.debug('Setting current asset index to: ' + index);
            self.currentAssetIndex = index;
            self.moveNext(self);
        }

        function playNextAssetSync() {
            //This is where we need to send the sync signal
            _MediaSyncController.sendPlayAsset(self.currentAssetIndex);
        }

        function getNextPlaylistMediaAsset() {
            var nextPlaylistMediaAsset = null;

            var nextMediaAssetIndex = self.currentAssetIndex;
            logger.info('Current media asset index: ' + self.currentAssetIndex);
            if (_direction == PlayDirection.FORWARDS) {
                logger.debug('Moving forwards');
                nextMediaAssetIndex = self.currentAssetIndex + 1;
                if (nextMediaAssetIndex >= self.currentPlaylist.playlistMediaAssets.length)
                    nextMediaAssetIndex = 0;
            } else {
                logger.debug('Moving backwards');
                nextMediaAssetIndex = self.currentAssetIndex - 1;
                if (nextMediaAssetIndex < 0)
                    nextMediaAssetIndex = self.currentPlaylist.playlistMediaAssets.length - 1;
            }

            nextPlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[nextMediaAssetIndex];
            logger.debug('Next media asset index: ' + nextMediaAssetIndex);
            self.currentAssetIndex = nextMediaAssetIndex;
            logger.debug('New current media asset index: ' + self.currentAssetIndex);

            // if the next asset in the playlist is not valid - perform a search of the playlist to find the next valid asset.
            if (nextPlaylistMediaAsset.isCurrentlyValid() == false) {
                logger.debug('Select Playlist media asset is not valid finding another');
                var startIndex = self.currentAssetIndex;

                var nextValidAsset = null;
                if (_direction == PlayDirection.FORWARDS) {
                    // check from the current asset to the end of the playlist
                    for (var i = startIndex; i < self.currentPlaylist.playlistMediaAssets.length; i++) {
                        var candidatePlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[i];
                        if (candidatePlaylistMediaAsset.isCurrentlyValid()) {
                            nextValidAsset = candidatePlaylistMediaAsset;
                            self.currentAssetIndex = i;
                            break;
                        }
                    }

                    // check from the beginning to the start position, if a valid asset was not found at the end
                    if (nextValidAsset == null) {
                        for (var i = 0; i < startIndex; i++) {
                            var candidatePlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[i];
                            if (candidatePlaylistMediaAsset.isCurrentlyValid()) {
                                nextValidAsset = candidatePlaylistMediaAsset;
                                self.currentAssetIndex = i;
                                break;
                            }
                        }
                    }
                } else {
                    logger.debug('Looking backwards...');
                    // check to the beginning of the playlist from the start position

                    for (var i = startIndex; i >= 0; i--) {
                        var candidatePlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[i];
                        if (candidatePlaylistMediaAsset.isCurrentlyValid()) {
                            nextValidAsset = candidatePlaylistMediaAsset;
                            self.currentAssetIndex = i;
                            break;
                        }
                    }

                    if (nextValidAsset == null) {
                        // check from the current asset to the end of the playlist
                        for (var i = self.currentPlaylist.playlistMediaAssets.length - 1; i > startIndex; i--) {
                            var candidatePlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[i];
                            if (candidatePlaylistMediaAsset.isCurrentlyValid()) {
                                nextValidAsset = candidatePlaylistMediaAsset;
                                self.currentAssetIndex = i;
                                break;
                            }
                        }
                    }
                }

                logger.debug('Rolled forwards - current media asset index: ' + self.currentAssetIndex);
                nextPlaylistMediaAsset = nextValidAsset;
            }
            return nextPlaylistMediaAsset;
        }

        function onVideoComplete() {
            logger.debug('Video Complete');
            if (self.videoComplete != null)
                self.videoComplete(self);

            performVideoCompleteActions();
            // Ensure that play is logged in proof of play
            clearAssetTimer();
            assetComplete(self);
        }

        function performVideoCompleteActions() {
            // Reset number of failures
            if (self.currentScheduledPlaylistMediaAsset && self.currentScheduledPlaylistMediaAsset.mediaAsset) {
                try {
                    self.currentScheduledPlaylistMediaAsset.mediaAsset.numberOfFailures = 0;
                } catch (e) {
                    logger.error(e);
                }
            }
            // Unlock video zone
            Settings.application.criticalSections.videoAvailable = true;
        }

        function onVideoFail() {
            logger.debug('Video fail');
            if (self.videoComplete != null)
                self.videoComplete(self);

            //This counts as asset complete as we don't log proof of play if an asset fails, we just want to move next.
            self.assetCompleteDone = true;
            // Make sure video is available again
            Settings.application.criticalSections.videoAvailable = true;
            // remember this asset has failed.
            if (self.currentScheduledPlaylistMediaAsset && self.currentScheduledPlaylistMediaAsset.mediaAsset) {
                self.currentScheduledPlaylistMediaAsset.mediaAsset.numberOfFailures++;
            }
            // If the same asset has failed more than 5 times then mark it as failed
            if (self.currentScheduledPlaylistMediaAsset && self.currentScheduledPlaylistMediaAsset.mediaAsset && self.currentScheduledPlaylistMediaAsset.mediaAsset.numberOfFailures === 5)
                self.currentScheduledPlaylistMediaAsset.mediaAsset.hasFailed = true;

            clearAssetTimer();
            //If sync we need to do something else
            if (self.syncEnabled && !self.syncMasterFailed) {
                logger.debug("Sync is enabled");

                if (self.isSyncMaster) {
                    logger.debug("Player is master - call playNextAssetSync()");
                    playNextAssetSync();
                } else {
                    //Do nothing - wait for sync event
                    logger.debug("Player is slave - wait for sync event");
                    //Set the timeout for sync failing
                    self.syncFailedTimerId = setTimeout(function() {
                        self.syncMasterFailed = true;
                        logger.debug("Sync failed, move next normally");
                        self.moveNext(sender);
                    }, self.syncFailWaitTime * 1000);
                }
            } else {
                // Move next
                self.moveNext(self);
            }
        }

        function onGaplessVideoComplete(mediaId) {
            if (self.currentAssetIndex == -1)
                self.currentAssetIndex = 0;
            // What media asset are we playing?
            self.currentScheduledPlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[self.currentAssetIndex];
            // Log PoP
            if (window.proofOfPlayLogger != null) {
                window.proofOfPlayLogger.debug(self.layout, self.viewModel, self.currentPlaylist, self.currentScheduledPlaylistMediaAsset, self.currentScheduledPlaylistMediaAsset.mediaAsset, self.currentAssetStartTime.toISOString(), DateTime.now().toISOString(), 'Schedule', 'Schedule');
            }
            // increment current asset id
            self.currentAssetIndex++;
            //Log now as the current asset start time
            self.currentAssetStartTime = DateTime.now();
            if (self.currentAssetIndex > self.currentPlaylist.playlistMediaAssets.length - 1)
                self.currentAssetIndex = 0;
        }

        function onGaplessVideoFail(error) {
            logger.error("Gapless Video failed. ");
            logger.error(JSON.stringify(error));
            self.gaplessPlaybackFailed = true;
            // Restart playlist
            setCurrentPlaylist();
        }

        function onWidgetReady(duration) {
            logger.info('Widget ready - hiding other elements');
            // Ensure image control is hidden
            $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
            // Ensure video and tv in elements are hidden
            self.videoViewerContainer.css('visibility', 'hidden');
            self.videoViewer.hide();
            self.tvViewerContainer.css('visibility', 'hidden');
            self.tvViewer.hide();
            // hide the other web control
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', 'about:blank');
            setAssetPlayTimer(duration);
        }

        function onVideoPlaying() {
            logger.info('onVideoPlaying()');
            // For webOS 4+ we can hide the other elements after video playback starts to provide a more seamless experience
            if (Settings.application.webOSVersion > 3.2) {
                logger.info('Video Ready and playing on panel > webOS 3.2 - hiding other elements');
                prepareForVideoPlayback();
            }
            else {
                logger.info('Video Ready but nothing to do as playing on webOS 3.2 or below');
            }
        }

        function playAsset(mediaAsset, duration) {
            logger.info('! Play Asset: ' + mediaAsset.name + ' for: ' + duration + 'ms');
            var assetType = mediaAsset.contentType;
            var mimeType = mediaAsset.mimeType;

            switch (assetType) {
                case MediaTypeConstants.IMAGE:
                    // Add Support for SVGs. SVGs will use the svgViewer rather than the imageViewer.
                    // The svgViewer is an <object> element and not an <img> element. This is the recommended
                    // way to embed svgs in a HTML page.
                    if (mimeType == "image/svg+xml") {
                        logger.error('SVGs are not currently supported');
                        self.currentScheduledPlaylistMediaAsset.mediaAsset.hasFailed = true;
                        setTimeout(function() {
                            self.moveNext(self);
                        }, 100);
                    }
                    // All other image mime types use the object tag
                    else {
                        // Remove existing event handlers
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).off();
                        // Add the load event handler. If the image has loaded then show it
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).load(function(eventData) {
                            // Log result
                            logger.debug('Image loaded: ' + eventData.currentTarget.src);
                            logger.debug('Target: ' + eventData.target.id);
                            logger.debug('Current Target: ' + eventData.currentTarget.id);
                            logger.debug('Timestamp: ' + eventData.timeStamp);
                            // Show Image
                            showImageControl(duration);
                        });

                        // Add the error event handler. If there was an error loading the image then move next
                        // and mark the asset as failed.
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).error(function() {
                            logger.error('Error loading image');
                            self.currentScheduledPlaylistMediaAsset.mediaAsset.hasFailed = true;
                            setTimeout(function() {
                                self.moveNext(self);
                            }, 100);
                        });

                        // Get the URL to play the Asset and update the image viewer source so that it
                        // attempes to load the image. If it is successful the load event will fire and if it fails
                        // the error event will fire
                        PlayUrlHelper.getMediaAssetPlayUrl(mediaAsset).done(function(url) {
                            if ($(zoneScreenElementSelector()).children(imageViewerElementSelector()).attr('src') == url) {
                                logger.debug('Url the same' + url);
                                showImageControl(duration);
                            } else {
                                logger.debug('Setting image to ' + url);
                                $(zoneScreenElementSelector()).children(imageViewerElementSelector()).attr('src', url);
                            }
                        });
                    }
                    break;
                case MediaTypeConstants.WIDGET:
                    logger.debug('Showing widget');
                    // Show Web loading Animation
                    if (Settings.application.features.showWebsiteLoadingAnimation) {
                        showWebLoadingAnimation();
                    }
                    var widgetGuid = mediaAsset.localFileName.replace('.wgt', '');
                    var widgetPath = Settings.application.fileSystem.root + '/' + Settings.application.fileSystem.widgetDirectory + '/' + widgetGuid + '/';
                    var configUrl = widgetPath + 'config.xml';

                    configUrl = './content/widgets/' + widgetGuid + '/config.xml';

                    $.ajax({
                        url: configUrl,
                        type: 'GET',
                        dataType: "xml",
                        async: false
                    }).done(function(data, textStatus, jqXHR) {
                        logger.debug('Config xml loaded ' + data);
                        if (data != null) {
                            var startfileName = null;
                            try {
                                startfileName = $(data).children('widget').children('content').attr('src');
                            } catch (err) {
                                logger.debug('Potential start file: ' + potentialStartFile);
                            }

                            var potentialStartFile = widgetPath + startfileName;
                            var startFileFound = false;
                            logger.debug('Potential start file: ' + potentialStartFile);

                            if (!(startfileName === undefined || startfileName == null)) {
                                File.exists2(widgetPath + startfileName).done(function(exists) {
                                    if (exists) {
                                        potentialStartFile = startfileName;
                                        startFileFound = true;
                                        logger.debug('Widget start file location:    ./content/widgets/' + widgetGuid + '/' + potentialStartFile);
                                        var src = './content/widgets/' + widgetGuid + '/' + potentialStartFile;
                                        var url = self.buildWebWidgetURL(src);
                                        logger.debug("Start File built with additional data: " + url);
                                        playWidget(url, duration);
                                    } else {
                                        logger.error('Start file was not found - this widget is invalid');
                                        isCancelled = true;
                                        self.assetTimerId = setTimeout(function() {
                                            assetComplete(self);
                                        }, 1500);
                                    }
                                });
                            } else {
                                logger.debug("No start file in config - must be a default start file");
                                //Get files and check in this order:  'index.htm',  'index.html',  'index.svg', index.xhtml index.xht
                                Directory.listFiles(widgetPath).done(function(files) {
                                    for (var i = 0; i < files.length; ++i) {
                                        switch (files[i].name) {
                                            case "index.htm":
                                            case "index.html":
                                            case "index.svg":
                                            case "index.xhtml":
                                            case "index.xht":
                                                startFileFound = true;
                                                potentialStartFile = files[i].name;
                                                break;
                                        }
                                    }
                                    if (startFileFound) {
                                        // Play Widget
                                        var src = './content/widgets/' + widgetGuid + '/' + potentialStartFile;
                                        var url = self.buildWebWidgetURL(src);
                                        logger.debug("Start File built with additional data: " + url);
                                        playWidget(url, duration);
                                    } else {
                                        if (Settings.application.features.showWebsiteLoadingAnimation) {
                                            hideLoadingAnimation();
                                        }
                                        logger.error('Start file was not found - this widget is invalid');
                                        isCancelled = true;
                                        self.assetTimerId = setTimeout(function() {
                                            assetComplete(self);
                                        }, 1500);
                                    }
                                });
                            }
                        } else {
                            logger.error('Config file data is null - this widget is invalid');
                            isCancelled = true;
                            self.assetTimerId = setTimeout(function() {
                                assetComplete(self);
                            }, 1500);
                        }
                    }).fail(function(jqXHR, textStatus, errorThrown) {
                        logger.error('Error loading widget config');
                        self.assetTimerId = setTimeout(function() {
                            assetComplete(self);
                        }, 1500);
                    });
                    break;
                case MediaTypeConstants.WEB:
                    var assetUrl = self.buildWebWidgetURL(mediaAsset.remoteUrl);
                    loadWebUrl(assetUrl, duration);
                    break;
                case MediaTypeConstants.CAPTURE:
                    // show background
                    self.tvViewerContainer.css('visibility', 'visible');
                    self.videoViewerContainer.css('visibility', 'hidden');
                    // hide all children
                    $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
                    $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', 'about:blank');
                    self.widgetViewer.hide();
                    // Previous asset was a capture card so just keep playing it (dont reinitialize)
                    if (self.previousScheduledPlaylistMediaAsset != null && self.previousScheduledPlaylistMediaAsset.mediaAsset.contentType == MediaTypeConstants.CAPTURE) {
                        self.tvViewer.setInitialized(true);
                        setAssetPlayTimer(duration);
                    } else {
                        // hide video viewer
                        self.videoViewer.hide();
                        // attempt to play video (this will initialize it and automatically start playback)
                        self.tvViewer.play();
                        // so we should set the timer (add half a second for initialisation)
                        setAssetPlayTimer(duration + 500);
                    }
                    break;
                case MediaTypeConstants.VIDEO:
                    // Play video
                    PlayUrlHelper.getMediaAssetPlayUrl(mediaAsset).done(function(url) {
                        // If this is webOS 3.2 or below then hide everything then play the video
                        // This leads to worse transitions but solves an issue where videos sometimes freeze
                        if (Settings.application.webOSVersion <= 3.2) {
                            logger.info('Playing video the old fashioned way');
                            prepareForVideoPlayback();
                        }
                        // Play video
                        self.videoViewer.play(url);
                        if (duration != 0) {
                            // set the timer
                            setVideoAssetPlayTimer(duration);
                        }
                        if (self.videoStart != null)
                            self.videoStart(self);
                    });
                    break;
                case MediaTypeConstants.IPTV:
                    self.videoViewerContainer.css('visibility', 'visible');
                    // hide all children
                    $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
                    self.widgetViewer.hide();
                    $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
                    $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
                    // hide tv viewer
                    self.tvViewerContainer.css('visibility', 'hidden');
                    self.tvViewer.hide();
                    self.tvViewer.stop();
                    // Play Video
                    self.videoViewer.playIPTV(mediaAsset.remoteUrl);
                    if (duration != 0) {
                        // set the timer
                        setVideoAssetPlayTimer(duration);
                    }
                    // Call Video Start if the function is defined
                    if (self.videoStart != null)
                        self.videoStart(self);
                    break;
                default:
                    setTimeout(function() {
                        logger.error('Unable to determine asset media type. Moving to next asset...');
                        self.moveNext(self);
                    }, 500);
            }
        }

        function prepareForVideoPlayback() {
            // show background
            self.videoViewerContainer.css('visibility', 'visible');
            // hide all children
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', 'about:blank');
            self.widgetViewer.hide();
            self.tvViewerContainer.css('visibility', 'hidden');
            self.tvViewer.hide();
            self.tvViewer.stop();    
        }

        function playWidget(src, duration) {
            self.widgetViewer.play(src, duration, self.currentScheduledPlaylistMediaAsset.mediaAsset);
        }

        function checkOnlineStatus() {
            // If this player is 3.2 or less just resolve that it is online and try to play the web content.
            // We never used to check and this API does not work correctly on webOS 3 and 3.2
            if (Settings.application.webOSVersion <= 3.2) {
                return $.Deferred().resolve({
                    network_connected: window.navigator.onLine,
					internet_access: true
                });
            }
            else {
                return NetworkInterface.getOnlineStatus();
            }
        }

        function loadWebUrl(remoteUrl, duration) {

            // Check if player is online before loading web url. On LG if you attempt to load a web page but the player is offline
            // the load event stills fires and the src is set to the page but it shows a defualt network error page instead.
            // Therefore, to avoid this if the network connection is unavailable we will skip the asset instead
            logger.info('Checking if device is online...')
            checkOnlineStatus().done(function(status) {
                logger.info('Device online: ' + status.internet_access);
                if (status.internet_access === false) {
                    logger.warning('Internet unavailable - skipping web asset');
                    // prevent the asset from loading
                    isCancelled = true;
                    // remove the broken web viewer - I'm not sure why this is required however if the iframe is left in the document it is not usable again (this does not happen in chrome or FF)
                    // by removing it and adding one stright back in all of the content plays as expected
                    $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).remove();
                    // add another - clean viewer
                    $('#' + self.id).append("<iframe class='viewer' id='" + self.id + self.inactiveWebViewerIdSuffix + "' frameborder='0' scrollbars='no' src='' style='position:absolute;left:0px;top:0px;width:" + self.width + "px;height:" + self.height + "px;visibility:hidden'></iframe>");
                    // swap the references to the viewers
                    swapWebViewers();
                    // cancelled - move to the next asset
                    self.assetTimerId = setTimeout(function() {
                        assetComplete(self);
                    }, 150);

                    return;
                }

                // Device is online
                logger.info('Device is online. Playing web asset');

                // If this is a remote widget then replace the url
                if (Settings.application.features.supportsTLS12 === false && Settings.application.api.useLegacyTlsUrl) {
                    remoteUrl = remoteUrl.replace(Settings.application.urls.widgetUrl, Settings.application.urls.legacyTlsWidgetUrl);
                }

                logger.info('URL is: ' + remoteUrl);
                var loaded = false;
                var isCancelled = false;
                // Ensure event is removed and readded
                $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).off('load').on('load', function() {
                    var $this = $(this);
                    if (isCancelled == false && $this.attr('src') != 'about:blank') {
                        loaded = true;
                        // Remove load event handler 
                        $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).off('load');
                        // clear the timeout timer
                        clearWebContentTimeoutTimer();
                        // Ensure image control is hidden
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
                        // Ensure video and tv in elements are hidden
                        self.videoViewerContainer.css('visibility', 'hidden');
                        self.videoViewer.hide();
                        self.tvViewerContainer.css('visibility', 'hidden');
                        self.tvViewer.hide();
                        self.widgetViewer.hide();
                        // hide the other web control
                        $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
                        // Finally show the new iframe
                        $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'visible');
                        // swap the references to the viewers
                        swapWebViewers();
                        // set the timer
                        setAssetPlayTimer(duration);
                    }
                });

                // set the url
                if ($(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src') != remoteUrl) {
                    /// handle assets that don't load
                    self.webZoneLoadTimer = setTimeout(function() {
                        if (!loaded) {
                            logger.info('Web Zone load timed out');
                            // prevent the asset from loading
                            isCancelled = true;
                            // remove the broken web viewer - I'm not sure why this is required however if the iframe is left in the document it is not usable again (this does not happen in chrome or FF)
                            // by removing it and adding one stright back in all of the content plays as expected
                            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).remove();
                            // add another - clean viewer
                            $('#' + self.id).append("<iframe class='viewer' id='" + self.id + self.inactiveWebViewerIdSuffix + "' frameborder='0' scrollbars='no' src='' style='position:absolute;left:0px;top:0px;width:" + self.width + "px;height:" + self.height + "px;visibility:hidden'></iframe>");
                            // swap the references to the viewers
                            swapWebViewers();
                            //cancelled - move to the next asset
                            assetComplete(self);
                        }
                    }, 10000);
                    $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', remoteUrl);
                }
            });
        }

        function showImageControl(duration) {
            logger.debug('showImageControl(' + duration + ')');
            // show the image control
            $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'visible');
            // Hide SVG Image viewer
            $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
            // hide the web controls
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', 'about:blank');
            self.widgetViewer.hide();
            self.videoViewerContainer.css('visibility', 'hidden');
            self.videoViewer.hide();
            self.tvViewerContainer.css('visibility', 'hidden');
            self.tvViewer.hide();
            // set the timer
            setAssetPlayTimer(duration);
        }

        function showWebLoadingAnimation() {
            logger.debug("Showing Website loading animation");
            //$('#' + self.id).css('background-color', '#FFFFFF');
            $('#' + self.id).append('<div id="webLoadingAnimation"><div class="bounce1"></div><div class="bounce2"></div><div class="bounce3"></div></div>');
            var margin = (self.height / 2) - 18;
            $('#webLoadingAnimation').css('margin', margin + 'px auto 0');
        }

        function hideLoadingAnimation() {
            logger.debug("Hiding Website loading animation");
            //$('#' + self.id).css('background-color', 'transparent');
            $('#webLoadingAnimation').remove();
        }

        function setAssetPlayTimer(duration) {
            self.assetTimerId = setTimeout(function() {
                assetComplete(self);
            }, duration);
        }

        function setVideoAssetPlayTimer(duration) {
            self.assetTimerId = setTimeout(function() {
                self.videoViewer.stop();
                onVideoComplete();
            }, duration);
        }

        function assetComplete(sender) {
            logger.debug('Asset Complete');
            //In case of race condition due to sync. This is set to false on movenext
            if (self.assetCompleteDone === true)
                return;
            //Do we need to check for asset complete done and stop processing?
            self.assetCompleteDone = true;
            // Log PoP
            logProofOfPlayOnAssetComplete(self.playReason);
            if (self.syncEnabled && !self.syncMasterFailed) {
                logger.debug("Sync is enabled");

                if (self.isSyncMaster) {
                    logger.debug("Player is master - call playNextAssetSync()");
                    playNextAssetSync();
                } else {
                    //Do nothing - wait for sync event
                    logger.debug("Player is slave - wait for sync event");

                    //Set the timeout for sync failing
                    self.syncFailedTimerId = setTimeout(function() {
                        self.syncMasterFailed = true;
                        logger.debug("Sync failed, move next normally");
                        self.moveNext(sender);
                    }, self.syncFailWaitTime * 1000);
                }
            } else {
                // Move next
                logger.debug("Not synced, move next normally");
                self.moveNext(sender);
            }
        }

        function logProofOfPlayOnAssetComplete(endReason) {
            try {
                if (window.proofOfPlayLogger != null) {
                    var playlistMediaAsset = null;
                    var mediaAsset = self.currentScheduledPlaylistMediaAsset;
                    if (mediaAsset.mediaAsset) {
                        mediaAsset = self.currentScheduledPlaylistMediaAsset.mediaAsset;
                        playlistMediaAsset = self.currentScheduledPlaylistMediaAsset;
                    }

                    logger.debug("PoP Log: Asset Name: " + mediaAsset.name + ", Start Time: " + self.currentAssetStartTime.toISOString() + ", End Time: " + DateTime.now().toISOString() + ", Play Reason: " + self.playReason + ", End Reason: " + endReason);
                    window.proofOfPlayLogger.debug(self.layout, self.viewModel, self.currentPlaylist, playlistMediaAsset, mediaAsset, self.currentAssetStartTime.toISOString(), DateTime.now().toISOString(), self.playReason, endReason);
                }
            } catch (e) {
                logger.warning("Error logging PoP: " + e);
            }
        }

        function startPlaylistLoop() {
            // reset the playlist
            self.currentAssetIndex = -1;
            logger.debug("Starting playlist loop..");

            if (self.gaplessPlaybackFailed) {
                logger.warning("Gapless playback has previously failed.");
            }
            if (self.syncEnabled) {
                logger.debug("Sync is enabled");

                if (self.isSyncMaster) {
                    logger.debug("Player is master");
                    _MediaSyncController.startMaster(self.syncConfig.groupId, onSyncEvent);
                } else {
                    //Do nothing - wait for sync event
                    logger.debug("Player is slave - find master and wait for sync event");
                    _MediaSyncController.startSlave(self.syncConfig.groupId, onSyncEvent);
                    syncStartTimeoutTimer = setTimeout(syncStartTimedOut, self.secondsToWaitForMasterOnStartup * 1000);
                }
            } else {
                logger.debug("Not synced, play normally");
                // Note: We need to determine ideally need to check if this is a WebOS 2.0 panel - to consider later!
                if (Settings.application.features.supportsGaplessPlayback && Settings.application.sdkVersion >= 1.2 && self.currentPlaylist != null && self.currentPlaylist.meetsGaplessPlaybackRequirements() && !self.gaplessPlaybackFailed) {
                    logger.debug("Starting gapless playback!");
                    //Settings.application.criticalSections.videoAvailable = false;
                    playGaplessVideos();
                } else {
                    logger.debug("Starting normal playback");
                    self.moveNext(self);
                }
            }

        }

        var syncStartTimeoutTimer = null;

        function syncStartTimedOut() {
            syncStartTimeoutTimer = null;
            self.currentAssetIndex = -1;
            self.syncMasterFailed = true;
            self.moveNext(self);
        }

        function playGaplessVideos() {
            // Set initial asset to play
            self.currentScheduledPlaylistMediaAsset = getNextPlaylistMediaAsset();
            if (self.currentScheduledPlaylistMediaAsset != null) {
                self.currentPlaylist.getGaplessPlaylistFile()
                    .done(function(gaplessFilePlaybackPath, gaplessFileLocalFilePath, endDate) {
                        logger.debug("MediaZoneView: Gapless file retrieved. Starting gapless playback...");
                        // Store gapless playlist file details
                        self.gaplessFileLocalFilePath = gaplessFileLocalFilePath;
                        self.gaplessFilePlaybackPath = gaplessFilePlaybackPath;
                        // Show container & Hide/Stop all other playback elements
                        self.videoViewerContainer.css('visibility', 'visible');
                        $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).attr('src', 'about:blank');
                        $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).attr('src', 'about:blank');
                        self.tvViewerContainer.css('visibility', 'hidden');
                        self.tvViewer.hide();
                        self.tvViewer.stop();
                        // Set timeout to restart playlist
                        var playlistTimeout = endDate.subtract(DateTime.now());
                        self.playlistTimerId = setTimeout(function() {
                            Settings.application.criticalSections.videoAvailable = true;
                            setCurrentPlaylist();
                        }, playlistTimeout);
                        // Start Playback
                        self.videoViewer.playGapless(self.gaplessFilePlaybackPath);
                        logger.info("MediaZoneView: Gapless playback started.");
                    })
                    .fail(function() {
                        logger.error("MediaZoneView: Failed to create gapless file!");
                    });
                }
                else {
                    logger.info('No valid assets to play. Will try again in 2 seconds...');
                    self.assetTimerId = setTimeout(function() {
                        playGaplessVideos();
                    }, 2000);
                }
        }

        function stopPlaylistLoop(endReason) {
            logger.debug('Stopping playlist loop');
            // If this is gapless we need to update the current asset before we log PoP
            if (Settings.application.features.supportsGaplessPlayback && Settings.application.sdkVersion >= 1.2 && self.currentPlaylist != null && self.currentPlaylist.meetsGaplessPlaybackRequirements() && !self.gaplessPlaybackFailed) {
                self.currentScheduledPlaylistMediaAsset = self.currentPlaylist.playlistMediaAssets[self.currentAssetIndex];
            }
            // Log that the current asset has stopped playing
            logProofOfPlayOnAssetComplete(endReason);
            // remove images and web
            $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
            $(zoneScreenElementSelector()).children(svgViewerElementSelector()).css('visibility', 'hidden');

            self.widgetViewer.hide();
            self.widgetViewer.remove();
            // stop video
            self.videoViewerContainer.css('visibility', 'hidden');
            self.videoViewer.removeEventListeners();
            self.videoViewer.hide();
            // stop TV in
            self.tvViewerContainer.css('visibility', 'hidden');
            self.tvViewer.hide();
            Settings.application.criticalSections.videoAvailable = true;
            // Delete the gapless playlist file as no longer needed
            // Reset gapless properties
            if (self.gaplessFileLocalFilePath != null) {
                self.currentPlaylist.deleteGaplessPlaylistFile();
                self.gaplessFileLocalFilePath = null;
                self.gaplessFilePlaybackPath = null;
            }

            // clear timers
            clearAllTimers();
        }

        function clearSyncFailedTimer() {
            clearTimeout(self.syncFailedTimerId);
        }

        function clearAssetTimer() {
            // stop the asset timer
            clearTimeout(self.assetTimerId);
        }

        function clearPlaylistTimer() {
            // stop the playlist timer
            clearTimeout(self.playlistTimerId);
        }

        function clearWebContentTimeoutTimer() {
            // clear the web content load timeout timer
            clearTimeout(self.webZoneLoadTimer);
        }

        function clearAllTimers() {
            clearAssetTimer();
            clearPlaylistTimer();
            clearWebContentTimeoutTimer();
        }

        /**
         * Move to the next playlist media asset
         * @param  {[type]} sender [description]
         * @return {[type]}        [description]
         */
        self.moveNext = function(sender) {
            // Let any current widgets that are chaining in to that event know that the widget will be closing as we are moving to the next asset
            if (self.widgetViewer) {
                self.widgetViewer.sendWidgetClosingNotification();
            }

            try {
                self.assetCompleteDone = false;
                if (self.currentPlaylist == null) {
                    // there is no current playlist - play the default media asset
                    self.currentScheduledPlaylistMediaAsset = { mediaAsset: self.viewModel.defaultMediaAsset };
                    playAsset(self.viewModel.defaultMediaAsset, 86400000);
                    self.currentAssetStartTime = DateTime.now();
                } else {
                    // reset/stop any previous assets
                    if (self.currentScheduledPlaylistMediaAsset != null) {
                        // store current media asset as previous media asset
                        self.previousScheduledPlaylistMediaAsset = self.currentScheduledPlaylistMediaAsset;
                        var assetType = self.currentScheduledPlaylistMediaAsset.mediaAsset.contentType;
                        switch (assetType) {
                            case MediaTypeConstants.CAPTURE:
                                //self.tvViewer.stop();
                                // need to set the video as available for the CAPTURE here as there is not a specific complete call back
                                Settings.application.criticalSections.videoAvailable = true;
                                break;
                            case MediaTypeConstants.VIDEO:
                                //self.videoViewer.stop();
                                // Nothing to do
                                break;
                            case MediaTypeConstants.IPTV:
                                // Nothing to do
                                break;
                        }
                    }
                    self.currentScheduledPlaylistMediaAsset = getNextPlaylistMediaAsset();
                    // has a valid media asset been found?
                    if (self.currentScheduledPlaylistMediaAsset != null) {
                        self.currentAssetStartTime = DateTime.now();
                        // Perform any PMA actions
                        // Note that we should disable NFC when assets aren't using it but this seems to break
                        // quite quickly
                        processPlaylistMediaAssetActions(self.currentScheduledPlaylistMediaAsset);
                        // is this a video asset?
                        var isVideo = (self.currentScheduledPlaylistMediaAsset.mediaAsset.contentType == MediaTypeConstants.VIDEO || self.currentScheduledPlaylistMediaAsset.mediaAsset.contentType == MediaTypeConstants.IPTV || self.currentScheduledPlaylistMediaAsset.mediaAsset.contentType == MediaTypeConstants.CAPTURE);
                        if (isVideo) {
                            self.currentAssetIsVideo = true;
                            // If the video is available or if the video is not available but the video in this zone is paused
                            if (Settings.application.criticalSections.videoAvailable) {
                                //Settings.application.criticalSections.videoAvailable = false;
                                logger.debug('Locking video to play: ' + self.currentScheduledPlaylistMediaAsset.mediaAsset.name);

                                playAsset(self.currentScheduledPlaylistMediaAsset.mediaAsset, self.currentScheduledPlaylistMediaAsset.duration);
                            } else {
                                logger.debug('Video in use - moving next');
                                self.currentScheduledPlaylistMediaAsset = null;
                                // set a timer to wait for the next valid asset
                                self.assetTimerId = setTimeout(function() {
                                    self.moveNext(self);
                                }, 1000);
                            }
                        } else {
                            self.currentAssetIsVideo = false;
                            playAsset(self.currentScheduledPlaylistMediaAsset.mediaAsset, self.currentScheduledPlaylistMediaAsset.duration);
                        }
                    } else {
                        logger.info('No valid media asset. Will try again in 3 seconds');
                        // hide invalid media
                        $(zoneScreenElementSelector()).children(inActiveWebViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(activeWebViewerElementSelector()).css('visibility', 'hidden');
                        $(zoneScreenElementSelector()).children(imageViewerElementSelector()).css('visibility', 'hidden');
                        self.currentScheduledPlaylistMediaAsset = null;
                        // set a timer to wait for the next valid asset
                        self.assetTimerId = setTimeout(function() {
                            self.moveNext(self);
                        }, 2000);
                    }
                }
            } catch (e) {
                logger.error('Error whilst moving next ' + e.message);
            }
        };

        function processPlaylistMediaAssetActions(playlistMediaAsset) {
            try {
                if (playlistMediaAsset.actions != null) {
                    var actions = playlistMediaAsset.actions;

                    $.each(actions, function(index, action) {
                        if (action.type == PlaylistMediaAssetActionType.WRITE_DATA_TO_FILE) {
                            NFCInterface.isEnabled()
                                .done(function(enabled) {
                                    if (enabled) {
                                        NFCInterface.writeTag(action.data);
                                    } else {
                                        NFCInterface.enable()
                                            .done(function() {
                                                NFCInterface.writeTag(action.data);
                                            })
                                            .fail(function() {
                                                logger.error("Failed to enable NFC tag");
                                            });
                                    }
                                });
                        } else if (action.type == PlaylistMediaAssetActionType.CALL_URL) {
                            callUrl(action.url)
                                .done(function(data, textStatus, jqXHR) {
                                    logger.debug("Successfully completed CALL_URL action. Called out to url: " + action.url);
                                })
                                .fail(function(jqXHR, textStatus, errorThrown) {
                                    logger.error("Successfully completed CALL_URL action. Called out to url: " + action.url);
                                });
                        }
                    });
                }

            } catch (e) {
                logger.error("Error trying to process PMA action");
            }
        }

        function callUrl(url) {
            try {
                return $.ajax({
                    url: url
                });

            } catch (e) {
                logger.error('Error executing call ' + e.message);
                return null;
            }
        }

        /**
         * Set the current playlist
         */
        function setCurrentPlaylist() {
            // is there a playlist already playing?
            if (self.currentPlaylist != null) {
                // stop the playlist loop
                logger.debug('Stopping playlist loop');
                stopPlaylistLoop();
            }
            // find the current schedule
            var currentScheduledPlaylist = null;
            $(self.viewModel.scheduledPlaylists).each(function(i, scheduledPlaylist) {
                try {
                    if (scheduledPlaylist.start <= DateTime.now() && scheduledPlaylist.end >= DateTime.now()) {
                        currentScheduledPlaylist = scheduledPlaylist;
                    }
                } catch (e) {

                }
            });
            // if there a playlist scheduled play it, otherwise play the default
            var playlistTimeout = 0;
            if (currentScheduledPlaylist != null) {
                self.currentPlaylist = currentScheduledPlaylist.playlist;
                playlistTimeout = currentScheduledPlaylist.end.subtract(DateTime.now());
            } else {
                self.currentPlaylist = self.viewModel.defaultPlaylist;

                // find the start of the next scheduled - if there is one

                // find the next scheduled playlist in the future
                var candidateNextScheduledPlaylist = null;

                $(self.viewModel.scheduledPlaylists).each(function(i, scheduledPlaylist) {
                    try {
                        if (scheduledPlaylist.start > DateTime.now()) {
                            if (candidateNextScheduledPlaylist == null)
                                candidateNextScheduledPlaylist = scheduledPlaylist;
                            else if (scheduledPlaylist.start < candidateNextScheduledPlaylist.start)
                                candidateNextScheduledPlaylist = scheduledPlaylist;
                        }
                    } catch (e) {

                    }
                });

                // if there is a candidate set the timeout of the default to the start of it - otherwise play for 12 hours
                if (candidateNextScheduledPlaylist != null)
                    playlistTimeout = candidateNextScheduledPlaylist.start.subtract(DateTime.now());
                else {
                    // if there are no scheduled playlists in the future don't time out
                    playlistTimeout = -1;
                }
            }

            // is there a playlist
            if (self.currentPlaylist != null) {
                logger.debug('There are ' + self.currentPlaylist.playlistMediaAssets.length + ' assets in the playlist');

                // start playing the playlist
                startPlaylistLoop();

                // has a timeout been set?
                if (playlistTimeout > 0) {
                    // set the timer to play the schedule at the appropriate time
                    self.playlistTimerId = setTimeout(function() {
                        setCurrentPlaylist();
                    }, playlistTimeout);
                }
            } else {
                // there is no current playlist - play the default media asset
                self.currentScheduledPlaylistMediaAsset = { mediaAsset: self.viewModel.defaultMediaAsset };
                playAsset(self.viewModel.defaultMediaAsset, 86400000);
            }
        }

        self.play = function() {
            logger.debug('Playing MediaZoneView');
            setCurrentPlaylist();
            _currentStatus = PlayStatus.PLAYING;
        };

        self.stop = function(endReason) {
            logger.debug('Stopping MediaZoneView');
            stopPlaylistLoop(endReason);

            if (self.syncEnabled)
                _MediaSyncController.stop();

            _currentStatus = PlayStatus.STOPPED;
        };

        self.pause = function() {
            _currentStatus = PlayStatus.PAUSED;
            clearAssetTimer();
        };

        self.skip = function() {
            clearAssetTimer();

            if (_direction != PlayDirection.FORWARDS) {
                logger.debug('Change of play direction to forwards');
                _direction = PlayDirection.FORWARDS;
            }

            self.moveNext(self);
        };

        self.back = function() {
            clearAssetTimer();

            if (_direction != PlayDirection.BACKWARDS) {
                logger.debug('Change of play direction to backwards');
                _direction = PlayDirection.BACKWARDS;
            }

            self.moveNext(self);
        };

        self.setMute = function(muted) {
            self.videoViewer.setMute(muted);
            self.tvViewer.setMute(muted);
        };

        self.volumeUp = function() {
            self.videoViewer.volumeUp();
            self.tvViewer.volumeUp();
        };

        self.volumeDown = function() {
            self.videoViewer.volumeDown();
            self.tvViewer.volumeDown();
        };

        self.buildWebWidgetURL = function(originalUrl) {
            var url = decodeURI(originalUrl);
            if (self.triggerAdditionalData != null) {
                if (url.includes('?')) {
                    url = url + '&';
                } else {
                    url = url + '?';
                }

                self.triggerAdditionalData.forEach(function(data) {
                    logger.debug('Appending additional data ' + data.Key + ":" + data.Value);
                    url = url + data.Key + "=" + data.Value + "&";
                });

                url = url.substr(0, url.lastIndexOf('&'));
            }
            logger.info('Built URL with additional data: ' + encodeURI(url));
            return encodeURI(url);
        };

        self.onKeyPress = function(key) {
            switch (self.currentScheduledPlaylistMediaAsset.mediaAsset.contentType) {
                case MediaTypeConstants.WIDGET:
                    self.widgetViewer.onKeyPress(key);
                    break;
                default:
                    //widget only
                    logger.debug("No widget playing in this window");
                    break;
            }
        };
    }

    return MediaZoneView;
});