define(["models/PlaylistMediaAssetModel", "models/MediaTypeConstants", "system/io/FileBase", "system/io/File", "properties/App.Settings", "system/DateTime"], function(PlaylistMediaAssetModel, MediaTypeConstants, FileBase, File, Settings, DateTime) {
	function PlaylistModel(data) {
		//logger.debug("Playlist Model: " + JSON.stringify(data));
		var self = this;
		self.externalId = 0;
		self.name = '';
		self.playlistMediaAssets = [];
		self.externalId = data.id;
		self.name = data.name;
		self.includeInProofOfPlay = data.include_in_proof_of_play;
		self.gaplessFilePlayPath = "./content/gaplessPlaylist_" + self.externalId + ".txt";
		self.gaplessFileLocalFilePath = Settings.application.fileSystem.root + "/gaplessPlaylist_" + self.externalId + ".txt";

		$(data.media_assets).each(function(index, value) {
			var playlistMediaAsset = new PlaylistMediaAssetModel(value);
			self.playlistMediaAssets.push(playlistMediaAsset);
		});

		/**
		 * Checks whether the playlist meets gapless requirements and returns true if it does, false otherwise.
		 * Note that ideally we would check codecs etc but for the time being we are only checking if the playlist is
		 * all videos.
		 * @return {Boolean} True if the playlist can be played in gapless, false otherwise
		 */
		self.meetsGaplessPlaybackRequirements = function() {
			var isAllMP4Videos = true;

			for (var i = 0; i < self.playlistMediaAssets.length; i++) {
				var plma = self.playlistMediaAssets[i];
				var ma = plma.mediaAsset;

				// If it is a video and the 
				if (ma.contentType == MediaTypeConstants.VIDEO && ma.mimeType == "video/mp4") {
					isAllMP4Videos = isAllMP4Videos & true;
				} else {
					isAllMP4Videos=false;
					break;
				}
			}

			return isAllMP4Videos;
		};

		/**
		 * Checks if a gapless playback file for this playlist has already been created
		 * @return {Boolean} Returns true if the playlist file already exists, false otherwise.
		 */
		self.gaplessPlaylistFileExists = function() {
			var dfd = $.Deferred();

			logger.debug("Checking if gapless playlist file already exists...");

			FileBase.fileExists(self.gaplessFileLocalFilePath)
			.done(function(exists){
				dfd.resolve(exists);
			});

			return dfd;
		};

		/**
		 * Returns a list if PMAs and the end date for playing
		 * taking into account validity events
		 * @return {[type]} [description]
		 */
		self.getGaplessPlaylist = function() {
			// set the next validity event high so that the checks are always less than this
			var nextValidityEvent = DateTime.now().addDays(365);

			// calculate the validity events for the next 7 days
			var startOfPeriod = DateTime.now();
			var endOfPeriod = startOfPeriod.addDays(7);

			// for each playlist media asset
			for (var i = 0; i < self.playlistMediaAssets.length; i++) {
				var plma = self.playlistMediaAssets[i];

				var validPeriods = plma.getValidPeriods(startOfPeriod, endOfPeriod);
				
				// and it's validity periods
				for (var j = 0; j < validPeriods.length; j++) {
					// find the next event - this could be the start or end of any given asset
					if (validPeriods[j].start > DateTime.now() && validPeriods[j].start < nextValidityEvent)
						nextValidityEvent = validPeriods[j].start;

					if (validPeriods[j].end > DateTime.now() && validPeriods[j].end < nextValidityEvent)
						nextValidityEvent = validPeriods[j].end;
				}
			}

			var plmas = [];
			// get the assets that are valid until the calculated date
			for (var i = 0; i < self.playlistMediaAssets.length; i++) {
				var plma = self.playlistMediaAssets[i];

				var validPeriods = plma.getValidPeriods(startOfPeriod, nextValidityEvent);
				logger.info('There are ' + validPeriods.length + ' validity periods');
				
				// and it's validity periods
				for (var j = 0; j < validPeriods.length; j++) {
					// find the next event - this could be the start or end of any given asset
					if (validPeriods[j].start <= DateTime.now() && validPeriods[j].end >= nextValidityEvent)
						plmas.push(plma);
				}
			}

			return {
				playlistMediaAssets : plmas,
				endDate : nextValidityEvent
			};
		};

		/**
		 * Gets the URL that should be set to the src attribute for the video player.
		 * If the file does not exist already it will attempt to create and return it.
		 * @return {String} the path to the gapless file.
		 */
		self.getGaplessPlaylistFile = function() {

			var dfd = $.Deferred();

			logger.debug("Getting the gapless playlist file...");

			var playlist = self.getGaplessPlaylist();

			self.gaplessPlaylistFileExists()
			.always(function(fileExists){

				if (fileExists)
				{
					logger.debug("File " + self.gaplessFilePlayPath + " found.");
					dfd.resolve(self.gaplessFilePlayPath, self.gaplessFileLocalFilePath, playlist.endDate);
				}
				else
				{
					logger.debug("File " + self.gaplessFilePlayPath + " not found. Attempting to write the gapless file...");
					self.writeGaplessPlaylistFile(playlist.playlistMediaAssets)
					.done(function(){
						logger.debug("File " + self.gaplessFilePlayPath + " not found. Attempting to write the gapless file...");
						dfd.resolve(self.gaplessFilePlayPath, self.gaplessFileLocalFilePath, playlist.endDate);
					})
					.fail(function(){
						logger.error("Error writing playlist file.");
						dfd.reject();
					});
				}

			});

			return dfd;

		};

		/**
		 * Writes the gapless playlist file
		 * @return {[type]} [description]
		 */
		self.writeGaplessPlaylistFile = function(playlistMediaAssets)
		{
			var dfd = $.Deferred();

			var PMAs = [];
			if (playlistMediaAssets == "undefined")
				PMAs = self.playlistMediaAssets;
			else
				PMAs = playlistMediaAssets;

			// Create file data
			logger.debug("WriteGaplessPlaylistFile(): Starting to generate gapless playlist file...");
			var playlistFileData = "";
			$.each(PMAs, function(index, pma){
				playlistFileData += Settings.application.gaplessPlayback.videoFileDir + pma.mediaAsset.localFileName + "\n";
			});

			logger.debug("WriteGaplessPlaylistFile(): Gapless playlist file generated: " + playlistFileData);
			logger.debug("WriteGaplessPlaylistFile(): Starting writing to: " + self.gaplessFileLocalFilePath);

			// Write file
			FileBase.writeToFileSystem(self.gaplessFileLocalFilePath, playlistFileData, 'truncate')
			.done(function(){
				logger.debug("Successfully wrote  gapless playlist file!");
				dfd.resolve();
			})
			.fail(function(){
				logger.error("Error writing gapless playlist file!");
				dfd.reject();
			});

			return dfd;
		};

		/**
		 * [deleteGaplessPlaylistFile description]
		 * @return {[type]} [description]
		 */
		self.deleteGaplessPlaylistFile = function()
		{
			logger.debug("Deleting gapless playlist file.");

			var dfd = $.Deferred();

			File.delete(self.gaplessFileLocalFilePath)
			.done(function(){
				logger.debug("Gapless playlist file deleted.");
				dfd.resolve();
			})
			.fail(function(){
				logger.error("There was an error deleting the gapless playlist file.");
				dfd.reject();
			});
			
			return dfd;
		};
	}

	return PlaylistModel;
});
