define(["system/power/PowerInterface", "system/sound/SoundInterface", "system/graphics/GraphicsInterface", "system/remotecontrol/RemoteControlInterface", "system/SystemInterface"], function(PowerInterface, SoundInterface, GraphicsInterface, RemoteControlInterface, SystemInterface) {
	/*
		System Settings Model - Provides a model for System Settings
		-------------------------------------------------------------------------
		This model uses a number of interfaces to populate its data including:
		- system/power/PowerInterface
		- system/sound/SoundInterface
		- system/graphics/GraphicsInterface
		- system/remotecontrol/RemoteControlInterface

		It originally used system/plugins/LFDPlugin for all of its methods but the data
		and functions exposed in this class have been split out into a number of Interface
		classes to provide a more modular approach to making low level panel functions and information
		available to the application and system reporting model.
	*/
	function SystemReport_SystemSettingsModel() {

		var self = this;

		//
		// Attempts to populate the system settings report model
		//
		self.generate = function() {
			var dfd = $.Deferred();

			// Get Picture Contrast
			logger.info("Getting contrast..");
			GraphicsInterface.getContrastRGB().done(function(data) {
				self.contrast = data;
			}).always(function() {

				// Get picture brightness
				logger.info("Getting brightness..");
				GraphicsInterface.getBrightnessRGB().done(function(data) {
					self.brightness = data;
				}).always(function() {
					getSLRotationSetting().done(function(slRotationValue) {
						// Only set if available
						if (slRotationValue) {
							self.signagelive_setup_rotation = slRotationValue;
						}
					}).always(function() {
						getNativePortraitSetting().done(function(nativePortraitRotationValue) {
							if (nativePortraitRotationValue) {
								self.native_portrait_rotation = nativePortraitRotationValue;
							}
						}).always(function() {
							// Get volume level
							logger.info("Getting volume..");
							SoundInterface.getVolume().done(function(data) {
								self.volume = data;
							}).always(function() {

								// Get mute status
								logger.info("Getting mute status...");
								SoundInterface.getMuteStatus().done(function(muteStatus) {
									self.mute_status = (muteStatus)
										? "muted"
										: "not muted";
								}).always(function() {

									// Get remote control status
									logger.info("Getting remote control status..");
									RemoteControlInterface.getRemoteControlStatus().done(function(data) {
										self.remote_control_status = data;
									}).always(function() {

										// Get saftey lock status
										logger.info("Getting safety lock status.");
										RemoteControlInterface.getSafetyLockStatus().done(function(data) {
											self.saftey_lock_status = data;
										}).always(function() {

											// Get button lock status
											logger.info("Getting button lock status..");
											RemoteControlInterface.getButtonLockStatus().done(function(data) {
												self.button_lock_status = data;
											}).always(function() {

												// Get screen on/off status
												logger.info("Getting screen on off status..");
												PowerInterface.getScreenOnOffStatus().done(function(data) {
													self.screen_status = data;
												}).always(function() {
													PowerInterface.getPMMode().done(function(data) {
														self.pm_mode = data;
													}).always(function() {
														// return FW version if we can
														logger.info("Getting firmware version...");
														if (SystemInterface.getFirmwareVersion != null) {
															SystemInterface.getFirmwareVersion().always(function(fwVersion) {
																self.firmware_version = fwVersion.toUpperCase();
																// Resolve system settings information generation
																logger.info("System settings info loaded");
																dfd.resolve();
															});
														} else {
															// Resolve system settings information generation
															logger.info("System settings info loaded");
															dfd.resolve();
														}
													});
												});
											});
										});
									});
								});
							});
						});
					});
				});
			});
			return dfd;
		}

		function getSLRotationSetting() {
			if (GraphicsInterface.getSignageliveRotation) {
				return GraphicsInterface.getSignageliveRotation();
			}
			else {
				return $.Deferred().resolve(undefined);
			}
		}

		function getNativePortraitSetting() {
			if (GraphicsInterface.getNativePortraitMode) {
				return GraphicsInterface.getNativePortraitMode();
			}
			else {
				return $.Deferred().resolve(undefined)
			}
		}
	}

	return SystemReport_SystemSettingsModel;
});
