/*
 * The DateTime plugin returns a singleton object that allows the caller to get the current system time, and parse a number of common
 * Javascript datetime string formats.
 * Please note that the Samsung SSP does not support timezones so all dates are returned GMT
 */
define(["system/SystemInterface", "properties/App.Settings", "controllers/UserConfigurationController"], function(SystemInterface, Settings, UserConfigurationController) {
    function DateTimePlugin() {
        var self = this;

        // Update the DateTime on the LG Panel
        self.setDateTime = function(dt, useNTP) {

            var dfd = $.Deferred();

            // Set the options
            var options = {};

            // Figure out if this is webOS 1.0 (if it is we use UTC time else we use local time)
            SystemInterface.getModel()
                .done(function(model) {
                    logger.debug("Panel Model: " + model);
                    // WEB OS 1.0
                    if (model.indexOf("LS") > -1 || model.indexOf("MS") > -1) {
                        logger.debug("WebOS 1.0 Panel found - saving time as UTC Time.");
                        options = getUTCTimeOptions(dt);
                    }
                    // WEB OS 2.0+
                    else {
                        logger.debug("WebOS 2.0 Panel found - saving time as Local Time.");   
                        useNTP = useNTP !== undefined ? useNTP : UserConfigurationController.useNTP();
                        options = getLocalTimeOptions(dt, useNTP);
                        logger.info('Setting DateTime: ' + JSON.stringify(options));
                    }

                    // Save datetime
                    setTime(options)
                        .done(function() {
                            dfd.resolve();
                        })
                        .fail(function() {
                            dfd.reject();
                        });

                })
                // Just default to local time otherwise
                .fail(function() {
                    // Set local options
                    logger.error("Error checking model while setting date time. Defaulting to Local time.");
                    options = getLocalTimeOptions(dt);

                    // Save datetime
                    setTime(options)
                        .done(function() {
                            dfd.resolve();
                        })
                        .fail(function() {
                            dfd.reject();
                        });
                });


            return dfd;
        };

        function getUTCTimeOptions(dt) {
            return {
                year: dt.getUTCFullYear(),
                month: dt.getUTCMonth() + 1,
                day: dt.getUTCDate(),
                hour: dt.getUTCHours(),
                minute: dt.getUTCMinutes(),
                sec: dt.getUTCSeconds()
            };
        }

        function getLocalTimeOptions(dt, useNTP) {
            return {
                year: dt.getFullYear(),
                month: dt.getMonth() + 1,
                day: dt.getDate(),
                hour: dt.getHours(),
                minute: dt.getMinutes(),
                sec: dt.getSeconds(),
                ntp: useNTP
            };
        }

        function setTime(options) {
            var dfd = $.Deferred();
            try
            {
                // Set Time
                logger.info("new Configuration.setCurrentTime()");
                new Configuration().setCurrentTime(
                    // Successfully Updated
                    function(successResult) {
                        logger.info("DateTime updated on panel successfully.");
                        dfd.resolve();
                    },
                    // Update Failed
                    function(errorResult) {
                        logger.error("There was an error updating the DateTime on the panel [" + errorResult.errorCode + "] " + errorResult.errorText);
                        dfd.reject();
                    },
                    options
                );
            }
            catch (error)
            {
                logger.error(error);
                dfd.reject(error);
            }

            return dfd;
        }


        // Get DateTime from LG Panel
        self.now = function() {
            /*var dfd = $.Deferred();
		     
		    // Set Time
		    new Configuration().getCurrentTime(
		        // Successfully Updated
		        function(successResult) {
		        	logger.debug("Successfully loaded current DateTime: " + JSON.stringify(successResult));
		        	var currentDateTime = new Date(successResult.year, successResult.month - 1, successResult.day, successResult.hour, successResult.minute, successResult.sec, 0);
		            dfd.resolve(currentDateTime);
		        },
		        // Update Failed
		        function(errorResult) {
		        	logger.error("Failed to get the current DateTime:  [" + errorResult.errorCode + "] " + errorResult.errorText);
		            dfd.reject();
		        }
		    );
		    return dfd;*/

            return new Date();
        };

        function ensure4PositionHex(hex) {
            if (hex.length == 1)
                hex = '000' + hex;

            if (hex.length == 2)
                hex = '00' + hex;

            if (hex.length == 3)
                hex = '0' + hex;

            return hex;
        }

        function ensure2PositionHex(hex) {
            if (hex.length == 1)
                hex = '0' + hex;
            return hex;
        }

        self.parse = function(string) {
            /*
             *Parses a date time string in the following formats to a Javascript datetime
             *
             *2013-06-03
             *2013-06-03T11:00:00
             *2013-06-04T09:07:29Z
             *9999-12-30T23:59:59.9999
             *2013-06-04T12:30:05.2926Z
             */

            var year = "1970";
            var month = "01";
            var day = "01";
            var hours = "00";
            var minutes = "00";
            var seconds = "00";
            var milliseconds = "0000";

            // has a date and a time been passed? - e.g. 2013-06-03T11:00:00
            var dateTimeComponents = splitDateTimeString(string);
            if (dateTimeComponents.length == 2) {
                var dateComponents = splitDateString(dateTimeComponents[0]);
                year = dateComponents[0];
                month = dateComponents[1];
                month = parseInt(month, 10) - 1;
                day = dateComponents[2];

                // we have a string in at least format - 2013-06-03T11:00:00
                var timeComponents = splitTimeString(dateTimeComponents[1]);
                hours = timeComponents[0];
                minutes = timeComponents[1];
                seconds = timeComponents[2];

                // check the seconds component as it may have milliseconds or a timezone attached to it
                if (seconds.length > 2) {
                    var secondsComponents = splitSecondsString(seconds);

                    // do we have a string in the format (for seconds only):
                    // 29Z or 59.9999 or 05.2926Z
                    // note that 00 is already handled by the check for a length greater than 2 places
                    if (secondsComponents.length == 2) {
                        // the format is either 59.9999 or 05.2926Z
                        // the first two places are the seconds
                        seconds = secondsComponents[0];
                    } else {
                        // the format is 29Z
                        // use the first two places as the seconds and ignore the timezone (SSSP does not interpret timezone)
                        seconds = seconds.substring(0, 2);
                    }
                }

            } else {
                // only a date e.g 2013-06-03
                var dateComponents = splitDateString(dateTimeComponents[0]);
                year = dateComponents[0];
                month = parseInt(dateComponents[1], 10) - 1;
                day = dateComponents[2];
            }

            return new Date(year, month, day, hours, minutes, seconds, milliseconds);
        };

        function splitDateTimeString(string) {
            return string.split('T');
        }

        function splitDateString(string) {
            return string.split('-');
        }

        function splitTimeString(string) {
            return string.split(':');
        }

        function splitSecondsString(string) {
            return string.split(".");
        }

    }

    Date.prototype.getStringRepresentation = function() {
        var month = this.getMonth() + 1;
        if (month < 10)
            month = "0" + month;

        var day = this.getDate();
        if (day < 10)
            day = "0" + day;

        var hour = this.getHours();
        if (hour < 10)
            hour = "0" + hour;

        var minutes = this.getMinutes();
        if (minutes < 10)
            minutes = "0" + minutes;

        var seconds = this.getSeconds();
        if (seconds < 10)
            seconds = "0" + seconds;

        return this.getFullYear() + "-" + month + "-" + day + "T" + hour + ":" + minutes + ":" + seconds + "Z";
    };

    Date.prototype.addDays = function(n) {
        var newDate = new Date(this);
        newDate.setDate(newDate.getDate() + n);
        return newDate;
    };
    /*
     *Adds the number of hours to the specified date and returns a NEW date object
     */
    Date.prototype.addHours = function(h) {
        var newDate = new Date(this);
        newDate.setHours(newDate.getHours() + h);
        return newDate;
    };

    Date.prototype.addMinutes = function(n) {
        var newDate = new Date(this);
        newDate.setMinutes(newDate.getMinutes() + n);
        return newDate;
    };

    Date.prototype.addSeconds = function(n) {
        var newDate = new Date(this);
        newDate.setSeconds(newDate.getSeconds() + n);
        return newDate;
    };
    /*
     * Returns the difference in milliseconds between the two dates
     */
    Date.prototype.subtract = function(dt) {
        return this - dt;
    };
    /*
     * Accepts a time in the format HH:mm:ss and returns a NEW date object
     *
     */
    Date.prototype.setTime = function(time) {
        var timeParts = time.split(':');
        var newDate = new Date(this);
        newDate.setHours(timeParts[0]);
        newDate.setMinutes(timeParts[1]);
        newDate.setSeconds(timeParts[2]);

        return newDate;
    };

    return new DateTimePlugin();
});