define(["properties/App.Settings", "system/SystemInterface", "models/LocalFileList", "system/power/PowerInterface"], function(Settings, SystemInterface, LocalFileList, PowerInterface) {
    function SystemInitializer() {
        var self = this;
        self.webOSVersion = null;

        var WEBOS_VERIONS = {
            '1.0': 1,
            '2.0': 2,
            '3.0': 3,
            '3.2': 3.2,
            '4.0': 4,
            '4.1': 4.1,
            '5.0': 5,
            '6.0.0': 6
        };

        var SCAP_1_1_Library = {
            'path': './scripts/cordova-cd/1.1/',
            'versionString': '1.1',
            'version': 1.1,
            'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js']
        };

        var SCAP_LIBRARIES = {
            '1.0': {
                'path': './scripts/cordova-cd/1.2/',
                'versionString': '1.2',
                'version': 1.2,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js']
            },
            '2.0': {
                'path': './scripts/cordova-cd/1.3/',
                'versionString': '1.3',
                'version': 1.3,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js']
            },
            '3.0': {
                'path': './scripts/cordova-cd/1.4.4/',
                'versionString': '1.4.4',
                'version': 1.4,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js']
            },
            '3.2': {
                'path': './scripts/cordova-cd/1.5.7/',
                'versionString': '1.5.7',
                'version': 1.5,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js']
            },
            '4.0': {
                'path': './scripts/cordova-cd/1.6.3/',
                'versionString': '1.6.3',
                'version': 1.6,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js']
            },
            '4.1': {
                'path': './scripts/cordova-cd/1.7.1/',
                'versionString': '1.7.1',
                'version': 1.7,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js', 'iot.js']
            },
            '5.0': {
                'path': './scripts/cordova-cd/1.7.1/',
                'versionString': '1.7.1',
                'version': 1.7,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js', 'iot.js']
            },
            '6.0.0': {
                'path': './scripts/cordova-cd/1.8/',
                'versionString': '1.8.0',
                'version': 1.8,
                'files': ['configuration.js', 'deviceInfo.js', 'inputSource.js', 'power.js', 'signage.js', 'sound.js', 'storage.js','video.js', 'security.js', 'time.js', 'utility.js', 'iot.js']
            }
        }

        self.initialize = function() {
            var dfd = $.Deferred();
            // Load model details and correct scap version
            SystemInterface.getOSVersion()
                .done(function(webOSVersion) {
                    self.webOSVersion = webOSVersion;
                    // Check if this is a known version. If it is not then it could be a new generation so default to 4.1/5.0 so we load the latest libraries
                    switch(webOSVersion) {
                        // If webOS 1 then gapless playback and remote firmware updates are not supported
                        case '1.0':
                            Settings.application.api.useLegacyTlsUrl = true;
                            Settings.application.features.supportsGaplessPlayback = false;
                            Settings.application.features.canPerformFirmwareUpdates = false;
                            Settings.application.features.supportsVideoRotation = false;
                        // Known versions
                        case '2.0':
                            Settings.application.api.useLegacyTlsUrl = true;
                            Settings.application.features.supportsVideoRotation = false;
                        case '3.0':
                        case '3.2':
                        case '4.0':
                        case '4.1':
                        case '5.0':
                        case '6.0.0':
                            // Do nothing
                            break;
                        case 'Unknown':
                            logger.warning('An unexepected error occurred getting the webOS version. This is known to happen on webOS 3/3.2 panels. Defaulting webOS version to 3.0.');
                            webOSVersion = '3.0';
                        default:
                            var webOSVersionAsFloat = parseFloat(webOSVersion);
                            var webOSVersionIsNewerThanWebOS5 = webOSVersionAsFloat > 5.0;
                            if (webOSVersionIsNewerThanWebOS5) {
                                logger.warning('WebOS version is greater than 5.0. This may be a newer generation panel. Recommended SCAP version is not known. It is advised that the media player is updated to add support for the latest generation. Falling back to SCAP API for WebOS 4.1/5.0.');
                                webOSVersion = '5.0';
                            }

                    }
                    // Load the SCAP Library for the webOS version
                    logger.info('Loading SCAP API for WebOS version: ' + webOSVersion);
                    loadScapLibraryForWebOSVersion(webOSVersion)
                        .done(function() {
                            // Validate the SCAP version works and downgrade if neccessary
                            validateScapVersionAndDowngradeIfNeccessary()
                                .done(function() {
                                    // Load server settings and model details
                                    logger.info('Loading server settings and model details...');
                                    $.when(loadServerSettings(), loadModelDetails(), checkTLS12Support(webOSVersion))
                                        .done(function() {
                                            logger.info('Server settings and model details loaded.');
                                            // Update app settings file paths for IPK apps
                                            updateApplicatonSettingFilesPaths();
                                            // Check PM Mode
                                            // checkPMMode(webOSVersion);
                                            // Disable Portrait OSD (for BH5C only)
                                            disablePortraitOSDMode();
                                            // Initialize LocalFileList
                                            logger.info('loading local file list...');
                                            LocalFileList.initialize().always(function() {
                                                logger.info('Local file list initialized. System initialization completed sucessfully.');
                                                dfd.resolve();
                                            });
                                        })
                                        .fail(function() {
                                            logger.fatal('Unable to load server settings or model data. Fatal error, can not continue.');
                                            dfd.reject();
                                        });
                                })
                                .fail(function(err) {
                                    logger.fatal('Unable to validate SCAP version: ' + err);
                                    dfd.reject();
                                })
                        })
                        .fail(function() {
                            logger.fatal('Unable to load SCAP library. Fatal error, can not continue.');
                            dfd.reject();
                        });
                });
            return dfd;
        }

        function loadScapLibraryForWebOSVersion(webOSVersion) {
            var dfd = $.Deferred();
            // Get Scap library object based on webOS version mapping
            var scapLibrary = SCAP_LIBRARIES[webOSVersion];
            // If not found then log error and return
            if (scapLibrary == null) {
                logger.error('Unable to load scap library for webOS version: ' + webOSVersion);
                return dfd.reject();
            }
            // Update settings properties
            Settings.application.sdkVersion = scapLibrary.version;
            Settings.application.webOSVersion = WEBOS_VERIONS[webOSVersion];
            // Load library by creating scripts and adding to DOM
            logger.info('Loading SCAP API ' + scapLibrary.versionString + ' for webOS ' + webOSVersion);
            var scriptsLoaded = 0;
            for (var i = 0; i <= scapLibrary.files.length; i++) {
                var jsPath = scapLibrary.path + scapLibrary.files[i];
                var script = document.createElement('script');            
                script.src = jsPath;
                document.head.appendChild(script);
                // Listen for the loaded event for each script.
                // Track the number of scripts loaded and resolve once all have loaded.
                script.onload = function() {
                    logger.info('SCAP API script loaded.');
                    scriptsLoaded++;
                    if (scriptsLoaded === scapLibrary.files.length) {
                        logger.info('All SCAP files loaded.');
                        dfd.resolve();
                    }
                }
            }
            return dfd;
        }

        function loadServerSettings() {
            var dfd = $.Deferred();
            logger.debug("Loading Server Setting details..");
            SystemInterface.getServerProperties().done(function(properties) {
                if (properties && properties.appType) {
                    try {
                        Settings.application.packageType = properties.appType.toUpperCase();
                        dfd.resolve();
                    } catch (e) {
                        // Default to IPK
                        logger.warning('Unable to load package type. Defaulting to IPK.');
                        Settings.application.packageType = 'IPK';
                        dfd.resolve();
                    }
                }
                else {
                    // Default to IPK
                    logger.warning('Unable to load package type. Defaulting to IPK.');
                    Settings.application.packageType = 'IPK';
                    dfd.resolve();
                }
            }).fail(function() {
                logger.error('Unable to load server settings');
                dfd.reject();
            });
            return dfd;
        }

        function loadModelDetails() {
            var dfd = $.Deferred();
            logger.debug("Loading model details..");
            // Load the panel model
            SystemInterface.getModel()
                .done(function(model) {
                    logger.info('Model details: ' + model);
                    Settings.application.model = model;
                    dfd.resolve('model');
                })
                .fail(function() {
                    logger.error("Error detecting model details");
                    dfd.resolve('Unknown');
                });
            return dfd;
        }

        function checkTLS12Support(webOSVersion) {
            var dfd = $.Deferred();
            isTLS12Supported(webOSVersion)
                .done(function(supported) {
                    Settings.application.features.supportsTLS12 = supported;
                    Settings.application.api.useLegacyTlsUrl = (supported === false);
                    dfd.resolve();
                });
            return dfd;
        }

        self.checkPMMode = function() {
            var dfd = $.Deferred();
            isPMModeSettingsSupported(self.webOSVersion)
            .done(function(supported) {
                if(supported && Settings.user.settingsLoaded && Settings.user.mySettings.usePMModeForScreenManagement) {
                    PowerInterface.getPMMode().then(function(mode) {
                        if(mode !== "screenOffAlways") {
                            PowerInterface.setPMMode().then(function() {
                                logger.info("PM Mode set to screenOffAlways");
                                dfd.resolve();
                            });
                        } else {
                            dfd.resolve();
                        }
                    });
                } else {
                    dfd.resolve();
                }
            });
            return dfd;
        };

        function isPMModeSettingsSupported(webOSVersion) {
            var dfd = $.Deferred();
            logger.info('Checking if PM Mode is supported');
            switch (webOSVersion) {
                case "1.0":
                case "2.0":
                case "3.0":
                case "3.2":
                case "4.0":
                    // WebOS 1 and 2 do not support this PM option
                    // WebOS 3.0 does not support addKeyItem for RC functionality
                    // WebOS 3.2 and 4.0 do not emit a keycode when using the RC to turn the screen off - keycode is also 1537 (undocumented) instead of 409
                    logger.info('WebOS ' + webOSVersion + ' panel - PM Mode setting not supported.');
                    return dfd.resolve(false);
                default:
                    // webOS 4.1+ does
                    logger.info('WebOS ' + webOSVersion + ' panel. PM Mode setting supported.');
                    return dfd.resolve(true);
            }
        };

        function isTLS12Supported(webOSVersion) {
            var dfd = $.Deferred();
            logger.info('Checking TLS 1.2 Supported');
            switch (webOSVersion) {
                case "1.0":
                case "2.0":
                    // WebOS 1 and 2 do not support TLS 1.2
                    logger.info('WebOS ' + webOSVersion + ' panel - TLS 1.2 not supported.');
                    return dfd.resolve(false);
                default:
                    // All WebOS 3.0+ displays supports TLS 1.2
                    logger.info('WebOS ' + webOSVersion + ' panel. TLS 1.2 supported.');
                    return dfd.resolve(true);
            }
        }

        function convertFirmwareVersionFromStringToNumber(fwVersionString) {
			try {
				// String is in the format 03.00.90
				// This always works because LG never break this format
				var fwNumber = parseInt(fwVersionString.split('.').join(""));
				return fwNumber;
			}
			catch (e) {
				logger.error("Error converting version string to number: " + e.message);
				return null;
			}
		}

        function disablePortraitOSDMode() {
            if (Settings.application.model == 'BH5C') {
                logger.info("Disabling portrait OSD for BH5C...");
                SystemInterface.disableOSDPortraitMode();
            }
        }

        function updateApplicatonSettingFilesPaths() {
            // Update settings file paths if required
            if (Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                Settings.application.fileSystem.dataDirectory = "data";
                Settings.application.fileSystem.systemFiles.settingsFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.settingsFilePath;
                Settings.application.fileSystem.systemFiles.propertiesFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.propertiesFilePath;
                Settings.application.fileSystem.systemFiles.lastConfigFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.lastConfigFilePath;
                Settings.application.fileSystem.systemFiles.clientIdFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.clientIdFilePath;
                Settings.application.fileSystem.systemFiles.clientKeyFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.clientKeyFilePath;
                Settings.application.fileSystem.systemFiles.appVersionFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.appVersionFilePath;
                Settings.application.fileSystem.systemFiles.fileListFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.fileListFilePath;
                Settings.application.fileSystem.systemFiles.fileListUpgradeFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.fileListUpgradeFilePath;
                Settings.application.fileSystem.systemFiles.proofOfPlayFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.proofOfPlayFilePath;
                Settings.application.fileSystem.systemFiles.proofOfPlayUploadFilePath = Settings.application.fileSystem.root + "/" + Settings.application.fileSystem.systemFiles.proofOfPlayUploadFilePath;
            }
        }
        // I had initially hoped we could remove this when we switched to using CustonJS to get the webOS version
        // and mapped that to the recommended SCAP version. However, we have run into some issues with certain older models
        // like webOS 1.0 XS2B so we have added this SCAP validation code back in.
        function validateScapVersionAndDowngradeIfNeccessary() {
            if (Settings.application.sdkVersion <= 1.4) {
                var dfd = $.Deferred();
                logger.info('Validating SCAP Support..');
                // Validate SCAP 1.4
                validateScap14Support().done(function(scap14Supported) {
                    // SCAP 1.4 Supported
                    if (scap14Supported) {
                        logger.info('SCAP 1.4 Support Validated. Downgrade not neccessary.');
                        dfd.resolve();
                    }
                    // SCAP 1.4 Not supported - downgrade and test
                    else {
                        logger.info('Loading SCAP 1.3 library...');
                        loadScapLibrary('1.3', 1.3).done(function() {
                            logger.info('SCAP 1.3 loaded. Validating...');
                            // Validate SCAP 1.3
                            validateScap13Support().done(function(scap13Supported) {
                                // SCAP 1.3 Validated
                                if (scap13Supported) {
                                    logger.info('SCAP 1.3 Validated!');
                                    dfd.resolve();
                                }
                                // SCAP 1.3 validation failed, downgrade to 1.2 and validate
                                else {      
                                    logger.info('Loading SCAP 1.2 library...');
                                    loadScapLibrary('1.2', 1.2).done(function() {
                                        logger.info('SCAP 1.2 loaded. Validating...');
                                        // Validate SCAP 1.2
                                        validateScap12Support().done(function(scap12Supported) {
                                            // SCAP 1.2 Validated
                                            if (scap12Supported) {
                                                logger.info('SCAP 1.2 Validated!');
                                                dfd.resolve();
                                            }
                                            // SCAP 1.2 validation failed, downgrade to 1.1 and validate
                                            else {
                                                logger.info('Loading SCAP 1.1 library...');
                                                loadScapLibrary('1.1', 1.1).done(function() {
                                                    logger.info('SCAP 1.1 loaded. Validating...');
                                                    validateScap11Support().done(function(scap11Supported) {
                                                        // SCAP 1.1 Validated
                                                        if (scap11Supported) {
                                                            logger.info('SCAP 1.1 Validated!');
                                                            dfd.resolve();
                                                        }
                                                        else {
                                                            logger.fatal('Failed to validate SCAP 1.1');
                                                            dfd.reject();
                                                        }
                                                    });
                                                }).fail(function() {
                                                    logger.fatal('Failed to load SCAP 1.1');
                                                    dfd.reject();
                                                });
                                            }
                                        });
                                    }).fail(function() {
                                        logger.fatal('Failed to load SCAP 1.2');
                                        dfd.reject();
                                    });
                                }
                            });
                        }).fail(function() {
                            logger.fatal('Failed to load SCAP 1.3');
                            dfd.reject();
                        });
                    }
                });
                return dfd;
            }
            else {
                logger.info('SCAP Validation not required');
                return $.Deferred().resolve();
            }
        }
        // Validate SCAP 1.4 support by calling getEddyStoneInfo() which was added in 1.4
        // This should fail if 1.4 is not supported.
        function validateScap14Support() {
            if (Settings.application.sdkVersion < 1.4) {
                logger.info('Current SCAP version is less than 1.4. SCAP 1.4 not supported.');
                return $.Deferred().resolve(false);
            }
            else {
                var dfd = $.Deferred();
                logger.info("Validating SCAP 1.4 Support...");
                try {
                    new DeviceInfo().getEddystoneInfo(
                        function() {
                            logger.info("Successfully loaded. SCAP 1.4 SUPPORTED!");
                            dfd.resolve(true);
                        }, 
                        function(cbObject) {
                            logger.warning(" Error Code [" + cbObject.errorCode + "]: " + cbObject.errorText + ". SCAP 1.4 NOT SUPPORTED");
                            dfd.resolve(false);
                        });
                } catch (e) {
                    logger.warning("Error: " + e.message + ". SCAP 1.4 NOT SUPPORTED");
                    dfd.resolve(false);
                }
                return dfd;
            }
            
        }
        // Validate SCAP 1.3 support by calling getTimeZone() which was added in 1.3
        // This should fail if 1.3 is not supported.
        function validateScap13Support() {
            var dfd = $.Deferred();
            logger.info("Validating SCAP 1.3 Support...");
            try {
               new Configuration().getTimeZone(
                    function() {
                        logger.info("SCAP 1.3 SUPPORTED");
                        dfd.resolve(true);
                    },
                    function() {
                        logger.warning("Error. SCAP 1.3 NOT SUPPORTED");
                        dfd.resolve(false);
                    });

            } catch (e) {
                logger.warning("Error: " + e.message + ". SCAP 1.3 NOT SUPPORTED");
                dfd.resolve(false);
            }
            return dfd;
        }
        // Validate SCAP 1.2 support by attempting to write a file, which was added in 1.2
        // This should fail if 1.2 is not supported.
        function validateScap12Support() {
            var dfd = $.Deferred();
            logger.info("Validating SCAP 1.2 Support...");
            try {
                // Write Text data, use utf-8 encoding, write all text.
                var textData = "Hello SCAP!!!!!";
                new Storage().writeFile(
                    function(cbObject) {
                        logger.info("Successfully writen " + cbObject.written + " bytes. SCAP 1.2 SUPPORTED!");
                        dfd.resolve(true);
                    },
                    function(cbObject) {
                        logger.warning(" Error Code [" + cbObject.errorCode + "]: " + cbObject.errorText + ". SCAP 1.2 NOT SUPPORTED");
                        dfd.resolve(false);
                    },
                    {
                        data: textData,
                        path: 'file://internal/test.txt',
                        position: 0,
                        mode: 'truncate',
                        offset: 0,
                        length: textData.length,
                        encoding: 'utf8'
                    });

            } catch (e) {
                logger.warning("Error: " + e.message + ". SCAP 1.2 NOT SUPPORTED");
                dfd.resolve(false);
            }
            return dfd;
        }
        // Validate SCAP 1.1 support by attempting to get server settings, which was added in 1.1
        // This should fail if 1.1 is not supported.
        function validateScap11Support() {
            var dfd = $.Deferred();
            logger.info("Validating SCAP 1.1 Support...");
            try {
                new Configuration().getServerProperty(
                    function() {
                        logger.info("Successfully loaded server properties. SCAP 1.1 SUPPORTED!");
                        dfd.resolve(true);
                    },
                    function(cbObject) {
                        logger.warning(" Error Code [" + cbObject.errorCode + "]: " + cbObject.errorText + ". SCAP 1.1 NOT SUPPORTED");
                        dfd.resolve(false);
                    });

            } catch (e) {
                logger.warning("Error: " + e.message + ". SCAP 1.1 NOT SUPPORTED");
                dfd.resolve(false);
            }
            return dfd;
        }
        // This method is used to downgrade the loaded SCAP version if it is determined
        // that the current player does not support the currently loaded version
        function loadScapLibrary(version, sdkVersion) {
            if (Settings.application.sdkVersion === sdkVersion) {
                logger.info('Attempting to load SCAP version: ' + version + ' but this is already the current version')
                return $.Deferred().resolve();
            }
            var dfd = $.Deferred();
            logger.info('Loading SCAP version ' + version);
            try {
                // Remove currently configured cordova plugins
                cordova.define.remove("cordova/plugin/signage");
                cordova.define.remove("cordova/plugin/storage");
                cordova.define.remove("cordova/plugin/configuration");
                cordova.define.remove("cordova/plugin/deviceInfo");
                cordova.define.remove("cordova/plugin/inputSource");
                cordova.define.remove("cordova/plugin/power");
                cordova.define.remove("cordova/plugin/sound");
                cordova.define.remove("cordova/plugin/video");
                cordova.define.remove("cordova/plugin/security");
                cordova.define.remove("cordova/plugin/time");
                cordova.define.remove("cordova/plugin/utility");
                cordova.define.remove("cordova/plugin/broadcast");
                // Set Global Window Vars to NULL
                Signage = null;
                Storage = null;
                Configuration = null;
                DeviceInfo = null;
                InputSource = null;
                Power = null;
                Sound = null;
                Video = null;
                Utility = null;
                // Now load all scripts for this version using $.getScript
                $.when(
                    $.getScript("scripts/cordova-cd/" + version + "/signage.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/storage.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/configuration.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/deviceInfo.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/inputSource.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/power.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/sound.js"),
                    $.getScript("scripts/cordova-cd/" + version + "/video.js")
                ).done(function() {
                    logger.info('SCAP Version ' + version + ' loaded.');
                    Settings.application.sdkVersion = sdkVersion;
                    dfd.resolve();
                })
                .fail(function(jqxhr, settings, exception) {
                    logger.error("Error loading SCAP library: " + exception);
                    dfd.reject();
                });
            } catch (e) {
                logger.error("Error loading SCAP library: " + e);
                dfd.reject();
            }
            return dfd;
        }
    }
    return SystemInitializer;
});