define(["models/FirmwareUpdateStatusConstants","properties/App.Settings"], function(FirmwareUpdateStatusConstants, Settings) {
    /*
    	SystemInterface provides access to low level System Information
     */
    function SystemInterface() {

        var self = this;

        /**
         * Downloads a valid firmware file from the url defined in the options passed to the method
         * @param  {Object} options The only option is the url of the firmware file (options.url = "")
         * @return {JQuery Deffered}         resolves on success, rejected on failure
         */
        self.downloadFirmware = function(options) {
            var dfd = $.Deferred();

            logger.info("Beginning remote firmware download process...");

            new Storage().downloadFirmware(
                // success callback
                function() {
                    logger.info("New remote firmware download successful");
                    dfd.resolve();
                },
                // failure callback
                function(e) {
                    logger.error("Error downloading remote firmware: [" + e.errorCode + "] " + e.errorText);
                    dfd.reject(e);
                },
                options
            );

            return dfd;
        };

        /**
         * Upgrades firmware installed on the panel using the firmware file that has been downloaded.
         * Must be called after self.downloadFirmware(). If the panel is shutdown or restarted after a download but before this
         * method is called the firmware file will get deleted and will need to be redownlaoded
         * @return {JQuery Deffered}         resolves on success, rejected on failure
         */
        self.upgradeFirmware = function() {
            var dfd = $.Deferred();

            logger.info("Beginning firmware upgrade process...");

            new Storage().upgradeFirmware(
                // success callback
                function() {
                    logger.info("New firmware upgrade process started successfully.");
                    dfd.resolve();
                },
                // failure callback
                function(e) {
                    logger.error("Error upgrading to new firmware: [" + e.errorCode + "] " + e.errorText);
                    dfd.reject(e);
                }
            );

            return dfd;
        };

        /**
         * Returns the firmware upgrade status:
         * Idle, Downlaoding, Ready, In Progress, Completed, Fail, Unknown & Download progress and Install Progess
         * Unknown is returned if the method call fails for any reason.
         * @return {JQuery Deffered}         resolves status on success, resolves Unknown on failure.
         */
        self.getFirmwareUpgradeStatus = function() {
            var dfd = $.Deferred();

            logger.info("Checking firmware upgrade status...");

            new Storage().getFirmwareUpgradeStatus(
                // success callback
                function(result) {
                    logger.info("Firmware Upgrade Status: " + result.status);

                    // resolve status
                    dfd.resolve(result);
                },
                // failure callback
                function(e) {
                    logger.error("Error getting latest firmware update status: [" + e.errorCode + "] " + e.errorText);
                    result.status = FirmwareUpdateStatusConstants.UNKNOWN;
                    dfd.resolve(result);
                }
            );

            return dfd;
        };

        self.getSystemName = function() {

            var dfd = $.Deferred();

            setTimeout(function() {
                var systemName = "LG WebOS Signage";
                logger.info("System Interface: System Name: " + systemName);
                dfd.resolve(systemName);
            }, 100);

            return dfd;
        };

        self.getOSName = function() {

            var dfd = $.Deferred();

            setTimeout(function() {
                var OSName = "WebOS";
                logger.info("System Interface: OS Name: " + OSName);
                dfd.resolve(OSName);
            }, 100);

            return dfd;
        };

        self.getOSVersion = function() {
            var dfd = $.Deferred();
            var custom = new Custom();
            custom.Signage.getwebOSVersion(
                function successCallback(successObject) {
                    var webOSVersion = successObject.webOSVersion;
                    switch(webOSVersion) {
                        // Known versions
                        case '1.0':
                        case '2.0':
                        case '3.0':
                        case '3.2':
                        case '4.0':
                        case '4.1':
                        case '5.0':
                            logger.info('webOS Signage version: ' + webOSVersion);
                            dfd.resolve(webOSVersion); // 1.0, 2.0, 3.0, 3.2, 4.0, 4.1, 5.0
                            break;
                        default:
                            logger.warning('Unknown webOS Version: ' + webOSVersion);
                            dfd.resolve(webOSVersion); // 1.0, 2.0, 3.0, 3.2, 4.0, 4.1, 5.0
                    }
                },
                function failureCallback(failureObject) {
                    logger.error('Error loading webOS version using CustomJS: [' + failureObject.errorCode + ']' + failureObject.errorText);
                    try {
                        idcap.request( "idcap://configuration/property/get", {
                            "parameters": {
                                "key" : "webos_version"
                            },
                            "onSuccess": function(result) {
                                logger.info("IDCAP WebOS Version Result = " + JSON.stringify(result));
                                dfd.resolve(result.value);
                            },
                            "onFailure": function (err) {
                                logger.error("onFailure : errorMessage = " + err.errorMessage);
                                dfd.resolve('Unknown');
                            }
                        });
                    }
                    catch (error) {

                    }
                }
            );
            return dfd;
        };

        self.getManufacturer = function() {

            var dfd = $.Deferred();

            getPlatformInfo()
                .done(function(platformInfo) {
                    logger.info("System Interface: Manufacturer: " + platformInfo.manufacturer);
                    dfd.resolve(platformInfo.manufacturer);
                })
                .fail(function() {
                    dfd.reject();
                });

            return dfd;
        };

        self.getModel = function() {

            var dfd = $.Deferred();

            getPlatformInfo()
                .done(function(platformInfo) {
                    var model = platformInfo.modelName.split("-")[0];
                    logger.info("System Interface: Model Name: " + model);
                    dfd.resolve(model);
                })
                .fail(function(e) {
                    logger.error("Error getting model: [" + e.errorCode + "] " + e.errorText);
                    dfd.reject();
                });

            return dfd;
        };

        self.getSystemUpTime = function() {
            var dfd = $.Deferred();
            getUsageData()
                .done(function(usageData) {
                    logger.info("System Interface: System Uptime: " + usageData.uptime + " hours");
                    dfd.resolve(usageData.uptime);
                })
                .fail(function() {
                    dfd.reject();
                });
            return dfd;
        };

        self.getTimeZone = function() {
            var dfd = $.Deferred();
            setTimeout(function() {
                logger.warning("System Interface: Error: Get Time Zone Not Implemented.");
                dfd.reject();
            }, 100);
            return dfd;
        };

        self.getUTCOffset = function() {
            var dfd = $.Deferred();
            setTimeout(function() {
                logger.warning("System Interface: Error: Get UTC Offset Not Implemented.");
                dfd.reject();
            }, 100);
            return dfd;
        };

        self.switchOffOSD = function() {
            var dfd = $.Deferred();
            setTimeout(function() {
                logger.warning("System Interface: Error: Switch Off OSD Not Implemented.");
                dfd.reject();
            }, 100);
            return dfd;
        };

        /**
         * [getSDKVersion description]
         * @return {[type]} [description]
         */
        self.getSDKVersion = function() {
            var dfd = $.Deferred();

            getPlatformInfo()
                .done(function(platformInfo) {
                    logger.info("System Interface: SDK Version: " + platformInfo.sdkVersion);
                    dfd.resolve(platformInfo.sdkVersion);
                })
                .fail(function() {
                    dfd.reject();
                });

            return dfd;
        };

        /**
         * Resolves the firmware version of the client
         * @return {[type]} [description]
         */
        self.getFirmwareVersion = function() {
            var dfd = $.Deferred();

            getPlatformInfo()
                .done(function(platformInfo) {
                    logger.info("System Interface: Firmware Version: " + platformInfo.firmwareVersion);
                    dfd.resolve(platformInfo.firmwareVersion);
                })
                .fail(function() {
                    logger.error("Error getting firmware version: [" + e.errorCode + "] " + e.errorText);
                    dfd.reject();
                });

            return dfd;
        };

        // Checked Supported HTML5 features using Modernizr (https://modernizr.com/docs)
        self.checkSupportedHTML5Features = function (featuresToTest) {
            // Check Modernizr is defined
            if (!Modernizr) {
                return $.Deferred.resolve({ "error": "Modernizr undefined. Unable to check features" });
            }
            // Check we have been passed an array of features to test
            if (Array.isArray(featuresToTest) === false) {
                return $.Deferred.resolve({ "error": "Invalid parameter. Array of features to test expected." });
            }
            else {
                // Iterate over features and build a response object
                var response = {};
                $.each(featuresToTest, function (index, feature) {
                    var property = feature + "_supported";
                    response[property] = Modernizr[feature];
                });
                return $.Deferred().resolve(response);
            }
        };

        // Check Max Supported TLS Version
        var cachedSSLDetails = null;
        self.getMaxSupportedTLSVersion = function () {
            // If we already have a cached response then just return that
            if (cachedSSLDetails) {
                return $.Deferred().resolve(cachedSSLDetails);
            }
            // New deferred result to retrun
            var dfd = $.Deferred();
            // Start building result object
            if(Settings.application.features.supportsTLS12) {
                var result = {
                    "tls_version_string": Settings.application.ssl.maxSupportedTLSVersionString,
                    "tls_version": Settings.application.ssl.maxSupportedTLSVersion
                };
            } else {
                var result = {
                    "tls_version_string": Settings.application.ssl.legacyMaxSupportedTLSVersionString,
                    "tls_version": Settings.application.ssl.legacyMaxSupportedTLSVersion
                };
            }
            // Otherwise attempt to call out to https://www.howsmyssl.com/a/check to get detailed SSL info
            callSSLAPICheck()
                .done(function (detailedSSLResults) {
                    result.detailed_tls_settings_included = true;
                    result.detailed_tls_settings = detailedSSLResults;
                    // Cache result
                    cachedSSLDetails = result;
                    // Return detailed results
                    logger.info('Returning detailed TLS Info: ' + JSON.stringify(result, undefined, 2));
                    dfd.resolve(result);
                })
                .fail(function () {
                    result.detailed_tls_settings_included = false;
                    result.detailed_tls_settings = null;
                    logger.info('Returning basic TLS Info: ' + JSON.stringify(result, undefined, 2));
                    dfd.resolve(result);
                });

            return dfd;
        };

        function callSSLAPICheck() {
            var dfd = $.Deferred();
            try {
                logger.info('Calling out to Hows My SSL to check SSL details: ' + Settings.application.ssl.sslCheckAPIUrl);
                // Attempt to get detailed SSL info from HowsMySSL
                $.ajax({
                    url: Settings.application.ssl.sslCheckAPIUrl,
                    type: 'GET',
                    contentType: 'application/json',
                    timeout: 60000,
                })
                    .done(function (data, textStatus, jqXHR) {
                        logger.info('SSL API call successful: ' + textStatus);
                        dfd.resolve(data);
                    })
                    .fail(function (jqXHR, textStatus, errorThrown) {
                        logger.error('Error calling out to SSL API: ' + textStatus);
                        dfd.reject();
                    });
            } catch (e) {
                logger.error('Error calling out to SSL API: ' + e);
                dfd.reject();
            }
            return dfd;
        }
        
        self.getOSDPortraitMode = function() {
            var dfd = $.Deferred();

            getSignageInfo()
                .done(function(signageInfo) {
                    dfd.resolve(signageInfo.portraitMode);
                })
                .fail(function() {
                    dfd.reject();
                });

            return dfd;
        }

        self.disableOSDPortraitMode = function() {
            var dfd = $.Deferred();

            logger.info("System Interface: DISABLING OSD...");

            var options = {
                portraitMode: Signage.OsdPortraitMode.OFF
            };

            new Signage().setPortraitMode(
                function(result) {
                    // Success
                    logger.info("System Interface: OSD Potrait disabled");
                    dfd.resolve(result);
                },
                function(result) {
                    // Failed
                    logger.error("System Interface: Error disabling OSD Potrait mode. Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        }

        self.enableOSDPortraitMode = function() {
            var dfd = $.Deferred();

            var options = {
                portraitMode: Signage.OsdPortraitMode.ON
            };

            new Signage().setPortraitMode(
                function(result) {
                    // Success
                    logger.info("System Interface: OSD Potrait enabled");
                    dfd.resolve(result);
                },
                function(result) {
                    // Failed
                    logger.error("System Interface: Error enabling OSD Potrait mode. Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        }

        self.setIDCAPProperty = function(key, value) {
            var dfd = $.Deferred();
            idcap.request( "idcap://configuration/property/set" , {
                "parameters": {
                    "key" : key,
                    "value": value
                },
                "onSuccess": function (cbObject) {
                    logger.info("setIDCAPProperty success = " + JSON.stringify(cbObject));
                    dfd.resolve();
                },
                "onFailure": function (err) {
                    logger.error("setIDCAPProperty failed - error = " + err.errorMessage);
                    dfd.reject();
                }
            });
            return dfd;
        };

        self.getIDCAPProperty = function(key) {
            var dfd = $.Deferred();
            idcap.request( "idcap://configuration/property/get" , {
                "parameters": {
                    "key" : key
                },
                "onSuccess": function (cbObject) {
                    logger.info("getIDCAPProperty = " + JSON.stringify(cbObject));
                    dfd.resolve(cbObject.value);
                },
                "onFailure": function (err) {
                    logger.error("getIDCAPProperty failed - error = " + err.errorMessage);
                    dfd.reject();
                }
            });
            return dfd;
        };

        self.getServerProperties = function() {
            var dfd = $.Deferred();
            logger.info('Loading server properties...');
            new Configuration().getServerProperty(
                function(properties) {
                    logger.info('Server properties loaded');
                    dfd.resolve(properties);
                },
                function(err) {
                    var errorCode = err.errorCode;
                    var errorText = err.errorText;
                    logger.error("Error Code [" + errorCode + "]: " + errorText);
                    dfd.reject();
                }
            );
            return dfd;
        };

        //
        // Returns platform information including model, manufacturer, firmware version, sdk version, serialno
        //
        function getPlatformInfo() {
            var dfd = $.Deferred();

            new DeviceInfo().getPlatformInfo(
                function(result) {
                    // Success
                    logger.info("System Interface: Platform Info Loaded");
                    dfd.resolve(result);
                },
                function(result) {
                    // Failed
                    logger.error("System Interface: Error getting platform information. Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                }
            );

            return dfd;
        }

        function getSignageInfo() {
            var dfd = $.Deferred();

            new Signage().getSignageInfo(
                function(result) {
                    // Success
                    logger.info("System Interface: Signage Info Loaded");
                    dfd.resolve(result);
                },
                function(result) {
                    // Failed
                    logger.error("System Interface: Error getting signage information. Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                }
            );

            return dfd;
        }

        //
        // Get Usage Data using LG SDK Signage.getUsageData()
        // Used to get the system uptime
        //
        function getUsageData() {
            var dfd = $.Deferred();

            new Signage().getUsageData(
                function(result) {
                    // Success
                    logger.info("System Interface: System Usage Data Loaded");
                    dfd.resolve(result);
                },
                function(result) {
                    // Failed
                    logger.error("System Interface: Error getting usage information. Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                }
            );

            return dfd;
        }

        self.getPlayerType = function () {

            return "LG";
        };

        self.getFirmwareVersionSync = function()
        {
            return "";
        }

    }

    // The object is a singleton as there is only a single plugin added to the application
    return new SystemInterface();
});