define(["system/DateTime", "controllers/UserConfigurationController", "system/SystemInterface", "properties/App.Settings"], function(DateTime, UserConfigurationController, System, Settings) {
	/*
		GraphicsInterface provides access to low level Graphics Information
	 */
	function GraphicsInterface() {

		var self = this;

		self.getGraphicsName = function() {
			var graphicsName = "Internal Graphics";
			logger.info("Graphics Interface: Graphics Name: " + graphicsName);
			return $.Deferred().resolve(graphicsName);
		};

		self.getGraphicsHorizontalResolution = function() {
			var hRes = document.documentElement.clientWidth;
			logger.info("Graphics Interface: Horizontal Resolution: " + hRes);
			return $.Deferred().resolve(hRes);
		};

		self.getGraphicsVerticalResolution = function() {
			var vRes = document.documentElement.clientHeight;
			logger.info("Graphics Interface: Horizontal Resolution: " + vRes);
			return $.Deferred().resolve(vRes);
		};

		self.getContrastRGB = function() {
			logger.warning("Graphics Interface: Error: Get Contrast RGB Not Implemented.");
			return $.Deferred().reject();
		};

		self.getBrightnessRGB = function() {
			logger.warning("Graphics Interface: Error: Get Brightness RGB Not Implemented.");
		    return $.Deferred().reject();
		};

		// Returns the rotation angle dependent on whether or not native portrait is being used
		// if Native portrait setting is anything other than 0 then that is returned
		// otherwise the SL rotation setting is returned instead
		self.getRotationAngle = function() {
			var dfd = $.Deferred();
			self.getNativePortraitMode()
				.done(function(nativePortraitRotation) {
					switch (nativePortraitRotation) {
						case 'off':
						case 'Unsupported':
							dfd.resolve(UserConfigurationController.rotation().angle);
							break;
						case '90':
						case '180':
						case '270':
							dfd.resolve(parseInt(nativePortraitRotation));
							break;
					}
				});
			return dfd;	
		}

		self.getSignageliveRotation = function() {
			var rotation = UserConfigurationController.rotation().angle.toString();
			logger.info("Graphics Interface: Signagelive Rotation: " + rotation);
			return $.Deferred().resolve(rotation);
		}

		self.getNativePortraitMode = function() {
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call getNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve('Unsupported');
			}
			// TEST - USING SCAP ONLY#
			// Recommended by LG that we use SCAP rather than CustomJS
			// Need to think about any potential issues transitioning from customJS to SCAP
			if (Settings.application.webOSVersion >= 6) {
				return self.getNativePortraitModeUsingIDCAP();
			}
			// webOS 3 - 4.1
			else {
				return self.getNativePortraitModeUsingCustomJS();
			}
		};

		// Original Method used CustomJS library
		self.getNativePortraitModeUsingCustomJS = function() {
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call getNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve('Unsupported');
			}
			// Otherwise we can attempt to return it
			var dfd = $.Deferred();
			new Custom().Configuration.getNativePortraitMode(
				// Call successful
                function successCallback(successObject) {
                    logger.info('Native portrait setting returned by CustomJS API: ' + successObject.nativePortrait);
                    dfd.resolve(successObject.nativePortrait);
                },
				// Call failed
                function failureCallback(failureObject) {
                    logger.error('An error occurred getting the Native Portraity setting from the CustomJS API: [' + failureObject.errorCode + '] ' + failureObject.errorText);
                    dfd.resolve('Unknown');
                }
            );
			return dfd;
		};

		self.isVideoRotationSupported = function() {
			return $.Deferred().resolve(Settings.application.features.supportsVideoRotation);
		}

		self.getDisplayProperties = function() {
			var displayProperties = {};
			var dfd = $.Deferred();
			$.when(self.getGraphicsHorizontalResolution(), self.getGraphicsVerticalResolution(), self.isVideoRotationSupported(), self.getRotationAngle(), self.getNativePortraitMode()).done(function(width, height, videoRotationSupported, nativePortraitRotationSupport, nativePortraitMode) {
				displayProperties.is_native_portrait_app = false;
				displayProperties.is_native_portrait_enabled = (nativePortraitMode !== "Unsupported" && nativePortraitMode !== "off" && nativePortraitMode !== "Unknown") ? true : false;
				displayProperties.rotation = (nativePortraitRotationSupport !== "off" && nativePortraitRotationSupport !== "Unsupported") ? nativePortraitRotationSupport : UserConfigurationController.rotation().angle;
				displayProperties.video_rotation_supported = videoRotationSupported;
				displayProperties.width = width;
				displayProperties.height = height;
				dfd.resolve(displayProperties);
			});
			return dfd;
		}

		// Updated using SCAP
		self.getNativePortraitModeUsingIDCAP = function() {
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call getNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve('Unsupported');
			}
			// Otherwise we can attempt to return it
			var dfd = $.Deferred();
			System.getIDCAPProperty("screen_rotation").done(function(currentRotation) {
				if(currentRotation) {
					dfd.resolve(currentRotation);
				} else {
					logger.error('An unknown error occurred getting the Native Portrait setting from the IDCAP');
					dfd.reject("Unknown");
				}
			}).fail(function(error) {
				logger.error('An error occurred getting the native portrait mode. Unable to load current setting: ' + error);
				dfd.reject();
			});
			return dfd;
		};

		// TEST: LG recommend using SCAP rather than CustomJS
		self.setNativePortraitMode = function(updatedRotation) {
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}

			if (Settings.application.webOSVersion >= 6) {
				return self.setNativePortraitModeUsingIDCAP(updatedRotation);
			}
			// webOS 3 - 4.1
			else {
				return self.setNativePortraitModeUsingCustomJS(updatedRotation);
			}
		};

		// Update LG Native Portrait Mode Setting (webOS 3+ ONLY)
		// Note: that the media player should be stopped when this method is called
		// Returns a promise that:
		// 		1: Resolves false if called on webOS 1/2
		// 		2: Resolves false if no update is made because the updated setting matches the current setting
		// 		3: Resolves true if the setting is updated
		self.setNativePortraitModeUsingCustomJS = function(updatedRotation) {
			logger.info('Attempting to update native portrait mode to: ' + updatedRotation);
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}
			// Configure Settings
			var nativePortraitModeOptions = {
				nativePortrait: Custom.NATIVEPORTRAIT.OFF
			};
			switch (updatedRotation) {
				case 'off':
				case '0':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.OFF;
					break;
				case '90':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_90;
					break;
				case '180':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_180;
					break;
				case '270':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_270;
					break;
				default:
					// Nothing to do just leave as OFF
					break;

			}
			// Check if the setting has actually changed from the current setting. If it hasn't then just resolve successful
			var dfd = $.Deferred();
			self.getNativePortraitMode()
				.done(function(currentRotation) {
					// Resolve success if updatedRotation === currentRotation
					if (currentRotation === nativePortraitModeOptions.nativePortrait) {
						logger.info('No need to update native portrait setting as current setting already matches updated setting. ' + currentRotation + ' === ' + updatedRotation);
						dfd.resolve(false);
					}
					// Otherwise we do need to update the  native portrait mode
					new Custom().Configuration.setNativePortraitMode(
						// Call was successful
						function successCallback() {
							logger.info('Native Portrait Rotation was updated successfully.');
							// We must have a small delay here before we continue to resize the DOM and report that the update was successful so we can restart the media player
							// to avoid a race condition where clientWidth and clientHeight are still reported with their old values.
							setTimeout(function() {
								if (updatedRotation !== '0') {
									// Reset the SL Rotation setting back to 0
									UserConfigurationController.rotation({
										displayName : 'No rotation',
										angle:0,
										id : 0
									});
									UserConfigurationController.saveSettings();
								}					
								updateDOMElementClientHeightAndWidth();
								dfd.resolve();
							}, 5000);
						},
						function failureCallback(failureObject) {
							logger.error('Error updating video rotationL [' + failureObject.errorCode + '] ' + failureObject.errorText);
							dfd.reject();
						},
						nativePortraitModeOptions
					);
				})
				.fail(function(error) {
					logger.error('An error occurred setting the native portrait mode. Unable to load current setting: ' + error);
					dfd.reject();
				});

			return dfd;
		};

		self.setNativePortraitModeUsingIDCAP = function(updatedRotation) {
			logger.info('Attempting to update native portrait mode to: ' + updatedRotation);
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}

			if(updatedRotation == '180') {
				logger.warning('Attempted to setNativePortraitModeUsingIDCAP on LG screen with webOSVersion >= 6 where 180 is not supported, so the panel will revert back to 0 / off');
			}

			switch (updatedRotation) {
				case 'off':
				case '0':
				case '180':
					updatedRotation = 'off';
					break;
				case '90':
				case '270':
					break;
				default:
					// Nothing to do just leave as OFF
					break;
			}

			var dfd = $.Deferred();
			System.getIDCAPProperty("screen_rotation").done(function(currentRotation) {
				// Resolve success if updatedRotation === currentRotation
				if (currentRotation === updatedRotation) {
					logger.info('No need to update native portrait setting as current setting already matches updated setting. ' + currentRotation + ' === ' + updatedRotation);
					dfd.resolve();
				} else {
					System.setIDCAPProperty("screen_rotation", updatedRotation).done(function() { // no callback info returned
						logger.info('Native Portrait Rotation was updated successfully.');
						// We must have a small delay here before we continue to resize the DOM and report that the update was successful so we can restart the media player
						// to avoid a race condition where clientWidth and clientHeight are still reported with their old values.
						setTimeout(function() {
							if (updatedRotation !== '0' && updatedRotation !== 'off') {
								// Reset the SL Rotation setting back to 0
								UserConfigurationController.rotation({
									displayName : 'No rotation',
									angle:0,
									id : 0
								});
								UserConfigurationController.saveSettings();
							}		
							updateDOMElementClientHeightAndWidth();
							dfd.resolve();
						}, 5000);
					}).fail(function(error) {
						logger.error('An error occurred setting the native portrait mode. Unable to load current setting: ' + error);
						dfd.reject();
					});
				}
			});
			return dfd;
		};

		// TEST
		self.hasNativePortraitModeSettingChanged = function(updatedRotation) {
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}

			if (Settings.application.webOSVersion >= 6) {
				return self.hasNativePortraitModeSettingChangedUsingIDCAP(updatedRotation);
			}
			// webOS 3 - 4.1
			else {
				return self.hasNativePortraitModeSettingChangedUsingCustomJS(updatedRotation);
			}
		};

		self.hasNativePortraitModeSettingChangedUsingCustomJS = function(updatedRotation) {
			logger.info('Attempting to update native portrait mode to: ' + updatedRotation);
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}
			// Configure Settings
			var nativePortraitModeOptions = {
				nativePortrait: Custom.NATIVEPORTRAIT.OFF
			};
			switch (updatedRotation) {
				case 'off':
				case '0':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.OFF;
					break;
				case '90':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_90;
					break;
				case '180':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_180;
					break;
				case '270':
					nativePortraitModeOptions.nativePortrait = Custom.NATIVEPORTRAIT.DEGREE_270;
					break;
				default:
					// Nothing to do just leave as OFF
					break;

			}
			// Check if the setting has actually changed from the current setting. If it hasn't then just resolve successful
			var dfd = $.Deferred();
			self.getNativePortraitMode()
				.done(function(currentRotation) {
					// Resolve success if updatedRotation === currentRotation
					if (currentRotation === nativePortraitModeOptions.nativePortrait) {
						logger.info('No need to update native portrait setting as current setting already matches updated setting. ' + currentRotation + ' === ' + updatedRotation);
						dfd.resolve(false);
					}
					else {
						logger.info('Native portrait setting updated. ' + currentRotation + ' === ' + updatedRotation);
						dfd.resolve(true);
					}
				})
				.fail(function(error) {
					logger.error('An error occurred checking for changes to the native portrait mode. Unable to load current setting: ' + error);
					dfd.resolve(false);
				});
			return dfd;
		};
		self.hasNativePortraitModeSettingChangedUsingIDCAP = function(updatedRotation) {
			logger.info('Attempting to update native portrait mode to: ' + updatedRotation);
			// Do nothing if webOS 1/2
			if (Settings.application.webOSVersion < 3) {
				logger.warning('Attempted to call setNativePortraitMode on LG screen with webOSVersion < 3');
				return $.Deferred().resolve(false);
			}

			switch (updatedRotation) {
				case 'off':
				case '0':
					updatedRotation = 'off';
				case '180':
				case '90':
				case '270':
					break;
				default:
					// Nothing to do just leave as OFF
					break;
			}
			// Check if the setting has actually changed from the current setting. If it hasn't then just resolve successful
			var dfd = $.Deferred();
			System.getIDCAPProperty("screen_rotation").done(function(currentRotation) {
				// Resolve success if updatedRotation === currentRotation
				if (currentRotation === updatedRotation) {
					logger.info('No need to update native portrait setting as current setting already matches updated setting. ' + currentRotation + ' === ' + updatedRotation);
					dfd.resolve(false);
				} else {
					logger.info('Native portrait setting updated. ' + currentRotation + ' === ' + updatedRotation);
					dfd.resolve(true);
				}
			}).fail(function(error) {
				logger.error('An error occurred checking for changes to the native portrait mode. Unable to load current setting: ' + error);
				dfd.resolve(false);
			});
			return dfd;
		};

		function updateDOMElementClientHeightAndWidth() {
			$('body').attr('width', document.documentElement.clientWidth);
			$('body').attr('height', document.documentElement.clientHeight);
			$('body').css('width', document.documentElement.clientWidth + 'px');
			$('body').css('height', document.documentElement.clientHeight + 'px');

			$('#canvas').attr('width', document.documentElement.clientWidth);
			$('#canvas').attr('height', document.documentElement.clientHeight);
			$('#canvas').css('width', document.documentElement.clientWidth + 'px');
			$('#canvas').css('height', document.documentElement.clientHeight + 'px');

			$('#layoutBackgroundCanvas').attr('width', document.documentElement.clientWidth);
			$('#layoutBackgroundCanvas').attr('height', document.documentElement.clientHeight);
			$('#layoutBackgroundCanvas').css('width', document.documentElement.clientWidth + 'px');
			$('#layoutBackgroundCanvas').css('height', document.documentElement.clientHeight + 'px');

			$('#layoutBackgroundImage').attr('width', document.documentElement.clientWidth);
			$('#layoutBackgroundImage').attr('height', document.documentElement.clientHeight);
			$('#layoutBackgroundImage').css('width', document.documentElement.clientWidth + 'px');
			$('#layoutBackgroundImage').css('height', document.documentElement.clientHeight + 'px');
		}

		/**
		 * Allows us to take a screenshot
		 */
		self.takeScreenshot = function() {
			logger.debug('Taking Screenshot');

			var dfd = $.Deferred();

			var options = {};
			options.save = false;
			options.thumbnail = false;

			var signage = new Signage();
			signage.captureScreen(function(cbObject) {
				var screenshot = {};
				screenshot.data = cbObject.data;
				screenshot.date = DateTime.now();
				screenshot.tz = UserConfigurationController.timeZone().id;
				logger.debug("Returning screenshot");
				dfd.resolve(screenshot);
			},
			function(cbObject) {
				var errorCode = cbObject.errorCode;
				var errorText = cbObject.errorText;
				logger.error("Error Code [" + errorCode + "]: " + errorText);
				dfd.reject(cbObject);
			}, options);
		
			return dfd;
		};

	}
	
	// The object is a singleton as there is only a single plugin added to the application
	return new GraphicsInterface();
});
