define(["models/StorageTypeConstants", "properties/App.Settings", "models/LocalFileList", "system/io/FileEx"], function(StorageType, Settings, LocalFileList, FileEx) {
    /**
     * The Main File IO class
     */
    function File() {
        var self = this;

        /**
         * Create a new file in local storage
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.create = function(path, forceLocal) {
            return FileEx.create(path, forceLocal);
        };

        /**
         * Write data to a new or existing file in local storage
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        self.write = function(path, data, forceLocal) {
            return FileEx.write(path, data, forceLocal);
        };

        // Unzip a file
        self.unzip = function(zipPath, targetPath) {
            var dfd = $.Deferred();

            //unzip the file
            var options = {};
            options.zipPath = zipPath;
            options.targetPath = targetPath;

            new Storage().unzipFile(
                // Success
                function(result) {
                    dfd.resolve();
                },
                // Failed
                function(error) {
                    console.error(error.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        };

        /**
         * Append data to an existing file in local storage
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        self.append = function(path, data, mimeType, forceLocal) {
            return FileEx.append(path, data, forceLocal);
        };

        self.keyExists = function(path) {
            var dfd = $.Deferred();

            var keyExists = false;

            //logger.debug('FILE APPEND: Checking Key: ' + path + ' Exists in local storage');

            try {
                // Handle Success
                var get = localStorage.getItem(path);

                if (get != undefined) {
                    //logger.debug('FILE APPEND: Key Exists');
                    // Result returned
                    dfd.resolve(true);
                    keyExists = true;
                } else {
                    //logger.debug('FILE APPEND: Key Does Not Exist');
                    // result not returned
                    dfd.resolve(false);
                }
            } catch (e) {
                // Handle error
                logger.error("File: Error checking if key " + path + " exists in local storage.");
                dfd.resolve(false);
            };

            return dfd;
        };

        /**
         * Read data from a file in local storage
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.read = function(path, forceLocal) {
            logger.info('File.read(): ' + path);
            return FileEx.read(path, forceLocal);
        };

        /**
         * Check if a file exists. This uses the LocalFileList.exists method.
         * This can be a file in local storage or a physical media file on the device
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.exists = function(path) {
            return LocalFileList.exists(path);
        };

        self.exists2 = function(path) {
            var dfd = $.Deferred();

            var options = {};
            options.path = path;

            new Storage().exists(
                // Success
                function(success) {
                    dfd.resolve(success.exists);
                },
                // Error
                function(error) {
                    logger.error(error.errorText);
                    dfd.resolve(false);
                },
                options
            );

            return dfd;
        };

        /**
         * Delete file form physical storage only
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.delete = function(path) {
            var dfd = $.Deferred();

            // Remove files using LG SDK removeFile method
            new Storage().removeFile(
                // Success
                function(result) {
                    logger.info('File Plugin: Successfully deleted file: ' + path + '.');
                    LocalFileList.remove(path).always(function() {
                        logger.info('Updated file list');
                        dfd.resolve();
                    });
                },
                // Failed
                function(result) {
                    logger.error('File Plugin: Error whilst deleting file ' + path + '. Error: [' + result.errorCode + '] ' + result.errorText);
                    dfd.reject();
                },
                // options
                {
                    file: path
                }
            );

            return dfd;
        };

        /**
         * Delete entry from local storage - actually just empties it
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.deleteLocal = function(path) {
            // if (Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
            //     logger.info("IPK Client - Doing physical file delete of pop log");
            //     return self.delete(path);
            // } else {
            //     logger.info("ZIP Client - Clearing pop log local storage data");
            return self.write(path, '', true);
            //}
        };

        /**
         * This actually copies a file. At the moment this is used
         * for copying a temp file to a permanemt file, and the temp
         * file is then removed.
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        self.move = function(source, dest) {
            var _source = source;
            var _dest = dest;

            logger.debug('File Move: Start moving from: ' + _source + ' to: ' + _dest);

            _Deferred = $.Deferred();

            // Download file using LG SDK copyFile method
            new Storage().copyFile(
                // Move Succeeded
                function(result) {
                    logger.debug('File Move: Successfully Moved file to: ' + _dest);
                    LocalFileList.add(_dest);
                    _Deferred.resolve();
                },
                // Download Failed
                function(result) {
                    logger.error('File Move: Error whilst moving file from: ' + _source + ' to ' + _dest + '. Error: [' + result.errorCode + '] ' + result.errorText);
                    _Deferred.reject(_source, _dest);
                },
                // Copy FilesOptions
                {
                    source: _source,
                    destination: _dest
                }
            );

            return _Deferred.promise();
        };

        /**
         * This copies a local storage entry. 
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        self.moveLocal = function(source, dest) {
            var _source = source;
            var _dest = dest;

            //logger.debug('File Move: Start moving from: ' + _source + ' to: ' + _dest);

            _Deferred = $.Deferred();

            self.read(source, true)
                .done(function(sourceFile) {
                    // Write dest file
                    //logger.debug("MOVE LOCAL: " + sourceFile + " to " + dest);
                    self.append(dest, sourceFile, "text/csv", true)
                        .done(function() {
                            logger.info("File: Successfully moved file: " + source + " to " + dest);
                            _Deferred.resolve();
                        })
                        .fail(function() {
                            logger.error("File: Could not write file " + dest + " while attempting to move");
                            _Deferred.reject();
                        });
                }).fail(function() {
                    logger.error("File: Could not read file " + source + " while attempting to move");
                    _Deferred.reject();
                });

            return _Deferred;
        };

        // Reset PoP
        self.resetProofOfPlay = function(sourceLogURI, uploadLogURI) {
            var dfd = $.Deferred();
            logger.info("PoP: Reset Proof of Play");
            logger.info("PoP: Checking if upload log already exists");
            try {
                if (self.exists(uploadLogURI)) {
                    logger.info("PoP: Upload log already exists. Appending new log data to existing upload log file.");
                    self.read(sourceLogURI, true)
                        .done(function(sourcePoPReport) {
                            if (sourcePoPReport != null) {
                                logger.info("PoP: Source report data loaded: " + sourcePoPReport);
                                self.append(uploadLogURI, sourcePoPReport, "text/csv", true)
                                    .done(function() {
                                        logger.info("PoP: Data appended to existing upload log successfully");
                                        self.deleteLocal(sourceLogURI)
                                            .done(function() {
                                                logger.info("PoP: Source log deleted.");
                                                logger.info("PoP: Successfully reset proof of play.");
                                                dfd.resolve();
                                            }).fail(function() {
                                                logger.error("PoP: Could not delete file PoP source log.");
                                                dfd.reject();
                                            });
                                    })
                                    .fail(function() {
                                        logger.error("PoP: Unable to append new data to upload log file.");
                                        dfd.reject();
                                    });
                            } else {
                                logger.warning("PoP: Unable to read data from source log file and update the PoP upload file. This may be because a source log file has not been recreated since the last time the logs were rolled.");
                                dfd.resolve();
                            }

                        })
                        .fail(function() {
                            logger.warning("PoP: Unable to read data from source log file and update the PoP upload file. This may be because a source log file has not been recreated since the last time the logs were rolled.");
                            dfd.resolve();
                        });
                } else {
                    logger.info("PoP: Does not exist. Copying current log to create new upload log.");
                    self.moveLocal(sourceLogURI, uploadLogURI)
                        .done(function() {
                            logger.info("PoP: Upload log created.");
                            self.deleteLocal(sourceLogURI)
                                .done(function() {
                                    logger.info("PoP: Source log deleted.");
                                    logger.info("PoP: Successfully reset proof of play.");
                                    dfd.resolve();
                                }).fail(function() {
                                    logger.error("PoP: Could not delete file PoP source log.");
                                    dfd.reject();
                                });
                        }).fail(function() {
                            logger.warning("Unable to copy PoP data to upload file while attempting to reset proof of play. This is likely because no PoP data has been stored.");
                            dfd.resolve();
                        });
                }
            } catch (error) {
                logger.error("PoP: Exception thrown resetting PoP: " + error);
                dfd.reject();
            }
            return dfd;
        };
    }

    var file = new File();
    return file;
});