define(["properties/App.Settings", "models/ObjectProcessingQueueModel"], function(Settings, ObjectProcessingQueueModel) {
    function FileBase() {
        var self = this;

        var MAX_LG_WRITE_BUFFER_SIZE = 10240 // (10KB)
        var MAX_LG_READ_BUFFER_SIZE = 10240 // (10KB)

        /**
         * Resolves true if file exists or false if it doesnt
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.fileExists = function(path) {
            var dfd = $.Deferred();

            new Storage().exists(
                //success
                function(result) {
                    logger.debug(path + " exists: " + result.exists);
                    dfd.resolve(result.exists);
                },
                // failed
                function(error) {
                    logger.error("Error reading block from " + path + ". Error Code [" + error.errorCode + "]: " + error.errorText);
                    dfd.resolve(false);
                }, {
                    path: path
                }
            );

            return dfd;
        }

        /**
         * Allows to read data from a file on the filesystem
         * If the file is larger than 10230 bytes it will read it in chunks
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.readFromFileSystem = function(path) {
            var dfd = $.Deferred();

            // find out size of file
            logger.debug("Checking size of file...");
            var options = {
                path: path,
            };
            new Storage().statFile(
                // Success
                function(result) {
                    logger.debug("Successfully stat file " + path + " from filesystem. File is " + result.size + "bytes");

                    // If the data length is less than the max we can just write the data
                    //var fileSizeInKB = Math.ceil(result.size/1000);
                    var fileSizeInKB = result.size;
                    logger.debug("File size in KB " + fileSizeInKB);
                    if (fileSizeInKB <= MAX_LG_READ_BUFFER_SIZE) {
                        logger.debug("Filesize less than limit. Reading entire file in one go.");
                        // just do simple read
                        readDataFromFileSystem(path, 'utf8', 0, fileSizeInKB)
                            .done(function(data) {
                                dfd.resolve(data);
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    }
                    // we need to read in blocks and stitch together
                    else {
                        logger.debug("File must be read in blocks.");
                        // Calculate the number and position of each block we need to read
                        // and create a queue to process each read synchronously. We can then
                        // stitch the results together
                        var numberOfReadBlocks = Math.ceil(fileSizeInKB / MAX_LG_READ_BUFFER_SIZE);

                        logger.debug("Number of blocks required: " + numberOfReadBlocks);

                        var readBlocks = [];

                        for (var i = 0; i < numberOfReadBlocks; i++) {
                            // the length is always the max buffer size
                            var length = MAX_LG_READ_BUFFER_SIZE;
                            var position = i * MAX_LG_READ_BUFFER_SIZE;

                            // except if this is the last block in which case it is just the remainder
                            if (i == numberOfReadBlocks - 1) {

                                length = fileSizeInKB - (i * MAX_LG_READ_BUFFER_SIZE);
                            }

                            readBlocks.push({ position: position, length: length });
                        }

                        logger.debug(JSON.stringify(readBlocks));

                        var readBlocksQueue = new ObjectProcessingQueueModel(readBlocks);

                        var result = "";

                        logger.debug("Starting read in blocks..");
                        readDataFromFileSystemInBlocks(readBlocksQueue, path, result)
                            .done(function(data) {
                                logger.debug("Successfully red all data" + JSON.stringify(data));
                                dfd.resolve(data);
                            })
                            .fail(function() {
                                logger.error("Error reading file " + path + " in blocks from filesystem");
                                dfd.reject();
                            });

                    }


                },
                // failed
                function(error) {
                    logger.error("Error stating file " + path + " from filesystem");
                    dfd.reject();
                },
                options
            );

            return dfd;
        }

        function readDataFromFileSystemInBlocks(readBlocksQueue, path, result) {
            var dfd = $.Deferred();

            readNextDataBlockFromFileSystem(readBlocksQueue, dfd, path, result);

            return dfd.promise();
        }

        function readNextDataBlockFromFileSystem(readBlocksQueue, dfd, path, result) {
            logger.debug("Number of blocks remaining: " + readBlocksQueue.items.length);
            // is the queue empty - return the success of the queue processing?
            if (readBlocksQueue.items.length == 0) {
                if (readBlocksQueue.failedItems.length == 0) {
                    logger.debug("All blocks processed");
                    dfd.resolve(result);
                } else {
                    logger.error("Failed to read all blocks");
                    dfd.reject(readBlocksQueue.processedItems, readBlocksQueue.failedItems);
                }
                return;
            }

            // Get current item
            var currentReadBlock = readBlocksQueue.currentItem();

            logger.debug("Reading " + currentReadBlock.length + " block of " + path + " from position " + currentReadBlock.position);

            readDataFromFileSystem(path, 'utf8', currentReadBlock.position, currentReadBlock.length)
                .done(function(data) {
                    var block = readBlocksQueue.shift();
                    readBlocksQueue.processedItems.push(block);
                    result += data;
                    logger.debug("Successfully read block");
                })
                .fail(function() {
                    var block = readBlocksQueue.shift();
                    readBlocksQueue.failedItems.push(block);
                    logger.error("Failed reading block");
                })
                .always(function() {
                    logger.debug("Moving on to next read block");
                    readNextDataBlockFromFileSystem(readBlocksQueue, dfd, path, result);
                });
        }


        function readDataFromFileSystem(path, encoding, position, length) {
            var dfd = $.Deferred();

            var options = {
                path: path,
                encoding: encoding,
                position: position,
                length: length
            };

            new Storage().readFile(
                // Success
                function(result) {
                    logger.debug("Successfully read file " + path + " from filesystem");
                    dfd.resolve(result.data);
                },
                // failed
                function(error) {
                    logger.error("Error reading block from " + path + ". Error Code [" + error.errorCode + "]: " + error.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        }

        /**
         * This method writes data to a file in the filesystem.
         * If files are too large it will do it in chunks
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        self.writeToFileSystem = function(path, data, mode) {
            var dfd = $.Deferred();

            // By default truncate and override existing file
            if (typeof(mode) == 'undefined')
                mode = 'truncate';

            if (data == null || data == '') {
                logger.warning("Unable to write file: " + path + ". Data is null or empty string.");
                dfd.reject();
                return dfd;
            }

            // If the data length is less than the max we can just write the data
            if (data.length <= MAX_LG_WRITE_BUFFER_SIZE) {
                writeDataToFileSystem(path, data, mode, 'utf8', data.length)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            // otherwise we need to break it down into chunks of at most MAX_LG_WRITE_BUFFER_SIZE and write each sequentially.
            else {
                // split data into blocks that can be written
                logger.debug(data.length + " is greater than limit " + MAX_LG_WRITE_BUFFER_SIZE + ". Writing data in blocks.");
                var dataBlocks = data.match(/(.|[\r\n]){1,10240}/g);

                var dataBlocksQueue = new ObjectProcessingQueueModel(dataBlocks);

                logger.debug("Data broken down into " + dataBlocks.length);

                // if mode == truncate then truncate the file first then write each chunk in append mode
                if (mode == 'truncate') {
                    // truncate the file
                    logger.debug("Truncating " + path);
                    // Truncate and write the first block
                    var firstDataBlock = dataBlocksQueue.shift();
                    writeDataToFileSystem(path, firstDataBlock, 'truncate', "utf8", firstDataBlock.length)
                        .done(function() {
                            dataBlocksQueue.processedItems.push(firstDataBlock);
                            logger.debug("Successfully truncated " + path + ". Starting writing data blocks...");

                            writeDataToFileSystemInBlocks(dataBlocksQueue, path)
                                .done(function() {
                                    logger.debug("Wrote all data blocks to file. Updateding local file list");
                                    dfd.resolve();
                                })
                                .fail(function() {
                                    dataBlocksQueue.failedItems.push(firstDataBlock);
                                    logger.error("Error saving all data blocks to file.");
                                    dfd.reject();
                                });
                        })
                        .fail(function() {
                            logger.error("Failed to truncate " + path);
                            dfd.reject();
                        });
                } else {

                    writeDataToFileSystemInBlocks(dataBlocksQueue, path)
                        .done(function() {
                            logger.debug("Wrote all data blocks to file.");
                            dfd.resolve();
                        })
                        .fail(function() {
                            logger.error("Error saving all data blocks to file.");
                            dfd.reject();
                        });
                }

            }

            return dfd;
        }

        function writeDataToFileSystemInBlocks(dataBlocksQueue, path) {
            var dfd = $.Deferred();

            writeNextDataBlockToFileSystem(dataBlocksQueue, dfd, path);

            return dfd.promise();
        }

        function writeNextDataBlockToFileSystem(dataBlocksQueue, dfd, path) {
            logger.debug("Number of blocks remaining: " + dataBlocksQueue.items.length);
            // is the queue empty - return the success of the queue processing?
            if (dataBlocksQueue.items.length == 0) {
                if (dataBlocksQueue.failedItems.length == 0) {
                    logger.debug("All blocks processed");
                    dfd.resolve(dataBlocksQueue.processedItems);
                } else {
                    logger.error("Failed to write all blocks");
                    dfd.reject(dataBlocksQueue.processedItems, dataBlocksQueue.failedItems);
                }
                return;
            }

            // Get current item
            var dataBlock = dataBlocksQueue.currentItem();

            logger.debug("Writing " + dataBlock.length + " block of " + path);

            writeDataToFileSystem(path, dataBlock, 'append', 'utf8', dataBlock.length)
                .done(function() {
                    var dataBlock = dataBlocksQueue.shift();
                    dataBlocksQueue.processedItems.push(dataBlock);
                    logger.debug("Successfully wrote block");
                })
                .fail(function() {
                    var dataBlock = dataBlocksQueue.shift();
                    dataBlocksQueue.failedItems.push(dataBlock);
                    logger.error("Failed writing block");
                })
                .always(function() {
                    logger.debug("Moving on to next block");
                    writeNextDataBlockToFileSystem(dataBlocksQueue, dfd, path);
                });
        }

        function writeDataToFileSystem(path, data, mode, encoding, length) {
            var dfd = $.Deferred();

            var options = {
                data: data,
                path: path,
                mode: mode,
                encoding: encoding,
                length: length
            };

            new Storage().writeFile(
                // Success
                function(result) {
                    logger.debug("Successfully wrote " + result.written + " data to " + path);
                    dfd.resolve();
                },
                // Error
                function(error) {
                    logger.error("Error writing to " + path + ". Error Code [" + error.errorCode + "]: " + error.errorText);
                    dfd.reject();
                },
                options
            );


            return dfd;
        }
    }

    return new FileBase();
});