define(["models/StorageTypeConstants", "properties/App.Settings", "models/LocalFileList", "models/ObjectProcessingQueueModel", "system/io/FileBase"], function(StorageType, Settings, LocalFileList, ObjectProcessingQueueModel, FileBase) {
    /**
     * The Main File IO class
     */
    function File() {
        var self = this;

        /**
         * Create a new empty file.
         * For SCAP 1.2+ files will be written to the file system
         * SCAP 1.0 and 1.1 will fall back to local storage
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.create = function(path, forceLocal) {
            var dfd = $.Deferred();
            var data = ""; // Just write a new file with no data!
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Writing data to the file system.");
                FileBase.writeToFileSystem(path, data)
                    .done(function() {
                        // Upadte local file list
                        LocalFileList.add(path)
                            .done(function() {
                                dfd.resolve();
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            } else {
                logger.debug("Old scap version detected. Writing data to local storage.");
                writeToLocalStorage(path, data)
                    .done(function() {
                        LocalFileList.add(path)
                            .done(function() {
                                dfd.resolve();
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            return dfd;
        };

        /**
         * Write data to a new or existing file.
         * For SCAP 1.2+ files will be written to the file system
         * SCAP 1.0 and 1.1 will fall back to local storage
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        self.write = function(path, data, forceLocal) {
            var dfd = $.Deferred();
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Writing data to the file system.");
                FileBase.writeToFileSystem(path, data)
                    .done(function() {
                        LocalFileList.add(path)
                            .done(function() {
                                dfd.resolve();
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            } else {
                logger.debug("Writing data to local storage.");
                writeToLocalStorage(path, data)
                    .done(function() {
                        LocalFileList.add(path)
                            .done(function() {
                                dfd.resolve();
                            })
                            .fail(function() {
                                dfd.reject();
                            });
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            return dfd;
        };

        /**
         * Writes to localstorage - required for legacy panels that do not support SCAP 1.2
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        function writeToLocalStorage(path, data) {
            var dfd = $.Deferred();

            setTimeout(function() {
                localStorage[path] = data;
                dfd.resolve();
            }, 100);

            return dfd;
        }

        /**
         * Append data to the end of an existing file
         * For SCAP 1.2+ files will be written to the file system
         * SCAP 1.0 and 1.1 will fall back to local storage
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        self.append = function(path, data, forceLocal) {
            var dfd = $.Deferred();
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Appending to file system");
                // Just use the writeToFileSystem method in append mode
                FileBase.writeToFileSystem(path, data, 'append')
                    .done(function() {
                        //logger.info("Appended data to file: " + path);
                        if (LocalFileList.exists(path) == false) {
                            logger.info("Added new file to local file list: " + path);
                            LocalFileList.add(path)
                                .done(function() {
                                    dfd.resolve();
                                })
                                .fail(function() {
                                    dfd.reject();
                                });
                        } else {
                            //logger.info("File already exists in local file list: " + path);
                            dfd.resolve();
                        }

                    })
                    .fail(function() {
                        dfd.reject();
                    });
            } else {
                logger.debug("Appending to local storage");
                appendToLocalStorage(path, data, forceLocal)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            return dfd;
        };

        /**
         * Appends data to file in local storage for legacy
         * players SCAP 1.0 & 1.1
         * @param  {[type]} path [description]
         * @param  {[type]} data [description]
         * @return {[type]}      [description]
         */
        function appendToLocalStorage(path, data, forceLocal) {
            var dfd = $.Deferred();

            var appendData = "";

            // First check if the key exists, if it does then read it prior to writing
            keyExists(path)
                .done(function(exists) {
                    if (exists) {
                        self.read(path, forceLocal)
                            .done(function(readdata) {
                                readdata = readdata + data;
                                self.write(path, readdata, forceLocal)
                                    .done(function() {
                                        dfd.resolve();
                                    }).fail(function() {
                                        logger.error("File: Could not write file " + path + " while attempting to append");
                                        dfd.reject();
                                    });
                            }).fail(function() {
                                logger.error("File: Could not read file " + path + " while attempting to append");
                                dfd.reject();
                            });
                    } else {
                        self.write(path, data, forceLocal)
                            .done(function() {
                                dfd.resolve();
                            }).fail(function() {
                                logger.error("File: Could not write file " + path + " while attempting to append");
                                dfd.reject();
                            });
                    }
                });

            return dfd;
        }

        /**
         * True if the key exists in local storage, false otherwise.
         * Only used by legacy players.
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        function keyExists(path) {
            var dfd = $.Deferred();

            var keyExists = false;

            //logger.debug('FILE APPEND: Checking Key: ' + path + ' Exists in local storage');

            try {
                // Handle Success
                var get = localStorage.getItem(path);

                if (get != undefined) {
                    //logger.debug('FILE APPEND: Key Exists');
                    // Result returned
                    dfd.resolve(true);
                    keyExists = true;
                } else {
                    //logger.debug('FILE APPEND: Key Does Not Exist');
                    // result not returned
                    dfd.resolve(false);
                }
            } catch (e) {
                // Handle error
                logger.error("File: Error checking if key " + path + " exists in local storage.");
                dfd.resolve(false);
            };

            return dfd;
        };

        /**
         * Read data from a file in local storage
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.read = function(path, forceLocal) {
            var dfd = $.Deferred();
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Reading data from the file system.");
                FileBase.readFromFileSystem(path)
                    .done(function(data) {
                        dfd.resolve(data);
                    })
                    .fail(function() {
                        dfd.resolve(null);
                    });
            } else {
                logger.debug("Reading data from local storage.");
                readFromLocalStorage(path)
                    .done(function(data) {
                        dfd.resolve(data);
                    })
                    .fail(function() {
                        dfd.resolve(null);
                    });
            }
            return dfd;
        };

        /**
         * Reads data from a file in local storage for legacy players
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        function readFromLocalStorage(path) {
            var dfd = $.Deferred();

            setTimeout(function() {
                dfd.resolve(localStorage[path]);
            }, 100);

            return dfd;
        }

        /**
         * Check if a file exists. This uses the LocalFileList.exists method.
         * This can be a file in local storage or a physical media file on the device
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.exists = function(path) {
            return LocalFileList.exists(path);
        };

        /**
         * Delete file form physical storage only
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        self.delete = function(path, forceLocal) {
            var dfd = $.Deferred();
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Deleting " + path + " from the file system.");
                deleteFromFileSystem(path)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            } else {
                logger.debug("Deleting " + path + " from local storage.");
                deleteFromLocalStorage(path)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            return dfd;
        };

        /**
         * Deletes a file from the file system (SCAP 1.2+ only)
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        function deleteFromFileSystem(path) {
            var dfd = $.Deferred();

            // Remove files using LG SDK removeFile method
            new Storage().removeFile(
                // Success
                function(result) {
                    logger.debug('File Plugin: Successfully deleted file: ' + path + '.');
                    LocalFileList.remove(path)
                        .done(function() {
                            dfd.resolve();
                        })
                        .fail(function() {
                            dfd.reject();
                        });
                },
                // Failed
                function(error) {
                    //logger.error('File Plugin: Error whilst deleting file ' + path + '. Error: [' + error.errorCode + '] ' + error.errorText);
                    dfd.reject();
                },
                // options
                {
                    file: path
                }
            );
            return dfd;
        }


        /**
         * Delete entry from local storage - actually just empties it
         * @param  {[type]} path [description]
         * @return {[type]}      [description]
         */
        function deleteFromLocalStorage(path) {
            var dfd = $.Deferred();

            setTimeout(function() {
                self.write(path, '');
                dfd.resolve();
            }, 100);

            return dfd;
        }

        /**
         * Resets Proof of Play.
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        self.resetProofOfPlay = function(source, dest) {
            var dfd = $.Deferred();

            self.move(source, dest)
                .done(function() {
                    self.delete(source)
                        .done(function() {
                            logger.info("File: Successfully reset proof of play file: " + source + " to " + dest);
                            dfd.resolve();
                        }).fail(function() {
                            logger.error("File: Could not delete file " + source + " while attempting to reset proof of play");
                            dfd.reject();
                        });
                }).fail(function() {
                    logger.error("File: Could not move file " + source + " while attempting to reset proof of play");
                    dfd.reject();
                });

            return dfd;
        };



        /**
         * Copies the source file to the destination.
         * uses the local file system SCAP 1.2+
         * or local storage SCAP 1.0, 1.1
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        self.move = function(source, dest, forceLocal) {
            var dfd = $.Deferred();
            if (forceLocal !== true && Settings.application.sdkVersion >= 1.2 && Settings.application.packageType === "IPK") {
                logger.debug("Moving from the file system.");
                moveFileFromFileSystem(source, dest)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            } else {
                logger.debug("Old scap version detected. Moving from local storage.");
                moveFileFromLocalStorage(source, dest)
                    .done(function() {
                        dfd.resolve();
                    })
                    .fail(function() {
                        dfd.reject();
                    });
            }
            return dfd;
        };

        /**
         * Copies file from source to dest in local file system
         * SCAP 1.2+ only
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        function moveFileFromFileSystem(source, dest) {
            var dfd = $.Deferred();

            // Download file using LG SDK copyFile method
            new Storage().copyFile(
                // Move Succeeded
                function(result) {
                    logger.debug('File Move: Successfully Moved file to: ' + dest);
                    LocalFileList.add(dest)
                        .done(function() {
                            dfd.resolve();
                        })
                        .fail(function() {
                            dfd.reject();
                        });
                    dfd.resolve();
                },
                // Download Failed
                function(error) {
                    logger.error('File Move: Error whilst moving file from: ' + source + ' to ' + dest + '. Error: [' + error.errorCode + '] ' + error.errorText);
                    dfd.reject(source, dest);
                },
                // Copy FilesOptions
                {
                    source: source,
                    destination: dest
                }
            );

            return dfd;
        }

        /**
         * This copies a local storage entry.
         * SCAP 1.0, 1.1
         * @param  {[type]} source [description]
         * @param  {[type]} dest   [description]
         * @return {[type]}        [description]
         */
        function moveFileFromLocalStorage(source, dest) {
            var dfd = $.Deferred();

            self.read(source)
                .done(function(sourceFile) {
                    // Write dest file
                    //logger.debug("MOVE LOCAL: " + sourceFile + " to " + dest);
                    self.append(dest, sourceFile)
                        .done(function() {
                            logger.debug("File: Successfully moved file: " + source + " to " + dest);
                            dfd.resolve();
                        })
                        .fail(function() {
                            logger.error("File: Could not write file " + dest + " while attempting to move");
                            dfd.reject();
                        });
                }).fail(function() {
                    logger.error("File: Could not read file " + source + " while attempting to move");
                    dfd.reject();
                });

            return dfd;
        };
    }

    var file = new File();
    return file;
});