define(["properties/App.Settings"], function(Settings) {
	/*
		Network Interface - Provides access to Network information
		----------------------------------------------------------------
		ConnectionTypes: -1 = Unknown, 1 = Wired, 2 = Wifi, 3 = WifiDirect
	 */
	function NetworkInterface() {
		var self = this;
		var APIInterface = null;

		self.registerAPIInterface = function(i) {
			APIInterface = i;
		};

		// This is a dummy address returned by LG devices with no wired adapter
		// We use this to determine if the device is a cooler (or potentially any device without a wired adapter).
		// When we receive this we will use the wireless MAC instead
		var DUMMY_MAC_ADDRESS = "00:B0:00:A1:01:01";

		// Defaults
		self.networkInfo = null;
		self.networkMacInfo = null;
		self.connectionType = null;

		var APIInterface = null;

		self.getMAC = function() {
			var dfd = $.Deferred();
			getNetworkMacInfo()
				.done(function(info) {
					// If we receive the dummy address then return the wireless mac instead
					if (info.wiredInfo.macAddress === "00:B0:00:A1:01:01") {
						dfd.resolve(info.wifiInfo.macAddress);
					}
					// Use the wired Mac address
					{
						dfd.resolve(info.wiredInfo.macAddress);
					}
				})
				.fail(function() { 
						dfd.reject();
					});

			return dfd;	
		};

		self.getActiveNetworkConnectionType = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					dfd.resolve(self.connectionType);
				})
				.fail(function() {
					dfd.reject();
				});
			return dfd;
		};

		self.getIP = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					switch (self.connectionType) {
						case 1:
							dfd.resolve(info.wired.ipAddress);
						case 2:
							dfd.resolve(info.wifi.ipAddress);
						case 3:
							dfd.resolve(info.wifiDirect.ipAddress);
						default:
							dfd.reject();
					}
				})
				.fail(function() {
					dfd.reject();
				});
			return dfd;			
		};

		self.getSubnetMask = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					switch (self.connectionType) {
						case 1:
							dfd.resolve(info.wired.netmask);
						case 2:
							dfd.resolve(info.wifi.netmask);
						case 3:
							dfd.resolve(info.wifiDirect.netmask);
						default:
							dfd.reject();
					}
				})
				.fail(function() {
					dfd.reject();
				});
			return dfd;	
		};

		self.getDefaultGateway = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					switch (self.connectionType) {
						case 1:
							dfd.resolve(info.wired.gateway);
						case 2:
							dfd.resolve(info.wifi.gateway);
						case 3:
							dfd.resolve(info.wifiDirect.gateway);
						default:
							dfd.reject();
					}
				})
				.fail(
				function() {
					dfd.reject();
				});
			return dfd;		
		};

		self.getDNS = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					switch (self.connectionType) {
						case 1:
							dfd.resolve(info.wired.dns1);
						case 2:
							dfd.resolve(info.wifi.dns1);
						case 3:
							dfd.resolve(info.wifiDirect.dns1);
						default:
							dfd.reject();
					}
				})
				.fail(function() {
					dfd.reject();
				});
			return dfd;	
		};

		self.isDHCP = function() {
			var dfd = $.Deferred();
			getNetworkInfo()
				.done(function(info) {
					switch (self.connectionType) {
						case 1:
							if (info.wired.method == "dhcp")
								dfd.resolve(true);
							else
								dfd.resolve(false);
						case 2:
							if (info.wired.method == "dhcp")
								dfd.resolve(true);
							else
								dfd.resolve(false);
						case 3:
							if (info.wired.method == "dhcp")
								dfd.resolve(true);
							else
								dfd.resolve(false);
						default:
							dfd.reject();
					}
				})
				.fail(
				function() {
					dfd.reject();
				});
			return dfd;
		};

		self.registerAPIInterface = function(i) {
			APIInterface = i;
		};

		function getNetworkInfo() {
			
			var dfd = $.Deferred();
			new DeviceInfo().getNetworkInfo(
				// Success
				function(result) {
					logger.info('Network Interface: Successfully got Network Info: ' + JSON.stringify(result));
					self.networkInfo = result;
					setNetworkConnectionType(self.networkInfo);
					dfd.resolve(self.networkInfo);
				},
				// Failed
				function(result) {
					logger.error('Network Interface: Error getting network info. Error [' + result.errorCode + '] ' + result.errorText);
					dfd.reject();
				}

			);
			return dfd;
		}

		function getNetworkMacInfo() {

			var dfd = $.Deferred();

			new DeviceInfo().getNetworkMacInfo(
				// Success
				function(result) {
					logger.debug('Network Interface: Successfully got Network MAC Info: ' + JSON.stringify(result));
					self.networkMacInfo = result;
					dfd.resolve(self.networkMacInfo);
				},
				// Failed
				function(result) {
					logger.error('Network Interface: Error getting Network MAC info. Error [' + result.errorCode + '] ' + result.errorText);
					dfd.reject();
				}
			);

			return dfd;
		}

		// Sets the connection type based on the state returned by the LG getNetworkInfo method
		function setNetworkConnectionType(info) {
			if (info.wired.state == "connected")
				self.connectionType = 1;
			else if (info.wifi.state == "connected")
				self.connectionType = 2;
			else if (info.wifiDirect.state == "connected")
				self.connectionType = 3;
			else
				self.connectionType = -1;
		}

		self.getOnlineStatus = function() {
			var dfd = $.Deferred();

			getNetworkInfo()
				.always(function(networkInfo) {

					var onlineStatus = {
						network_connected: window.navigator.onLine,
						internet_access: false
					};

					var doOnlineCheckUsingNetworkInfo = Settings.application.webOSVersion > 3.2 && networkInfo && networkInfo.hasOwnProperty('isInternetConnectionAvailable');

					// First check the LG SCAP API method to determine if an internet connection is available
					if (doOnlineCheckUsingNetworkInfo) {
						onlineStatus.internet_access = networkInfo.isInternetConnectionAvailable;
						return dfd.resolve(onlineStatus);
					}
					
					// Fallback to our own method

					// When the player first starts up there is a delay starting the API Interface.
					// If this is null then set internet access to true and resolve 
					if (APIInterface == null) {
						onlineStatus.internet_access = true;
						return dfd.resolve(onlineStatus);
					}
		
					if (lastAPIConnectionIsWithinThePast5Minutes()) {
						onlineStatus.internet_access = true;
						return dfd.resolve(onlineStatus);
					}
					// Attempt a health check
					else {
						var healthCheckDfd = APIInterface.healthCheck();
						if (healthCheckDfd) {
							healthCheckDfd.always(function () {
								if (lastAPIConnectionIsWithinThePast5Minutes()) {
									onlineStatus.internet_access = true;
								}
								return dfd.resolve(onlineStatus);
							});
						} else {
							return dfd.resolve(onlineStatus);
						}
					}

				});
			
			return dfd;
		};

		function lastAPIConnectionIsWithinThePast5Minutes() {
			var currentTime = new Date();
			var threshold = new Date(currentTime.getTime() - 5 * 60000); // 5 minutes ago
			return APIInterface != null && APIInterface.lastConnectionTime != null && APIInterface.lastConnectionTime > threshold;
		}
	}

	return new NetworkInterface();
});
