define(["system/io/File", "properties/App.Settings", "models/RebootReasons"], function(File, Settings, RebootReasons) {
    /*
    	PowerInterface provides access to low level LG Panel Power functions and information
     */
    function PowerInterface() {

        var self = this;
        var _screenState = "on";

        /**
         * Implements LG setDisplayMode to turn the display on
         * 
         * @param  {[type]} persistState If optional parameter given this will store the state information to the filesystem, else the file will be cleared
         */
        self.switchScreenOn = function(persistState) {
            var dfd = $.Deferred();

            var options = {};
            options.displayMode = Power.DisplayMode.DISPLAY_ON;

            new Power().setDisplayMode(
                function(result) {
                    // Success
                    logger.info("Power Interface: Display Mode set to Display On.");
                    if(persistState) {
                        storeLastScreenStatus("ScreenOn");
                    } else { //if we dont want to persist the state then we should clear it due to a scheduled screen control change
                        clearLastScreenState();
                    }
                    dfd.resolve();
                },
                function(result) {
                    // Failed
                    logger.error("Power Interface: Error Setting Display Mode to Display On: Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        };

        /**
         * Implements LG setDisplayMode to turn the display off
         * 
         * @param  {[type]} persistState If optional parameter given this will store the state information to the filesystem, else the file will be cleared
         */
        self.switchScreenOff = function(persistState) {
            var dfd = $.Deferred();

            var options = {};
            options.displayMode = Power.DisplayMode.DISPLAY_OFF;

            new Power().setDisplayMode(
                function(result) {
                    // Success
                    logger.info("Power Interface: Display Mode set to Display Off.");
                    if(persistState) {
                        storeLastScreenStatus("ScreenOff");
                    } else { //if we dont want to persist the state then we should clear it due to a scheduled screen control change
                        clearLastScreenState();
                    }
                    dfd.resolve();
                },
                function(result) {
                    // Failed
                    logger.error("Power Interface: Error Setting Display Mode to Display On: Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                },
                options
            );

            return dfd;
        };

        /**
         * Implements LG getPMMode to return the current power management mode (webOS 3.0+)
         */
        self.getPMMode = function() {
            var dfd = $.Deferred();
            console.log("Get PM Mode");

            // NOTE: Dont need this?
            if(Settings.application.webOSVersion < 4.1) { //unsupported
                return dfd.resolve("Unsupported");
            }

            new Power().getPMMode(
                function(result) {
                    // Success
                    logger.info("Power Interface: getPMMode success: " + result.mode);
                    dfd.resolve(result.mode);
                },
                function(result) {
                    // Failed
                    logger.error("Power Interface: Error getting PM Mode: Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                });

            return dfd;
        };

        /**
         * Implements LG setPMMode to set the current power management mode (webOS 4.1+)
         */
        self.setPMMode = function() {
            var dfd = $.Deferred();

            var options = {};
            options.mode = Power.PMMode.ScreenOffAlways;

            // NOTE: Dont need this?
            if (Settings.application.webOSVersion < 4.1) { //unsupported
                return dfd.resolve();
            }

            console.log("Set PM Mode");

            new Power().setPMMode(
                function (result) {
                    // Success
                    logger.info("Power Interface: setPMMode success");
                    dfd.resolve();
                },
                function (result) {
                    // Failed
                    logger.error("Power Interface: Error setting PM Mode: Error [" + result.errorCode + "] " + result.errorText);
                    dfd.reject();
                }, options);

            return dfd;
        };

        //
        //	Returns the screen on / off status
        //
        self.getScreenOnOffStatus = function() {
            var dfd = $.Deferred();

            if(Settings.user.mySettings.usePMModeForScreenManagement && Settings.application.webOSVersion >= 4.1) {
                dfd.resolve(_screenState);
            } else {
                new Power().getPowerStatus(
                    function(result) {
                        // Success
                        logger.info("Power Interface: Display On/Off Status:" + result.displayMode);
                        if (result.displayMode == "Active") {
                            dfd.resolve("on");
                        } else {
                            dfd.resolve("off");
                        }
                    },
                    function(result) {
                        // Failed
                        logger.error("Power Interface: Error getting display On/Off Status: Error [" + result.errorCode + "] " + result.errorText);
                        dfd.reject();
                    }
                );
            }

            return dfd;
        };

        // Reboot - Reboots the Panel
        self.reboot = function(reason, appBuildNumber) {
            var dfd = $.Deferred();
            logger.info('Reboot called. Checking if we should allow the reboot...')
            checkRebootReason(reason, appBuildNumber)
                .done(function(continueWithReboot) {
                    // Skip reboot
                    if (continueWithReboot === false) {
                        logger.info('Skipping reboot');
                        dfd.resolve();
                    }
                    // Do reboot
                    else {
                        logger.info('logging reboot reason...');
                        logRebootReason(reason, appBuildNumber)
                            .done(function() {
                                doReboot();
                            });
                    }


                });
            return dfd;
        };

        /**
         * Screen Control - Store last screen state command sent by WBT API
         * @param  {[type]} screenState The value ("ScreenOn"/"ScreenOff") which should be stored to the filesystem
         */
        function storeLastScreenStatus(screenState) {
            var dfd = $.Deferred();
            var data = {
                screenStatus: screenState,
                screenStatusUpdatedDateTime: new Date()
            }
            var dataString = JSON.stringify(data);
            File.write(Settings.application.fileSystem.systemFiles.screenControlStateFilePath, dataString)
                .done(function() {
                    logger.info('Power Interface: Screen Control: Stored last stored screen control state.');
                }).always(function() {
                    dfd.resolve(); 
                });
            return dfd;
        };

        /**
         * Screen Control - Clear last screen state command stored to the filesystem
         */
        function clearLastScreenState() {
            var dfd = $.Deferred();
            File.write(Settings.application.fileSystem.systemFiles.screenControlStateFilePath, '{}')
            .done(function() {
                logger.info('Power Interface: Screen Control: Cleared last stored screen control state.');
            }).always(function() {
                dfd.resolve(); 
            });
            return dfd;
        };

        /**
         * Screen Control - Read last screen state command stored to the filesystem
         */
        self.readLastScreenStatus = function() {
            var dfd = $.Deferred();
            File.read(Settings.application.fileSystem.systemFiles.screenControlStateFilePath)
                .done(function(data) {
                    if (data) {
                        try {
                            var parsedData = JSON.parse(data);
                            logger.info('Power Interface: Screen Control: Read last stored screen control state: ' + parsedData.screenStatus + ' Last updated at: ' + new Date(parsedData.screenStatusUpdatedDateTime));
                            dfd.resolve(parsedData);
                        }
                        catch (err) {
                            logger.error('Power Interface: Screen Control: Error parsing data from last screen status file: ' + err);
                            File.write(Settings.application.fileSystem.systemFiles.screenControlStateFilePath, '{}')
                                .always(function() {
                                    dfd.resolve();
                                })
                        }
                    } else {
                        logger.info('Power Interface: Screen Control: Unable to read last screen control file.');
                        dfd.resolve();
                    }
                });
            return dfd;
        };

        // Note that this designed to prevent a rare case where
        // LG players get stuck bootlooping because the API reports that the app has updated
        // but the update is never applied
        function checkRebootReason(reason, appBuildNumber) {
            var dfd = $.Deferred();
            // Check if we have a stored last reboot file
            File.read(Settings.application.fileSystem.systemFiles.lastRebootFilePath)
                .done(function(data) {
                    if (data) {
                        try {
                            var parsedData = JSON.parse(data);
                            // Check if the last reboot was an app update for the same version
                            if (parsedData.reason === RebootReasons.APPLICATION_UPDATE && parsedData.reason === reason 
                                && parsedData.appBuildNumber && parsedData.appBuildNumber === appBuildNumber) {
                                // If it was check if it was within the last 24 hours to prevent multiple attempts to update
                                var loggedRebootDateTime = new Date(parsedData.rebootDateTime);
                                loggedRebootDateTime.setHours(loggedRebootDateTime.getHours() + 24);
                                var loggedRebootTime = loggedRebootDateTime.getTime();
                                var currentDateTime = new Date();
                                var currentTime = currentDateTime.getTime();
                                if (loggedRebootTime > currentTime) {
                                    logger.warning('Attempted to reboot but reboot reason is application update, version matches and last logged reboot was less than 24 hours ago.');
                                    dfd.resolve(false);
                                }
                                else {
                                    // Do reboot
                                    logger.info('Last update was a reboot for the same app version but was more than 24 hours ago');
                                    dfd.resolve(true);
                                }
                            }
                            else {
                                // Do reboot
                                logger.info('No reason to block update')
                                dfd.resolve(true);
                            }
                        }
                        catch (err) {
                            // Some error parsing the file - Don't reboot but reset the file
                            logger.error('Error parsing data from last reboot file: ' + err + '. Truncating file...');
                            File.Write(Settings.application.fileSystem.systemFiles.lastRebootFilePath, '{}')
                                .always(function() {
                                    dfd.resolve(false);
                                })
                        }
                    }
                    else {
                        // No file, reboot
                        logger.info('Unable to load or read last reboot file. Continuing with reboot..');
                        dfd.resolve(true);
                    }
                });
            return dfd;
        }

        function logRebootReason(reason, appBuildNumber) {
            if (reason == null) {
                reason = "Unknown";
            }
            var data = {
                reason: reason,
                rebootDateTime: new Date()
            }
            if (appBuildNumber) {
                data.appBuildNumber = appBuildNumber;
            }
            var dataString = JSON.stringify(data);
            return File.write(Settings.application.fileSystem.systemFiles.lastRebootFilePath, dataString);
        }

        function doReboot() {
            var dfd = $.Deferred();
            logger.info('do reboot...');
            new Power().executePowerCommand(
                // Success
                function(restartSuccessResult) {
                    logger.info("Power Interface: Panel rebooted.");
                    dfd.resolve();
                },
                // Failed
                function(restartFailedResult) {
                    logger.error("Power Interface: Failed to reboot panel: [" + restartFailedResult.errorCode + "] " + restartFailedResult.errorText);
                    dfd.reject();
                }, {
                    powerCommand: Power.PowerCommand.REBOOT
                }
            );
            return dfd;
        }

        self.restartApplication = function() {
            return self.reboot();
        };
    }

    // The object is a singleton as there is only a single plugin added to the application
    return new PowerInterface();
});