define(["system/sensors/LunaAPIInterface", "models/GPIOPinModel", "models/GPIOPinState"], function(LunaAPIInterface, GPIOPin, GPIOPinState) {
	
	/**
	 * GPIO Notification Interface
	 * ---------------------------------
	 * Allows other parts of the application to register a callback
	 * when the state of the GPIO pins changes
	 */
	function GPIONotificationInterface() {

		var self = this;

		self.onStateChange = {};

		/**
		 * Register callback that is fired when the idle status changes
		 * Returns Model/IdleState enum
		 * @param {Function}
		 */
		self.onStateChange.addListener = function(callback)
		{
			logger.debug("GPIO Interface: Subscribing to GPIO events");

			var dfd = $.Deferred();

			// Register listener
			var url = "luna://com.webos.service.externaldevice/gpio/readAll";
			var params = JSON.stringify({ "subscribe" : true });
			LunaAPIInterface.call(
				url,
				params,
				// Callback
				function(response){
					logger.debug("GPIO Interface: New GPIO Event. Pin Values: " + response.pinValues);

					// If pinValues isn't returned then this must be
					// the notification regarding initial subscription
					if (typeof response.pinValues == 'undefined')
					{
						if (response.returnValue === true)
						{
							logger.debug("Successfully subsribed to GPIO events");
							dfd.resolve();
						}
						else
						{
							logger.error("Error subscribing to GPIO events");
							dfd.reject();
						}
						
					}
					// We have pin values - this is a state update
					else
					{
						callback(getGPIOPinStates(response.pinValues));
					}
				}
			);

			return dfd;
		};

		/**
		 * Unsubscribes from GPIO events
		 * @return {Promise}
		 */
		self.onStateChange.removeListener = function()
		{
			var dfd = $.Deferred();

			logger.debug("GPIO Interface: Unsubscribing from GPIO events");

			var dfd = $.Deferred();

			setTimeout(function() {
				// Register listener
				var url = "luna://com.webos.service.externaldevice/gpio/readAll";
				var params = JSON.stringify({ "subscribe" : false });
				LunaAPIInterface.call(
					url,
					params,
					// Callback
					function(response){
						if (response.returnValue === true)
						{
							logger.debug("Successfully unsubsribed to GPIO events");
							dfd.resolve();
						}
						else
						{
							logger.error("Error unsubscribing to GPIO events");
							dfd.reject();
						}
					}
				);
				dfd.resolve();
			}, 50);

			return dfd;
		};

		/**
		 * Converts pin values into an array of pins with a pin number and pin state
		 * @param  {int total binary value of all pins}
		 * @return {Array of GPIO Pins}
		 */
		function getGPIOPinStates(pinValues)
		{
			var pins = [];

			// Pin 1
			if ((pinValues & 1) > 0)
			{
				pins[pins.length] = new GPIOPin(0, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(0, GPIOPinState.DOWN);
			}

			// Pin 2
			if ((pinValues & 2) > 0)
			{
				pins[pins.length] = new GPIOPin(1, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(1, GPIOPinState.DOWN);
			}

			// Pin 3
			if ((pinValues & 4) > 0)
			{
				pins[pins.length] = new GPIOPin(2, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(2, GPIOPinState.DOWN);
			}

			// Pin 4
			if ((pinValues & 8) > 0)
			{
				pins[pins.length] = new GPIOPin(3, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(3, GPIOPinState.DOWN);
			}

			// Pin 5
			if ((pinValues & 16) > 0)
			{
				pins[pins.length] = new GPIOPin(4, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(4, GPIOPinState.DOWN);
			}

			// Pin 6
			if ((pinValues & 32) > 0)
			{
				pins[pins.length] = new GPIOPin(5, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(5, GPIOPinState.DOWN);
			}

			// Pin 7
			if ((pinValues & 64) > 0)
			{
				pins[pins.length] = new GPIOPin(6, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(6, GPIOPinState.DOWN);
			}

			// Pin 8
			if ((pinValues & 128) > 0)
			{
				pins[pins.length] = new GPIOPin(7, GPIOPinState.UP);
			}
			else
			{
			    pins[pins.length] = new GPIOPin(7, GPIOPinState.DOWN);
			}

			return pins;
		}

	}
	
	// The object is a singleton as there is only a single inteface added to the application
	return new GPIONotificationInterface();
});